@@ -184,9 +184,9 @@ fn pulley_emit<P>(
             }
 
             if let Some(try_call) = info.try_call_info.as_ref() {
-                sink.add_call_site(&try_call.exception_dests);
+                sink.add_try_call_site(try_call.exception_handlers(&state.frame_layout));
             } else {
-                sink.add_call_site(&[]);
+                sink.add_call_site();
             }
 
             let adjust = -i32::try_from(info.callee_pop_size).unwrap();
@@ -224,9 +224,9 @@ fn pulley_emit<P>(
             }
 
             if let Some(try_call) = info.try_call_info.as_ref() {
-                sink.add_call_site(&try_call.exception_dests);
+                sink.add_try_call_site(try_call.exception_handlers(&state.frame_layout));
             } else {
-                sink.add_call_site(&[]);
+                sink.add_call_site();
             }
 
             let adjust = -i32::try_from(info.callee_pop_size).unwrap();
@@ -288,7 +288,12 @@ fn pulley_emit<P>(
                 let offset = sink.cur_offset();
                 sink.push_user_stack_map(state, offset, s);
             }
-            sink.add_call_site(&[]);
+
+            if let Some(try_call) = info.try_call_info.as_ref() {
+                sink.add_try_call_site(try_call.exception_handlers(&state.frame_layout));
+            } else {
+                sink.add_call_site();
+            }
 
             // If a callee pop is happening here that means that something has
             // messed up, these are expected to be "very simple" signatures.
@@ -555,17 +560,7 @@ fn pulley_emit<P>(
             *start_offset = sink.cur_offset();
         }
 
-        Inst::Raw { raw } => {
-            match raw {
-                RawInst::PushFrame
-                | RawInst::StackAlloc32 { .. }
-                | RawInst::PushFrameSave { .. } => {
-                    sink.add_trap(ir::TrapCode::STACK_OVERFLOW);
-                }
-                _ => {}
-            }
-            super::generated::emit(raw, sink)
-        }
+        Inst::Raw { raw } => super::generated::emit(raw, sink),
 
         Inst::EmitIsland { space_needed } => {
             if sink.island_needed(*space_needed) {
