@@ -7,77 +7,14 @@
 #[doc(hidden)]
 macro_rules! isle_common_prelude_methods {
     () => {
+        isle_numerics_methods!();
+
         /// We don't have a way of making a `()` value in isle directly.
         #[inline]
         fn unit(&mut self) -> Unit {
             ()
         }
 
-        #[inline]
-        fn u8_as_u32(&mut self, x: u8) -> u32 {
-            x.into()
-        }
-
-        #[inline]
-        fn u8_as_u64(&mut self, x: u8) -> u64 {
-            x.into()
-        }
-
-        #[inline]
-        fn u16_as_i16(&mut self, x: u16) -> i16 {
-            x as i16
-        }
-
-        #[inline]
-        fn u16_as_u32(&mut self, x: u16) -> u32 {
-            x.into()
-        }
-
-        #[inline]
-        fn u16_as_u64(&mut self, x: u16) -> u64 {
-            x.into()
-        }
-
-        #[inline]
-        fn u32_as_u64(&mut self, x: u32) -> u64 {
-            x.into()
-        }
-
-        #[inline]
-        fn i64_as_u64(&mut self, x: i64) -> u64 {
-            x as u64
-        }
-
-        #[inline]
-        fn u64_as_i32(&mut self, x: u64) -> i32 {
-            x as i32
-        }
-
-        #[inline]
-        fn u64_as_i64(&mut self, x: u64) -> i64 {
-            x as i64
-        }
-
-        #[inline]
-        fn i32_as_u32(&mut self, x: i32) -> u32 {
-            x as u32
-        }
-
-        #[inline]
-        fn i32_as_i64(&mut self, x: i32) -> i64 {
-            x.into()
-        }
-
-        #[inline]
-        fn i64_neg(&mut self, x: i64) -> i64 {
-            x.wrapping_neg()
-        }
-
-        #[inline]
-        fn i8_neg(&mut self, x: i8) -> i8 {
-            x.wrapping_neg()
-        }
-
         #[inline]
         fn checked_add_with_type(&mut self, ty: Type, a: u64, b: u64) -> Option<u64> {
             let c = a.checked_add(b)?;
@@ -90,21 +27,6 @@ macro_rules! isle_common_prelude_methods {
             self.checked_add_with_type(ty, a, b).is_none()
         }
 
-        #[inline]
-        fn u64_add(&mut self, x: u64, y: u64) -> u64 {
-            x.wrapping_add(y)
-        }
-
-        #[inline]
-        fn u64_sub(&mut self, x: u64, y: u64) -> u64 {
-            x.wrapping_sub(y)
-        }
-
-        #[inline]
-        fn u64_mul(&mut self, x: u64, y: u64) -> u64 {
-            x.wrapping_mul(y)
-        }
-
         #[inline]
         fn imm64_sdiv(&mut self, ty: Type, x: Imm64, y: Imm64) -> Option<Imm64> {
             // Sign extend `x` and `y`.
@@ -127,31 +49,6 @@ macro_rules! isle_common_prelude_methods {
             Some(Imm64::new(result))
         }
 
-        #[inline]
-        fn u64_udiv(&mut self, x: u64, y: u64) -> Option<u64> {
-            x.checked_div(y)
-        }
-
-        #[inline]
-        fn u64_and(&mut self, x: u64, y: u64) -> u64 {
-            x & y
-        }
-
-        #[inline]
-        fn u64_or(&mut self, x: u64, y: u64) -> u64 {
-            x | y
-        }
-
-        #[inline]
-        fn u64_xor(&mut self, x: u64, y: u64) -> u64 {
-            x ^ y
-        }
-
-        #[inline]
-        fn u64_shl(&mut self, x: u64, y: u64) -> u64 {
-            x << y
-        }
-
         #[inline]
         fn imm64_shl(&mut self, ty: Type, x: Imm64, y: Imm64) -> Imm64 {
             // Mask off any excess shift bits.
@@ -191,48 +88,6 @@ macro_rules! isle_common_prelude_methods {
             Imm64::new((x >> y) & ty_mask)
         }
 
-        #[inline]
-        fn u64_not(&mut self, x: u64) -> u64 {
-            !x
-        }
-
-        #[inline]
-        fn u64_eq(&mut self, x: u64, y: u64) -> bool {
-            x == y
-        }
-
-        #[inline]
-        fn u64_le(&mut self, x: u64, y: u64) -> bool {
-            x <= y
-        }
-
-        #[inline]
-        fn u64_lt(&mut self, x: u64, y: u64) -> bool {
-            x < y
-        }
-
-        #[inline]
-        fn u64_is_zero(&mut self, value: u64) -> bool {
-            0 == value
-        }
-
-        fn i64_is_zero(&mut self, value: i64) -> bool {
-            0 == value
-        }
-
-        #[inline]
-        fn u64_is_odd(&mut self, x: u64) -> bool {
-            x & 1 == 1
-        }
-
-        fn i64_shr(&mut self, a: i64, b: i64) -> i64 {
-            a >> b
-        }
-
-        fn i64_ctz(&mut self, a: i64) -> i64 {
-            a.trailing_zeros().into()
-        }
-
         #[inline]
         fn i64_sextend_u64(&mut self, ty: Type, x: u64) -> i64 {
             let shift_amt = std::cmp::max(0, 64 - ty.bits());
@@ -684,51 +539,11 @@ macro_rules! isle_common_prelude_methods {
             }
         }
 
-        #[inline]
-        fn u32_add(&mut self, a: u32, b: u32) -> u32 {
-            a.wrapping_add(b)
-        }
-
-        #[inline]
-        fn u32_sub(&mut self, a: u32, b: u32) -> u32 {
-            a.wrapping_sub(b)
-        }
-
-        #[inline]
-        fn u32_and(&mut self, a: u32, b: u32) -> u32 {
-            a & b
-        }
-
-        #[inline]
-        fn u32_shl(&mut self, x: u32, y: u32) -> u32 {
-            x << y
-        }
-
-        #[inline]
-        fn s32_add_fallible(&mut self, a: i32, b: i32) -> Option<i32> {
-            a.checked_add(b)
-        }
-
         #[inline]
         fn u32_nonnegative(&mut self, x: u32) -> Option<u32> {
             if (x as i32) >= 0 { Some(x) } else { None }
         }
 
-        #[inline]
-        fn u32_lteq(&mut self, a: u32, b: u32) -> Option<()> {
-            if a <= b { Some(()) } else { None }
-        }
-
-        #[inline]
-        fn u8_lteq(&mut self, a: u8, b: u8) -> Option<()> {
-            if a <= b { Some(()) } else { None }
-        }
-
-        #[inline]
-        fn u8_lt(&mut self, a: u8, b: u8) -> Option<()> {
-            if a < b { Some(()) } else { None }
-        }
-
         #[inline]
         fn imm64(&mut self, x: u64) -> Imm64 {
             Imm64::new(x as i64)
@@ -744,26 +559,6 @@ macro_rules! isle_common_prelude_methods {
             x.into()
         }
 
-        #[inline]
-        fn u8_and(&mut self, a: u8, b: u8) -> u8 {
-            a & b
-        }
-
-        #[inline]
-        fn u8_shl(&mut self, a: u8, b: u8) -> u8 {
-            a << b
-        }
-
-        #[inline]
-        fn u8_shr(&mut self, a: u8, b: u8) -> u8 {
-            a >> b
-        }
-
-        #[inline]
-        fn u8_sub(&mut self, a: u8, b: u8) -> u8 {
-            a.wrapping_sub(b)
-        }
-
         #[inline]
         fn lane_type(&mut self, ty: Type) -> Type {
             ty.lane_type()
@@ -902,86 +697,6 @@ macro_rules! isle_common_prelude_methods {
             [a, b]
         }
 
-        fn u128_as_u64(&mut self, val: u128) -> Option<u64> {
-            u64::try_from(val).ok()
-        }
-
-        fn u64_as_u32(&mut self, val: u64) -> Option<u32> {
-            u32::try_from(val).ok()
-        }
-
-        fn u32_as_u16(&mut self, val: u32) -> Option<u16> {
-            val.try_into().ok()
-        }
-
-        fn i32_as_i8(&mut self, val: i32) -> Option<i8> {
-            val.try_into().ok()
-        }
-
-        fn u8_as_i8(&mut self, val: u8) -> i8 {
-            val as i8
-        }
-
-        fn u64_as_u8(&mut self, val: u64) -> u8 {
-            val as u8
-        }
-
-        fn u64_as_u16(&mut self, val: u64) -> u16 {
-            val as u16
-        }
-
-        fn u8_try_from_u64(&mut self, val: u64) -> Option<u8> {
-            u8::try_from(val).ok()
-        }
-
-        fn u8_try_from_u16(&mut self, val: u16) -> Option<u8> {
-            u8::try_from(val).ok()
-        }
-
-        fn u8_try_from_i32(&mut self, val: i32) -> Option<u8> {
-            u8::try_from(val).ok()
-        }
-
-        fn u64_try_from_i64(&mut self, val: i64) -> Option<u64> {
-            u64::try_from(val).ok()
-        }
-
-        fn u16_try_from_i32(&mut self, val: i32) -> Option<u16> {
-            u16::try_from(val).ok()
-        }
-
-        fn u16_try_from_u64(&mut self, val: u64) -> Option<u16> {
-            u16::try_from(val).ok()
-        }
-
-        fn u32_try_from_u64(&mut self, val: u64) -> Option<u32> {
-            u32::try_from(val).ok()
-        }
-
-        fn i8_try_from_i32(&mut self, val: i32) -> Option<i8> {
-            i8::try_from(val).ok()
-        }
-
-        fn i8_try_from_u64(&mut self, val: u64) -> Option<i8> {
-            i8::try_from(val).ok()
-        }
-
-        fn i16_as_u16(&mut self, val: i16) -> u16 {
-            val as u16
-        }
-
-        fn i16_try_from_i32(&mut self, val: i32) -> Option<i16> {
-            i16::try_from(val).ok()
-        }
-
-        fn i16_try_from_u64(&mut self, val: u64) -> Option<i16> {
-            i16::try_from(val).ok()
-        }
-
-        fn i32_try_from_u64(&mut self, val: u64) -> Option<i32> {
-            i32::try_from(val).ok()
-        }
-
         fn u128_replicated_u64(&mut self, val: u128) -> Option<u64> {
             let low64 = val as u64 as u128;
             if (low64 | (low64 << 64)) == val {
