@@ -1,5 +1,5 @@
-use crate::{name_map, Naming, Printer, State};
-use anyhow::{bail, Result};
+use crate::{Naming, Printer, State, name_map};
+use anyhow::{Result, bail};
 use wasmparser::*;
 
 impl Printer<'_, '_> {
@@ -1028,13 +1028,13 @@ impl Printer<'_, '_> {
                         Ok(())
                     })?;
                 }
-                CanonicalFunction::StreamCloseReadable { ty } => {
-                    self.print_intrinsic(state, "canon stream.close-readable ", &|me, state| {
+                CanonicalFunction::StreamDropReadable { ty } => {
+                    self.print_intrinsic(state, "canon stream.drop-readable ", &|me, state| {
                         me.print_idx(&state.component.type_names, ty)
                     })?;
                 }
-                CanonicalFunction::StreamCloseWritable { ty } => {
-                    self.print_intrinsic(state, "canon stream.close-writable ", &|me, state| {
+                CanonicalFunction::StreamDropWritable { ty } => {
+                    self.print_intrinsic(state, "canon stream.drop-writable ", &|me, state| {
                         me.print_idx(&state.component.type_names, ty)
                     })?;
                 }
@@ -1073,13 +1073,13 @@ impl Printer<'_, '_> {
                         Ok(())
                     })?;
                 }
-                CanonicalFunction::FutureCloseReadable { ty } => {
-                    self.print_intrinsic(state, "canon future.close-readable ", &|me, state| {
+                CanonicalFunction::FutureDropReadable { ty } => {
+                    self.print_intrinsic(state, "canon future.drop-readable ", &|me, state| {
                         me.print_idx(&state.component.type_names, ty)
                     })?;
                 }
-                CanonicalFunction::FutureCloseWritable { ty } => {
-                    self.print_intrinsic(state, "canon future.close-writable ", &|me, state| {
+                CanonicalFunction::FutureDropWritable { ty } => {
+                    self.print_intrinsic(state, "canon future.drop-writable ", &|me, state| {
                         me.print_idx(&state.component.type_names, ty)
                     })?;
                 }
