@@ -195,6 +195,40 @@ impl Engine {
             .collect::<Result<Vec<B>, E>>()
     }
 
+    #[cfg(any(feature = "cranelift", feature = "winch"))]
+    pub(crate) fn run_maybe_parallel_mut<
+        T: Send,
+        E: Send,
+        F: Fn(&mut T) -> Result<(), E> + Send + Sync,
+    >(
+        &self,
+        input: &mut [T],
+        f: F,
+    ) -> Result<(), E> {
+        if self.config().parallel_compilation {
+            #[cfg(feature = "parallel-compilation")]
+            {
+                use rayon::prelude::*;
+                // If we collect into `Result<(), E>` directly, the returned
+                // error is not deterministic, because any error could be
+                // returned early. So we first materialize all results in order
+                // and then return the first error deterministically, or
+                // `Ok(_)`.
+                return input
+                    .into_par_iter()
+                    .map(|a| f(a))
+                    .collect::<Vec<Result<(), E>>>()
+                    .into_iter()
+                    .collect::<Result<(), E>>();
+            }
+        }
+
+        // In case the parallel-compilation feature is disabled or the
+        // parallel_compilation config was turned off dynamically fallback to
+        // the non-parallel version.
+        input.into_iter().map(|a| f(a)).collect::<Result<(), E>>()
+    }
+
     /// Take a weak reference to this engine.
     pub fn weak(&self) -> EngineWeak {
         EngineWeak {
@@ -387,6 +421,24 @@ impl Engine {
                 }
             }
 
+            // stack switch model must match the current OS
+            "stack_switch_model" => {
+                if self.features().contains(WasmFeatures::STACK_SWITCHING) {
+                    use target_lexicon::OperatingSystem;
+                    let expected =
+                    match target.operating_system  {
+                        OperatingSystem::Windows => "update_windows_tib",
+                        OperatingSystem::Linux
+                        | OperatingSystem::MacOSX(_)
+                        | OperatingSystem::Darwin(_)  => "basic",
+                        _ => { return Err(String::from("stack-switching feature not supported on this platform")); }
+                    };
+                    *value == FlagValue::Enum(expected)
+                } else {
+                    return Ok(())
+                }
+            }
+
             // These settings don't affect the interface or functionality of
             // the module itself, so their configuration values shouldn't
             // matter.
@@ -405,7 +457,6 @@ impl Engine {
             | "log2_min_function_alignment"
             | "machine_code_cfg_info"
             | "tls_model" // wasmtime doesn't use tls right now
-            | "stack_switch_model" // wasmtime doesn't use stack switching right now
             | "opt_level" // opt level doesn't change semantics
             | "enable_alias_analysis" // alias analysis-based opts don't change semantics
             | "probestack_size_log2" // probestack above asserted disabled
@@ -491,7 +542,9 @@ impl Engine {
 
             // s390x features to detect
             "has_vxrs_ext2" => "vxrs_ext2",
-            "has_mie2" => "mie2",
+            "has_vxrs_ext3" => "vxrs_ext3",
+            "has_mie3" => "mie3",
+            "has_mie4" => "mie4",
 
             // x64 features to detect
             "has_cmpxchg16b" => "cmpxchg16b",
@@ -744,25 +797,6 @@ impl Engine {
         crate::compile::HashedEngineCompileEnv(self)
     }
 
-    /// Executes `f1` and `f2` in parallel if parallel compilation is enabled at
-    /// both runtime and compile time, otherwise runs them synchronously.
-    #[allow(dead_code)] // only used for the component-model feature right now
-    pub(crate) fn join_maybe_parallel<T, U>(
-        &self,
-        f1: impl FnOnce() -> T + Send,
-        f2: impl FnOnce() -> U + Send,
-    ) -> (T, U)
-    where
-        T: Send,
-        U: Send,
-    {
-        if self.config().parallel_compilation {
-            #[cfg(feature = "parallel-compilation")]
-            return rayon::join(f1, f2);
-        }
-        (f1(), f2())
-    }
-
     /// Returns the required alignment for a code image, if we
     /// allocate in a way that is not a system `mmap()` that naturally
     /// aligns it.
