@@ -1,6 +1,6 @@
-use crate::component::ResourceType;
 use crate::component::matching::InstanceType;
 use crate::component::resources::{HostResourceData, HostResourceIndex, HostResourceTables};
+use crate::component::{Instance, ResourceType};
 use crate::prelude::*;
 use crate::runtime::vm::component::{
     CallContexts, ComponentInstance, InstanceFlags, ResourceTable, ResourceTables,
@@ -9,8 +9,12 @@ use crate::runtime::vm::{VMFuncRef, VMMemoryDefinition};
 use crate::store::{StoreId, StoreOpaque};
 use crate::{FuncType, StoreContextMut};
 use alloc::sync::Arc;
+use core::pin::Pin;
 use core::ptr::NonNull;
-use wasmtime_environ::component::{ComponentTypes, StringEncoding, TypeResourceTableIndex};
+use wasmtime_environ::component::{
+    CanonicalOptions, CanonicalOptionsDataModel, ComponentTypes, OptionsIndex, StringEncoding,
+    TypeResourceTableIndex,
+};
 
 /// Runtime representation of canonical ABI options in the component model.
 ///
@@ -22,7 +26,7 @@ use wasmtime_environ::component::{ComponentTypes, StringEncoding, TypeResourceTa
 /// reference to actually use the pointers.
 #[derive(Copy, Clone)]
 pub struct Options {
-    /// The store from which this options originated from.
+    /// The store from which this options originated.
     store_id: StoreId,
 
     /// An optional pointer for the memory that this set of options is referring
@@ -43,6 +47,12 @@ pub struct Options {
     ///
     /// This defaults to utf-8 but can be changed if necessary.
     string_encoding: StringEncoding,
+
+    /// Whether or not this the async option was set when lowering.
+    async_: bool,
+
+    #[cfg(feature = "component-model-async")]
+    callback: Option<NonNull<VMFuncRef>>,
 }
 
 // The `Options` structure stores raw pointers but they're never used unless a
@@ -54,24 +64,39 @@ unsafe impl Sync for Options {}
 impl Options {
     // FIXME(#4311): prevent a ctor where the memory is memory64
 
-    /// Creates a new set of options with the specified components.
+    /// Creates a new [`Options`] from the given [`OptionsIndex`] belonging to
+    /// the specified [`Instance`]
     ///
-    /// # Unsafety
+    /// # Panics
     ///
-    /// This is unsafety as there is no way to statically verify the validity of
-    /// the arguments. For example pointers must be valid pointers, the
-    /// `StoreId` must be valid for the pointers, etc.
-    pub unsafe fn new(
-        store_id: StoreId,
-        memory: Option<NonNull<VMMemoryDefinition>>,
-        realloc: Option<NonNull<VMFuncRef>>,
-        string_encoding: StringEncoding,
-    ) -> Options {
+    /// Panics if `instance` is not owned by `store` or if `index` is not valid
+    /// for `instance`'s component.
+    pub fn new_index(store: &StoreOpaque, instance: Instance, index: OptionsIndex) -> Options {
+        let instance = instance.id().get(store);
+        let CanonicalOptions {
+            string_encoding,
+            async_,
+            callback,
+            ref data_model,
+            ..
+        } = instance.component().env_component().options[index];
+        let (memory, realloc) = match data_model {
+            CanonicalOptionsDataModel::Gc { .. } => (None, None),
+            CanonicalOptionsDataModel::LinearMemory(o) => (o.memory, o.realloc),
+        };
+        let memory = memory.map(|i| NonNull::new(instance.runtime_memory(i)).unwrap());
+        let realloc = realloc.map(|i| instance.runtime_realloc(i));
+        let callback = callback.map(|i| instance.runtime_callback(i));
+        let _ = callback;
+
         Options {
-            store_id,
+            store_id: store.id(),
             memory,
             realloc,
             string_encoding,
+            async_,
+            #[cfg(feature = "component-model-async")]
+            callback,
         }
     }
 
@@ -97,11 +122,6 @@ impl Options {
 
         type ReallocFunc = crate::TypedFunc<(u32, u32, u32, u32), u32>;
 
-        // This call doesn't take any GC refs, and therefore we shouldn't ever
-        // need to GC before entering Wasm.
-        #[cfg(feature = "gc")]
-        debug_assert!(!ReallocFunc::need_gc_before_call_raw(store.0, &params));
-
         // Invoke the wasm malloc function using its raw and statically known
         // signature.
         let result = unsafe { ReallocFunc::call_raw(store, realloc_ty, realloc, params)? };
@@ -164,6 +184,21 @@ impl Options {
     pub fn store_id(&self) -> StoreId {
         self.store_id
     }
+
+    /// Returns whether this lifting or lowering uses the async ABI.
+    pub fn async_(&self) -> bool {
+        self.async_
+    }
+
+    #[cfg(feature = "component-model-async")]
+    pub(crate) fn callback(&self) -> Option<NonNull<VMFuncRef>> {
+        self.callback
+    }
+
+    #[cfg(feature = "component-model-async")]
+    pub(crate) fn memory_raw(&self) -> Option<NonNull<VMMemoryDefinition>> {
+        self.memory
+    }
 }
 
 /// A helper structure which is a "package" of the context used during lowering
@@ -193,36 +228,71 @@ pub struct LowerContext<'a, T: 'static> {
     /// lifting/lowering process.
     pub types: &'a ComponentTypes,
 
-    /// A raw unsafe pointer to the component instance that's being lowered
-    /// into.
-    ///
-    /// This pointer is required to be owned by the `store` provided.
-    instance: *mut ComponentInstance,
+    /// Index of the component instance that's being lowered into.
+    instance: Instance,
+
+    /// Whether to allow `options.realloc` to be used when lowering.
+    allow_realloc: bool,
 }
 
 #[doc(hidden)]
 impl<'a, T: 'static> LowerContext<'a, T> {
     /// Creates a new lowering context from the specified parameters.
-    ///
-    /// # Unsafety
-    ///
-    /// This function is unsafe as it needs to be guaranteed by the caller that
-    /// the `instance` here is valid within `store` and is a valid component
-    /// instance.
-    pub unsafe fn new(
+    pub fn new(
         store: StoreContextMut<'a, T>,
         options: &'a Options,
         types: &'a ComponentTypes,
-        instance: *mut ComponentInstance,
+        instance: Instance,
+    ) -> LowerContext<'a, T> {
+        #[cfg(all(debug_assertions, feature = "component-model-async"))]
+        if store.engine().config().async_support {
+            // Assert that we're running on a fiber, which is necessary in
+            // case we call the guest's realloc function.
+            store.0.with_blocking(|_, _| {});
+        }
+        LowerContext {
+            store,
+            options,
+            types,
+            instance,
+            allow_realloc: true,
+        }
+    }
+
+    /// Like `new`, except disallows use of `options.realloc`.
+    ///
+    /// The returned object will panic if its `realloc` method is called.
+    ///
+    /// This is meant for use when lowering "flat" values (i.e. values which
+    /// require no allocations) into already-allocated memory or into stack
+    /// slots, in which case the lowering may safely be done outside of a fiber
+    /// since there is no need to make any guest calls.
+    #[cfg(feature = "component-model-async")]
+    pub(crate) fn new_without_realloc(
+        store: StoreContextMut<'a, T>,
+        options: &'a Options,
+        types: &'a ComponentTypes,
+        instance: Instance,
     ) -> LowerContext<'a, T> {
         LowerContext {
             store,
             options,
             types,
             instance,
+            allow_realloc: false,
         }
     }
 
+    /// Returns the `&ComponentInstance` that's being lowered into.
+    pub fn instance(&self) -> &ComponentInstance {
+        self.instance.id().get(self.store.0)
+    }
+
+    /// Returns the `&mut ComponentInstance` that's being lowered into.
+    pub fn instance_mut(&mut self) -> Pin<&mut ComponentInstance> {
+        self.instance.id().get_mut(self.store.0)
+    }
+
     /// Returns a view into memory as a mutable slice of bytes.
     ///
     /// # Panics
@@ -247,7 +317,9 @@ impl<'a, T: 'static> LowerContext<'a, T> {
         old_align: u32,
         new_size: usize,
     ) -> Result<usize> {
-        let realloc_func_ty = Arc::clone(unsafe { (*self.instance).component().realloc_func_ty() });
+        assert!(self.allow_realloc);
+
+        let realloc_func_ty = Arc::clone(self.instance().component().realloc_func_ty());
         self.options
             .realloc(
                 &mut self.store,
@@ -311,10 +383,7 @@ impl<'a, T: 'static> LowerContext<'a, T> {
         // This check is performed by comparing the owning instance of `ty`
         // against the owning instance of the resource that `ty` is working
         // with.
-        //
-        // Note that the unsafety here should be valid given the contract of
-        // `LowerContext::new`.
-        if unsafe { (*self.instance).resource_owned_by_own_instance(ty) } {
+        if self.instance().resource_owned_by_own_instance(ty) {
             return Ok(rep);
         }
         self.resource_tables().guest_resource_lower_borrow(rep, ty)
@@ -355,20 +424,19 @@ impl<'a, T: 'static> LowerContext<'a, T> {
     /// Returns the instance type information corresponding to the instance that
     /// this context is lowering into.
     pub fn instance_type(&self) -> InstanceType<'_> {
-        // Note that the unsafety here should be valid given the contract of
-        // `LowerContext::new`.
-        InstanceType::new(unsafe { &*self.instance })
+        InstanceType::new(self.instance())
     }
 
     fn resource_tables(&mut self) -> HostResourceTables<'_> {
-        let (calls, host_table, host_resource_data) = self.store.0.component_resource_state();
+        let (calls, host_table, host_resource_data, instance) = self
+            .store
+            .0
+            .component_resource_state_with_instance(self.instance);
         HostResourceTables::from_parts(
             ResourceTables {
                 host_table: Some(host_table),
                 calls,
-                // Note that the unsafety here should be valid given the contract of
-                // `LowerContext::new`.
-                guest: Some(unsafe { (*self.instance).guest_tables() }),
+                guest: Some(instance.guest_tables()),
             },
             host_resource_data,
         )
@@ -402,7 +470,8 @@ pub struct LiftContext<'a> {
 
     memory: Option<&'a [u8]>,
 
-    instance: *mut ComponentInstance,
+    instance: Pin<&'a mut ComponentInstance>,
+    instance_handle: Instance,
 
     host_table: &'a mut ResourceTable,
     host_resource_data: &'a mut HostResourceData,
@@ -413,17 +482,11 @@ pub struct LiftContext<'a> {
 #[doc(hidden)]
 impl<'a> LiftContext<'a> {
     /// Creates a new lifting context given the provided context.
-    ///
-    /// # Unsafety
-    ///
-    /// This is unsafe for the same reasons as `LowerContext::new` where the
-    /// validity of `instance` is required to be upheld by the caller.
     #[inline]
-    pub unsafe fn new(
+    pub fn new(
         store: &'a mut StoreOpaque,
         options: &'a Options,
-        types: &'a Arc<ComponentTypes>,
-        instance: *mut ComponentInstance,
+        instance_handle: Instance,
     ) -> LiftContext<'a> {
         // From `&mut StoreOpaque` provided the goal here is to project out
         // three different disjoint fields owned by the store: memory,
@@ -431,15 +494,19 @@ impl<'a> LiftContext<'a> {
         // so it's hacked around a bit. This unsafe pointer cast could be fixed
         // with more methods in more places, but it doesn't seem worth doing it
         // at this time.
-        let (calls, host_table, host_resource_data) =
-            (&mut *(store as *mut StoreOpaque)).component_resource_state();
-        let memory = options.memory.map(|_| options.memory(store));
+        let memory = options
+            .memory
+            .map(|_| options.memory(unsafe { &*(store as *const StoreOpaque) }));
+        let (calls, host_table, host_resource_data, instance) =
+            store.component_resource_state_with_instance(instance_handle);
+        let (component, instance) = instance.component_and_self();
 
         LiftContext {
             memory,
             options,
-            types,
+            types: component.types(),
             instance,
+            instance_handle,
             calls,
             host_table,
             host_resource_data,
@@ -463,9 +530,13 @@ impl<'a> LiftContext<'a> {
         self.options.store_id
     }
 
-    /// Returns the component instance raw pointer that is being lifted from.
-    pub fn instance_ptr(&self) -> *mut ComponentInstance {
-        self.instance
+    /// Returns the component instance that is being lifted from.
+    pub fn instance_mut(&mut self) -> Pin<&mut ComponentInstance> {
+        self.instance.as_mut()
+    }
+    /// Returns the component instance that is being lifted from.
+    pub fn instance_handle(&self) -> Instance {
+        self.instance_handle
     }
 
     /// Lifts an `own` resource from the guest at the `idx` specified into its
@@ -479,9 +550,7 @@ impl<'a> LiftContext<'a> {
         idx: u32,
     ) -> Result<(u32, Option<NonNull<VMFuncRef>>, Option<InstanceFlags>)> {
         let idx = self.resource_tables().guest_resource_lift_own(idx, ty)?;
-        // Note that the unsafety here should be valid given the contract of
-        // `LiftContext::new`.
-        let (dtor, flags) = unsafe { (*self.instance).dtor_and_flags(ty) };
+        let (dtor, flags) = self.instance.dtor_and_flags(ty);
         Ok((idx, dtor, flags))
     }
 
@@ -520,9 +589,7 @@ impl<'a> LiftContext<'a> {
     /// Returns instance type information for the component instance that is
     /// being lifted from.
     pub fn instance_type(&self) -> InstanceType<'_> {
-        // Note that the unsafety here should be valid given the contract of
-        // `LiftContext::new`.
-        InstanceType::new(unsafe { &*self.instance })
+        InstanceType::new(&self.instance)
     }
 
     fn resource_tables(&mut self) -> HostResourceTables<'_> {
@@ -532,7 +599,7 @@ impl<'a> LiftContext<'a> {
                 calls: self.calls,
                 // Note that the unsafety here should be valid given the contract of
                 // `LiftContext::new`.
-                guest: Some(unsafe { (*self.instance).guest_tables() }),
+                guest: Some(self.instance.as_mut().guest_tables()),
             },
             self.host_resource_data,
         )
