@@ -8,15 +8,18 @@ use crate::component::{
 use crate::instance::OwnedImports;
 use crate::linker::DefinitionType;
 use crate::prelude::*;
-use crate::runtime::vm::VMFuncRef;
-use crate::runtime::vm::component::{ComponentInstance, OwnedComponentInstance};
+use crate::runtime::vm::component::{
+    CallContexts, ComponentInstance, ResourceTables, TypedResource, TypedResourceIndex,
+};
+use crate::runtime::vm::{self, VMFuncRef};
 use crate::store::StoreOpaque;
 use crate::{AsContext, AsContextMut, Engine, Module, StoreContextMut};
 use alloc::sync::Arc;
 use core::marker;
+use core::pin::Pin;
 use core::ptr::NonNull;
 use wasmtime_environ::{EngineOrModuleTypeIndex, component::*};
-use wasmtime_environ::{EntityType, PrimaryMap};
+use wasmtime_environ::{EntityIndex, EntityType, PrimaryMap};
 
 /// An instantiated component.
 ///
@@ -50,12 +53,7 @@ const _: () = {
 
 impl Instance {
     /// Creates a raw `Instance` from the internal identifiers within the store.
-    ///
-    /// # Safety
-    ///
-    /// The safety of this function relies on `id` belonging to the instance
-    /// within `store`.
-    pub(crate) unsafe fn from_wasmtime(store: &StoreOpaque, id: ComponentInstanceId) -> Instance {
+    pub(crate) fn from_wasmtime(store: &StoreOpaque, id: ComponentInstanceId) -> Instance {
         Instance {
             id: StoreComponentInstanceId::new(store.id(), id),
         }
@@ -161,7 +159,7 @@ impl Instance {
         name: impl InstanceExportLookup,
     ) -> Option<Func> {
         let store = store.as_context_mut().0;
-        let instance = &store[self.id];
+        let instance = self.id.get(store);
         let component = instance.component();
 
         // Validate that `name` exists within `self.`
@@ -304,7 +302,7 @@ impl Instance {
         instance: Option<&ComponentExportIndex>,
         name: &str,
     ) -> Option<(ComponentItem, ComponentExportIndex)> {
-        let data = self.instance(store);
+        let data = self.id().get(store);
         let component = data.component();
         let index = component.lookup_export_index(instance, name)?;
         let item = ComponentItem::from_export(
@@ -340,7 +338,7 @@ impl Instance {
         instance: Option<&ComponentExportIndex>,
         name: &str,
     ) -> Option<ComponentExportIndex> {
-        let data = self.instance(store.as_context_mut().0);
+        let data = self.id().get(store.as_context_mut().0);
         let index = data.component().lookup_export_index(instance, name)?;
         Some(ComponentExportIndex {
             id: data.component().id(),
@@ -353,7 +351,7 @@ impl Instance {
         store: &'a StoreOpaque,
         name: impl InstanceExportLookup,
     ) -> Option<(&'a ComponentInstance, &'a Export)> {
-        let data = self.instance(store);
+        let data = self.id().get(store);
         let index = name.lookup(data.component())?;
         Some((data, &data.component().env_component().export_items[index]))
     }
@@ -362,7 +360,7 @@ impl Instance {
     pub fn instance_pre<T>(&self, store: impl AsContext<Data = T>) -> InstancePre<T> {
         // This indexing operation asserts the Store owns the Instance.
         // Therefore, the InstancePre<T> must match the Store<T>.
-        let data = self.instance(store.as_context().0);
+        let data = self.id().get(store.as_context().0);
 
         // SAFETY: calling this method safely here relies on matching the `T`
         // in `InstancePre<T>` to the store itself, which is happening in the
@@ -371,22 +369,167 @@ impl Instance {
         unsafe { data.instance_pre() }
     }
 
-    /// Returns the VM/runtime state for this instance as belonging to the
-    /// store provided.
-    pub(crate) fn instance<'a>(&self, store: &'a StoreOpaque) -> &'a ComponentInstance {
-        &store[self.id]
-    }
-
-    /// Returns the VM/runtime state for this instance as belonging to the
-    /// store provided.
-    pub(crate) fn instance_ptr(&self, store: &StoreOpaque) -> NonNull<ComponentInstance> {
-        self.id.assert_belongs_to(store.id());
-        store.component_instance_ptr(self.id.instance())
-    }
-
     pub(crate) fn id(&self) -> StoreComponentInstanceId {
         self.id
     }
+
+    /// Implementation of the `resource.new` intrinsic for `i32`
+    /// representations.
+    pub(crate) fn resource_new32(
+        self,
+        store: &mut StoreOpaque,
+        ty: TypeResourceTableIndex,
+        rep: u32,
+    ) -> Result<u32> {
+        let (calls, _, _, instance) = store.component_resource_state_with_instance(self);
+        resource_tables(calls, instance).resource_new(TypedResource::Component { ty, rep })
+    }
+
+    /// Implementation of the `resource.rep` intrinsic for `i32`
+    /// representations.
+    pub(crate) fn resource_rep32(
+        self,
+        store: &mut StoreOpaque,
+        ty: TypeResourceTableIndex,
+        index: u32,
+    ) -> Result<u32> {
+        let (calls, _, _, instance) = store.component_resource_state_with_instance(self);
+        resource_tables(calls, instance).resource_rep(TypedResourceIndex::Component { ty, index })
+    }
+
+    /// Implementation of the `resource.drop` intrinsic.
+    pub(crate) fn resource_drop(
+        self,
+        store: &mut StoreOpaque,
+        ty: TypeResourceTableIndex,
+        index: u32,
+    ) -> Result<Option<u32>> {
+        let (calls, _, _, instance) = store.component_resource_state_with_instance(self);
+        resource_tables(calls, instance).resource_drop(TypedResourceIndex::Component { ty, index })
+    }
+
+    pub(crate) fn resource_transfer_own(
+        self,
+        store: &mut StoreOpaque,
+        index: u32,
+        src: TypeResourceTableIndex,
+        dst: TypeResourceTableIndex,
+    ) -> Result<u32> {
+        let (calls, _, _, instance) = store.component_resource_state_with_instance(self);
+        let mut tables = resource_tables(calls, instance);
+        let rep = tables.resource_lift_own(TypedResourceIndex::Component { ty: src, index })?;
+        tables.resource_lower_own(TypedResource::Component { ty: dst, rep })
+    }
+
+    pub(crate) fn resource_transfer_borrow(
+        self,
+        store: &mut StoreOpaque,
+        index: u32,
+        src: TypeResourceTableIndex,
+        dst: TypeResourceTableIndex,
+    ) -> Result<u32> {
+        let dst_owns_resource = self.id().get(store).resource_owned_by_own_instance(dst);
+        let (calls, _, _, instance) = store.component_resource_state_with_instance(self);
+        let mut tables = resource_tables(calls, instance);
+        let rep = tables.resource_lift_borrow(TypedResourceIndex::Component { ty: src, index })?;
+        // Implement `lower_borrow`'s special case here where if a borrow's
+        // resource type is owned by `dst` then the destination receives the
+        // representation directly rather than a handle to the representation.
+        //
+        // This can perhaps become a different libcall in the future to avoid
+        // this check at runtime since we know at compile time whether the
+        // destination type owns the resource, but that's left as a future
+        // refactoring if truly necessary.
+        if dst_owns_resource {
+            return Ok(rep);
+        }
+        tables.resource_lower_borrow(TypedResource::Component { ty: dst, rep })
+    }
+
+    pub(crate) fn resource_enter_call(self, store: &mut StoreOpaque) {
+        let (calls, _, _, instance) = store.component_resource_state_with_instance(self);
+        resource_tables(calls, instance).enter_call()
+    }
+
+    pub(crate) fn resource_exit_call(self, store: &mut StoreOpaque) -> Result<()> {
+        let (calls, _, _, instance) = store.component_resource_state_with_instance(self);
+        resource_tables(calls, instance).exit_call()
+    }
+
+    pub(crate) fn lookup_vmdef(&self, store: &mut StoreOpaque, def: &CoreDef) -> vm::Export {
+        lookup_vmdef(store, self.id.instance(), def)
+    }
+}
+
+/// Translates a `CoreDef`, a definition of a core wasm item, to an
+/// [`Export`] which is the runtime core wasm definition.
+pub(crate) fn lookup_vmdef(
+    store: &mut StoreOpaque,
+    id: ComponentInstanceId,
+    def: &CoreDef,
+) -> vm::Export {
+    match def {
+        CoreDef::Export(e) => lookup_vmexport(store, id, e),
+        CoreDef::Trampoline(idx) => {
+            let funcref = store
+                .store_data_mut()
+                .component_instance_mut(id)
+                .trampoline_func_ref(*idx);
+            // SAFETY: the `funcref` is owned by `store` and is valid within
+            // that store, so it's safe to create a `Func`.
+            vm::Export::Function(unsafe { crate::Func::from_vm_func_ref(store.id(), funcref) })
+        }
+        CoreDef::InstanceFlags(idx) => {
+            let id = StoreComponentInstanceId::new(store.id(), id);
+            vm::Export::Global(crate::Global::from_component_flags(id, *idx))
+        }
+    }
+}
+
+/// Translates a `CoreExport<T>`, an export of some core instance within
+/// this component, to the actual runtime definition of that item.
+pub(crate) fn lookup_vmexport<T>(
+    store: &mut StoreOpaque,
+    id: ComponentInstanceId,
+    item: &CoreExport<T>,
+) -> vm::Export
+where
+    T: Copy + Into<EntityIndex>,
+{
+    let store_id = store.id();
+    let id = store
+        .store_data_mut()
+        .component_instance_mut(id)
+        .instance(item.instance);
+    let instance = store.instance_mut(id);
+    let idx = match &item.item {
+        ExportItem::Index(idx) => (*idx).into(),
+
+        // FIXME: ideally at runtime we don't actually do any name lookups
+        // here. This will only happen when the host supplies an imported
+        // module so while the structure can't be known at compile time we
+        // do know at `InstancePre` time, for example, what all the host
+        // imports are. In theory we should be able to, as part of
+        // `InstancePre` construction, perform all name=>index mappings
+        // during that phase so the actual instantiation of an `InstancePre`
+        // skips all string lookups. This should probably only be
+        // investigated if this becomes a performance issue though.
+        ExportItem::Name(name) => instance.env_module().exports[name],
+    };
+    // SAFETY: the `store_id` owns this instance and all exports contained
+    // within.
+    unsafe { instance.get_export_by_index_mut(store_id, idx) }
+}
+
+fn resource_tables<'a>(
+    calls: &'a mut CallContexts,
+    instance: Pin<&'a mut ComponentInstance>,
+) -> ResourceTables<'a> {
+    ResourceTables {
+        host_table: None,
+        calls,
+        guest: Some(instance.guest_tables()),
+    }
 }
 
 /// Trait used to lookup the export of a component instance.
@@ -432,7 +575,7 @@ impl InstanceExportLookup for String {
 
 struct Instantiator<'a> {
     component: &'a Component,
-    instance: OwnedComponentInstance,
+    id: ComponentInstanceId,
     core_imports: OwnedImports,
     imports: &'a PrimaryMap<RuntimeImportIndex, RuntimeImport>,
 }
@@ -473,17 +616,21 @@ impl<'a> Instantiator<'a> {
         store.modules_mut().register_component(component);
         let imported_resources: ImportedResources =
             PrimaryMap::with_capacity(env_component.imported_resources.len());
+
+        let instance = ComponentInstance::new(
+            store.store_data().components.next_component_instance_id(),
+            component,
+            Arc::new(imported_resources),
+            imports,
+            store.traitobj(),
+        );
+        let id = store.store_data_mut().push_component_instance(instance);
+
         Instantiator {
             component,
             imports,
             core_imports: OwnedImports::empty(),
-            instance: OwnedComponentInstance::new(
-                store.store_data().components.next_component_instance_id(),
-                component,
-                Arc::new(imported_resources),
-                imports,
-                store.traitobj(),
-            ),
+            id,
         }
     }
 
@@ -500,9 +647,10 @@ impl<'a> Instantiator<'a> {
                 } => (*ty, NonNull::from(dtor_funcref)),
                 _ => unreachable!(),
             };
-            let i = self.instance_resource_types_mut().push(ty);
+            let i = self.instance_resource_types_mut(store.0).push(ty);
             assert_eq!(i, idx);
-            self.instance.set_resource_destructor(idx, Some(func_ref));
+            self.instance_mut(store.0)
+                .set_resource_destructor(idx, Some(func_ref));
         }
 
         // Next configure all `VMFuncRef`s for trampolines that this component
@@ -516,8 +664,12 @@ impl<'a> Instantiator<'a> {
                 None => panic!("found unregistered signature: {sig:?}"),
             };
 
-            self.instance
-                .set_trampoline(idx, ptrs.wasm_call, ptrs.array_call, signature);
+            self.instance_mut(store.0).set_trampoline(
+                idx,
+                ptrs.wasm_call,
+                ptrs.array_call,
+                signature,
+            );
         }
 
         for initializer in env_component.initializers.iter() {
@@ -564,7 +716,7 @@ impl<'a> Instantiator<'a> {
                     let i = unsafe {
                         crate::Instance::new_started_impl(store, module, imports.as_ref())?
                     };
-                    self.instance.push_instance_id(i.id());
+                    self.instance_mut(store.0).push_instance_id(i.id());
                 }
 
                 GlobalInitializer::LowerImport { import, index } => {
@@ -572,7 +724,8 @@ impl<'a> Instantiator<'a> {
                         RuntimeImport::Func(func) => func,
                         _ => unreachable!(),
                     };
-                    self.instance.set_lowering(*index, func.lowering());
+                    self.instance_mut(store.0)
+                        .set_lowering(*index, func.lowering());
                 }
 
                 GlobalInitializer::ExtractTable(table) => self.extract_table(store.0, table),
@@ -597,71 +750,77 @@ impl<'a> Instantiator<'a> {
         Ok(())
     }
 
-    fn resource(&mut self, store: &StoreOpaque, resource: &Resource) {
+    fn resource(&mut self, store: &mut StoreOpaque, resource: &Resource) {
         let dtor = resource
             .dtor
             .as_ref()
-            .map(|dtor| self.instance.lookup_def(store, dtor));
+            .map(|dtor| lookup_vmdef(store, self.id, dtor));
         let dtor = dtor.map(|export| match export {
-            crate::runtime::vm::Export::Function(f) => f.func_ref,
+            crate::runtime::vm::Export::Function(f) => f.vm_func_ref(store),
             _ => unreachable!(),
         });
         let index = self
             .component
             .env_component()
             .resource_index(resource.index);
-        self.instance.set_resource_destructor(index, dtor);
-        let ty = ResourceType::guest(store.id(), &self.instance, resource.index);
-        let i = self.instance_resource_types_mut().push(ty);
+        let instance = self.instance(store);
+        let ty = ResourceType::guest(store.id(), instance, resource.index);
+        self.instance_mut(store)
+            .set_resource_destructor(index, dtor);
+        let i = self.instance_resource_types_mut(store).push(ty);
         debug_assert_eq!(i, index);
     }
 
-    fn extract_memory(&mut self, store: &StoreOpaque, memory: &ExtractMemory) {
-        let mem = match self.instance.lookup_export(store, &memory.export) {
-            crate::runtime::vm::Export::Memory(m) => m,
+    fn extract_memory(&mut self, store: &mut StoreOpaque, memory: &ExtractMemory) {
+        let mem = match lookup_vmexport(store, self.id, &memory.export) {
+            crate::runtime::vm::Export::Memory { memory, .. } => memory,
             _ => unreachable!(),
         };
-        self.instance
-            .set_runtime_memory(memory.index, mem.definition);
+        let import = mem.vmimport(store);
+        self.instance_mut(store)
+            .set_runtime_memory(memory.index, import.from.as_non_null());
     }
 
-    fn extract_realloc(&mut self, store: &StoreOpaque, realloc: &ExtractRealloc) {
-        let func_ref = match self.instance.lookup_def(store, &realloc.def) {
-            crate::runtime::vm::Export::Function(f) => f.func_ref,
+    fn extract_realloc(&mut self, store: &mut StoreOpaque, realloc: &ExtractRealloc) {
+        let func_ref = match lookup_vmdef(store, self.id, &realloc.def) {
+            crate::runtime::vm::Export::Function(f) => f.vm_func_ref(store),
             _ => unreachable!(),
         };
-        self.instance.set_runtime_realloc(realloc.index, func_ref);
+        self.instance_mut(store)
+            .set_runtime_realloc(realloc.index, func_ref);
     }
 
-    fn extract_callback(&mut self, store: &StoreOpaque, callback: &ExtractCallback) {
-        let func_ref = match self.instance.lookup_def(store, &callback.def) {
-            crate::runtime::vm::Export::Function(f) => f.func_ref,
+    fn extract_callback(&mut self, store: &mut StoreOpaque, callback: &ExtractCallback) {
+        let func_ref = match lookup_vmdef(store, self.id, &callback.def) {
+            crate::runtime::vm::Export::Function(f) => f.vm_func_ref(store),
             _ => unreachable!(),
         };
-        self.instance.set_runtime_callback(callback.index, func_ref);
+        self.instance_mut(store)
+            .set_runtime_callback(callback.index, func_ref);
     }
 
-    fn extract_post_return(&mut self, store: &StoreOpaque, post_return: &ExtractPostReturn) {
-        let func_ref = match self.instance.lookup_def(store, &post_return.def) {
-            crate::runtime::vm::Export::Function(f) => f.func_ref,
+    fn extract_post_return(&mut self, store: &mut StoreOpaque, post_return: &ExtractPostReturn) {
+        let func_ref = match lookup_vmdef(store, self.id, &post_return.def) {
+            crate::runtime::vm::Export::Function(f) => f.vm_func_ref(store),
             _ => unreachable!(),
         };
-        self.instance
+        self.instance_mut(store)
             .set_runtime_post_return(post_return.index, func_ref);
     }
 
-    fn extract_table(&mut self, store: &StoreOpaque, table: &ExtractTable) {
-        let export = match self.instance.lookup_export(store, &table.export) {
+    fn extract_table(&mut self, store: &mut StoreOpaque, table: &ExtractTable) {
+        let export = match lookup_vmexport(store, self.id, &table.export) {
             crate::runtime::vm::Export::Table(t) => t,
             _ => unreachable!(),
         };
-        self.instance
-            .set_runtime_table(table.index, export.definition, export.vmctx, export.index);
+        let import = export.vmimport(store);
+        self.instance_mut(store)
+            .set_runtime_table(table.index, import);
     }
 
     fn build_imports<'b>(
         &mut self,
-        store: &StoreOpaque,
+        store: &mut StoreOpaque,
         module: &Module,
         args: impl Iterator<Item = &'b CoreDef>,
     ) -> &OwnedImports {
@@ -683,10 +842,8 @@ impl<'a> Instantiator<'a> {
             // The unsafety here should be ok since the `export` is loaded
             // directly from an instance which should only give us valid export
             // items.
-            let export = self.instance.lookup_def(store, arg);
-            unsafe {
-                self.core_imports.push_export(&export);
-            }
+            let export = lookup_vmdef(store, self.id, arg);
+            self.core_imports.push_export(store, &export);
         }
         debug_assert!(imports.next().is_none());
 
@@ -695,14 +852,14 @@ impl<'a> Instantiator<'a> {
 
     fn assert_type_matches(
         &self,
-        store: &StoreOpaque,
+        store: &mut StoreOpaque,
         module: &Module,
         arg: &CoreDef,
         imp_module: &str,
         imp_name: &str,
         expected: EntityType,
     ) {
-        let export = self.instance.lookup_def(store, arg);
+        let export = lookup_vmdef(store, self.id, arg);
 
         // If this value is a core wasm function then the type check is inlined
         // here. This can otherwise fail `Extern::from_wasmtime_export` because
@@ -714,7 +871,7 @@ impl<'a> Instantiator<'a> {
                 EngineOrModuleTypeIndex::Module(m) => module.signatures().shared_type(m),
                 EngineOrModuleTypeIndex::RecGroup(_) => unreachable!(),
             };
-            let actual = unsafe { f.func_ref.as_ref().type_index };
+            let actual = unsafe { f.vm_func_ref(store).as_ref().type_index };
             assert_eq!(
                 expected,
                 Some(actual),
@@ -734,13 +891,27 @@ impl<'a> Instantiator<'a> {
             .expect("unexpected typecheck failure");
     }
 
+    /// Convenience helper to return the `&ComponentInstance` that's being
+    /// instantiated.
+    fn instance<'b>(&self, store: &'b StoreOpaque) -> &'b ComponentInstance {
+        store.store_data().component_instance(self.id)
+    }
+
+    /// Same as [`Self::instance`], but for mutability.
+    fn instance_mut<'b>(&self, store: &'b mut StoreOpaque) -> Pin<&'b mut ComponentInstance> {
+        store.store_data_mut().component_instance_mut(self.id)
+    }
+
     // NB: This method is only intended to be called during the instantiation
     // process because the `Arc::get_mut` here is fallible and won't generally
     // succeed once the instance has been handed to the embedder. Before that
     // though it should be guaranteed that the single owning reference currently
     // lives within the `ComponentInstance` that's being built.
-    fn instance_resource_types_mut(&mut self) -> &mut ImportedResources {
-        Arc::get_mut(self.instance.resource_types_mut()).unwrap()
+    fn instance_resource_types_mut<'b>(
+        &self,
+        store: &'b mut StoreOpaque,
+    ) -> &'b mut ImportedResources {
+        Arc::get_mut(self.instance_mut(store).resource_types_mut()).unwrap()
     }
 }
 
@@ -857,14 +1028,7 @@ impl<T: 'static> InstancePre<T> {
                 .decrement_component_instance_count();
             e
         })?;
-        let id = store
-            .0
-            .store_data_mut()
-            .push_component_instance(instantiator.instance);
-        // SAFETY: `from_wasmtime` requires that `id` belongs to the store
-        // provided, and it was just inserted above so the condition should be
-        // satisfied.
-        let instance = unsafe { Instance::from_wasmtime(store.0, id) };
+        let instance = Instance::from_wasmtime(store.0, instantiator.id);
         store.0.push_component_instance(instance);
         Ok(instance)
     }
