@@ -1,3 +1,5 @@
+#[cfg(feature = "component-model-async")]
+use crate::component::concurrent::Accessor;
 use crate::component::func::HostFunc;
 use crate::component::instance::RuntimeImport;
 use crate::component::matching::{InstanceType, TypeChecker};
@@ -390,7 +392,7 @@ impl<T: 'static> LinkerInstance<'_, T> {
         }
     }
 
-    /// Defines a new host-provided function into this [`Linker`].
+    /// Defines a new host-provided function into this [`LinkerInstance`].
     ///
     /// This method is used to give host functions to wasm components. The
     /// `func` provided will be callable from linked components with the type
@@ -406,6 +408,19 @@ impl<T: 'static> LinkerInstance<'_, T> {
     /// type parameter from [`Store<T>`](crate::Store) which is accessible
     /// through the leading [`StoreContextMut<'_, T>`](crate::StoreContextMut)
     /// argument which can be provided to the `func` given here.
+    ///
+    /// # Blocking / Async Behavior
+    ///
+    /// The host function `func` provided here is a blocking function from the
+    /// perspective of WebAssembly, regardless of how [`Config::async_support`]
+    /// is defined. This function can be used both with sync and async stores
+    /// and will always define a blocking function.
+    ///
+    /// To define a function which is async on the host, but blocking to the
+    /// guest, see the [`func_wrap_async`] method.
+    ///
+    /// [`Config::async_support`]: crate::Config::async_support
+    /// [`func_wrap_async`]: LinkerInstance::func_wrap_async
     //
     // TODO: needs more words and examples
     pub fn func_wrap<F, Params, Return>(&mut self, name: &str, func: F) -> Result<()>
@@ -418,17 +433,48 @@ impl<T: 'static> LinkerInstance<'_, T> {
         Ok(())
     }
 
-    /// Defines a new host-provided async function into this [`Linker`].
+    /// Defines a new host-provided async function into this [`LinkerInstance`].
     ///
-    /// This is exactly like [`Self::func_wrap`] except it takes an async
-    /// host function.
+    /// This function is similar to [`Self::func_wrap`] except it takes an async
+    /// host function instead of a blocking host function. The `F` function here
+    /// is intended to be:
+    ///
+    /// ```ignore
+    /// F: AsyncFn(StoreContextMut<'_, T>, Params) -> Result<Return>
+    /// ```
+    ///
+    /// however the returned future must be `Send` which is not possible to
+    /// bound at this time. This will be switched to an async closure once Rust
+    /// supports it.
+    ///
+    /// # Blocking / Async Behavior
+    ///
+    /// This function can only be called when [`Config::async_support`] is
+    /// enabled. The function defined which WebAssembly calls will still appear
+    /// as blocking from the perspective of WebAssembly itself. The host,
+    /// however, can perform asynchronous operations without blocking the thread
+    /// performing a call.
+    ///
+    /// When [`Config::async_support`] is enabled then all WebAssembly is
+    /// invoked on a separate stack within a Wasmtime-managed fiber. This means
+    /// that if the future returned by `F` is not immediately ready then the
+    /// fiber will be suspended to block WebAssembly but not the host. When the
+    /// future becomes ready again the fiber will be resumed to continue
+    /// execution within WebAssembly.
+    ///
+    /// # Panics
+    ///
+    /// This function panics if [`Config::async_support`] is set to `false`.
+    ///
+    /// [`Config::async_support`]: crate::Config::async_support
+    /// [`func_wrap_async`]: LinkerInstance::func_wrap_async
     #[cfg(feature = "async")]
     pub fn func_wrap_async<Params, Return, F>(&mut self, name: &str, f: F) -> Result<()>
     where
-        F: for<'a> Fn(
-                StoreContextMut<'a, T>,
+        F: Fn(
+                StoreContextMut<'_, T>,
                 Params,
-            ) -> Box<dyn Future<Output = Result<Return>> + Send + 'a>
+            ) -> Box<dyn Future<Output = Result<Return>> + Send + '_>
             + Send
             + Sync
             + 'static,
@@ -439,47 +485,82 @@ impl<T: 'static> LinkerInstance<'_, T> {
             self.engine.config().async_support,
             "cannot use `func_wrap_async` without enabling async support in the config"
         );
-        let ff = move |mut store: StoreContextMut<'_, T>, params: Params| -> Result<Return> {
-            let async_cx = store.as_context_mut().0.async_cx().expect("async cx");
-            let future = f(store.as_context_mut(), params);
-            unsafe { async_cx.block_on(Pin::from(future)) }?
+        let ff = move |store: StoreContextMut<'_, T>, params: Params| -> Result<Return> {
+            store.block_on(|store| f(store, params).into())?
         };
         self.func_wrap(name, ff)
     }
 
     /// Defines a new host-provided async function into this [`LinkerInstance`].
     ///
-    /// This allows the caller to register host functions with the
-    /// LinkerInstance such that multiple calls to such functions can run
-    /// concurrently. This isn't possible with the existing func_wrap_async
-    /// method because it takes a function which returns a future that owns a
-    /// unique reference to the Store, meaning the Store can't be used for
-    /// anything else until the future resolves.
+    /// This function defines a host function available to call from
+    /// WebAssembly. WebAssembly may additionally make multiple invocations of
+    /// this function concurrently all at the same time. This function requires
+    /// the [`Config::wasm_component_model_async`] feature to be enabled.
     ///
-    /// Ideally, we'd have a way to thread a `StoreContextMut<T>` through an
-    /// arbitrary `Future` such that it has access to the `Store` only while
-    /// being polled (i.e. between, but not across, await points). However,
-    /// there's currently no way to express that in async Rust, so we make do
-    /// with a more awkward scheme: each function registered using
-    /// `func_wrap_concurrent` gets access to the `Store` twice: once before
-    /// doing any concurrent operations (i.e. before awaiting) and once
-    /// afterward. This allows multiple calls to proceed concurrently without
-    /// any one of them monopolizing the store.
+    /// The function `f` provided will be invoked when called by WebAssembly.
+    /// WebAssembly components may then call `f` multiple times while previous
+    /// invocations of `f` are already running. Additionally while `f` is
+    /// running other host functions may be invoked.
+    ///
+    /// The `F` function here is intended to be:
+    ///
+    /// ```ignore
+    /// F: AsyncFn(&Accessor<T>, Params) -> Result<Return>
+    /// ```
+    ///
+    /// however the returned future must be `Send` which is not possible to
+    /// bound at this time. This will be switched to an async closure once Rust
+    /// supports it.
+    ///
+    /// The closure `f` is provided an [`Accessor`] which can be used to acquire
+    /// temporary, blocking, access to a [`StoreContextMut`] (through
+    /// [`Access`]). This models how a store is not available to `f` across
+    /// `await` points but it is temporarily available while actively being
+    /// polled.
+    ///
+    /// # Blocking / Async Behavior
+    ///
+    /// This function can only be called when [`Config::async_support`] is
+    /// enabled. Unlike [`Self::func_wrap`] and [`Self::func_wrap_async`] this
+    /// function is asynchronous even from the perspective of guest WebAssembly.
+    /// This means that if `f` is not immediately resolved then the call from
+    /// WebAssembly will still return immediately (assuming it was lowered with
+    /// `async`). The closure `f` should not block the current thread and should
+    /// only perform blocking via `async` meaning that `f` won't block either
+    /// WebAssembly nor the host.
+    ///
+    /// Note that WebAssembly components can lower host functions both with and
+    /// without `async`. That means that even if a host function is defined in
+    /// the "concurrent" mode here a guest may still lower it synchronously. In
+    /// this situation Wasmtime will manage blocking the guest while the closure
+    /// `f` provided here completes. If a guest lowers this function with
+    /// `async`, though, then no blocking will happen.
+    ///
+    /// # Panics
+    ///
+    /// This function panics if [`Config::async_support`] is set to `false`.
+    ///
+    /// [`Config::async_support`]: crate::Config::async_support
+    /// [`Config::wasm_component_model_async`]: crate::Config::wasm_component_model_async
+    /// [`func_wrap_async`]: LinkerInstance::func_wrap_async
     #[cfg(feature = "component-model-async")]
-    pub fn func_wrap_concurrent<Params, Return, F, N, FN>(&mut self, name: &str, f: F) -> Result<()>
+    pub fn func_wrap_concurrent<Params, Return, F>(&mut self, name: &str, f: F) -> Result<()>
     where
-        N: FnOnce(StoreContextMut<T>) -> Result<Return> + Send + Sync + 'static,
-        FN: Future<Output = N> + Send + Sync + 'static,
-        F: Fn(StoreContextMut<T>, Params) -> FN + Send + Sync + 'static,
+        T: 'static,
+        F: Fn(&Accessor<T>, Params) -> Pin<Box<dyn Future<Output = Result<Return>> + Send + '_>>
+            + Send
+            + Sync
+            + 'static,
         Params: ComponentNamedList + Lift + 'static,
-        Return: ComponentNamedList + Lower + Send + Sync + 'static,
+        Return: ComponentNamedList + Lower + 'static,
     {
         assert!(
             self.engine.config().async_support,
             "cannot use `func_wrap_concurrent` without enabling async support in the config"
         );
-        _ = (name, f);
-        todo!()
+        self.insert(name, Definition::Func(HostFunc::from_concurrent(f)))?;
+        Ok(())
     }
 
     /// Define a new host-provided function using dynamically typed values.
@@ -592,8 +673,10 @@ impl<T: 'static> LinkerInstance<'_, T> {
 
     /// Define a new host-provided async function using dynamic types.
     ///
-    /// This is exactly like [`Self::func_new`] except it takes an async
-    /// host function.
+    /// As [`Self::func_wrap_async`] is a dual of [`Self::func_wrap`], this
+    /// function is the dual of [`Self::func_new`].
+    ///
+    /// For documentation on blocking behavior see [`Self::func_wrap_async`].
     #[cfg(feature = "async")]
     pub fn func_new_async<F>(&mut self, name: &str, f: F) -> Result<()>
     where
@@ -610,12 +693,38 @@ impl<T: 'static> LinkerInstance<'_, T> {
             self.engine.config().async_support,
             "cannot use `func_new_async` without enabling async support in the config"
         );
-        let ff = move |mut store: StoreContextMut<'_, T>, params: &[Val], results: &mut [Val]| {
-            let async_cx = store.as_context_mut().0.async_cx().expect("async cx");
-            let future = f(store.as_context_mut(), params, results);
-            unsafe { async_cx.block_on(Pin::from(future)) }?
+        let ff = move |store: StoreContextMut<'_, T>, params: &[Val], results: &mut [Val]| {
+            store.with_blocking(|store, cx| cx.block_on(Pin::from(f(store, params, results)))?)
         };
-        self.func_new(name, ff)
+        return self.func_new(name, ff);
+    }
+
+    /// Define a new host-provided async function using dynamic types.
+    ///
+    /// As [`Self::func_wrap_concurrent`] is a dual of [`Self::func_wrap`], this
+    /// function is the dual of [`Self::func_new`].
+    ///
+    /// For documentation on async/blocking behavior see
+    /// [`Self::func_wrap_concurrent`].
+    #[cfg(feature = "component-model-async")]
+    pub fn func_new_concurrent<F>(&mut self, name: &str, f: F) -> Result<()>
+    where
+        T: 'static,
+        F: for<'a> Fn(
+                &'a Accessor<T>,
+                &'a [Val],
+                &'a mut [Val],
+            ) -> Pin<Box<dyn Future<Output = Result<()>> + Send + 'a>>
+            + Send
+            + Sync
+            + 'static,
+    {
+        assert!(
+            self.engine.config().async_support,
+            "cannot use `func_wrap_concurrent` without enabling async support in the config"
+        );
+        self.insert(name, Definition::Func(HostFunc::new_dynamic_concurrent(f)))?;
+        Ok(())
     }
 
     /// Defines a [`Module`] within this instance.
@@ -668,10 +777,7 @@ impl<T: 'static> LinkerInstance<'_, T> {
     #[cfg(feature = "async")]
     pub fn resource_async<F>(&mut self, name: &str, ty: ResourceType, dtor: F) -> Result<()>
     where
-        F: for<'a> Fn(
-                StoreContextMut<'a, T>,
-                u32,
-            ) -> Box<dyn Future<Output = Result<()>> + Send + 'a>
+        F: Fn(StoreContextMut<'_, T>, u32) -> Box<dyn Future<Output = Result<()>> + Send + '_>
             + Send
             + Sync
             + 'static,
@@ -683,12 +789,61 @@ impl<T: 'static> LinkerInstance<'_, T> {
         let dtor = Arc::new(crate::func::HostFunc::wrap_inner(
             &self.engine,
             move |mut cx: crate::Caller<'_, T>, (param,): (u32,)| {
-                let async_cx = cx.as_context_mut().0.async_cx().expect("async cx");
-                let future = dtor(cx.as_context_mut(), param);
-                match unsafe { async_cx.block_on(Pin::from(future)) } {
-                    Ok(Ok(())) => Ok(()),
-                    Ok(Err(trap)) | Err(trap) => Err(trap),
-                }
+                cx.as_context_mut()
+                    .block_on(|store| dtor(store, param).into())?
+            },
+        ));
+        self.insert(name, Definition::Resource(ty, dtor))?;
+        Ok(())
+    }
+
+    /// Identical to [`Self::resource`], except that it takes a concurrent destructor.
+    #[cfg(feature = "component-model-async")]
+    pub fn resource_concurrent<F>(&mut self, name: &str, ty: ResourceType, dtor: F) -> Result<()>
+    where
+        T: Send + 'static,
+        F: Fn(&Accessor<T>, u32) -> Pin<Box<dyn Future<Output = Result<()>> + Send + '_>>
+            + Send
+            + Sync
+            + 'static,
+    {
+        assert!(
+            self.engine.config().async_support,
+            "cannot use `resource_concurrent` without enabling async support in the config"
+        );
+        // TODO: This isn't really concurrent -- it requires exclusive access to
+        // the store for the duration of the call, preventing guest code from
+        // running until it completes.  We should make it concurrent and clean
+        // up the implementation to avoid using e.g. `Accessor::new` and
+        // `tls::set` directly.
+        let dtor = Arc::new(dtor);
+        let dtor = Arc::new(crate::func::HostFunc::wrap_inner(
+            &self.engine,
+            move |mut cx: crate::Caller<'_, T>, (param,): (u32,)| {
+                let dtor = dtor.clone();
+                cx.as_context_mut().block_on(move |mut store| {
+                    Box::pin(async move {
+                        // NOTE: We currently pass `None` as the `instance`
+                        // parameter to `Accessor::new` because we don't have ready
+                        // access to it, meaning `dtor` will panic if it tries to
+                        // use `Accessor::instance`.  We could plumb that through
+                        // from the `wasmtime-cranelift`-generated code, but we plan
+                        // to remove `Accessor::instance` once all instances in a
+                        // store share the same concurrent state, at which point we
+                        // won't need it anyway.
+                        let accessor = &Accessor::new(
+                            crate::store::StoreToken::new(store.as_context_mut()),
+                            None,
+                        );
+                        let mut future = std::pin::pin!(dtor(accessor, param));
+                        std::future::poll_fn(|cx| {
+                            crate::component::concurrent::tls::set(store.0.traitobj_mut(), || {
+                                future.as_mut().poll(cx)
+                            })
+                        })
+                        .await
+                    })
+                })?
             },
         ));
         self.insert(name, Definition::Resource(ty, dtor))?;
