@@ -1,8 +1,8 @@
 use crate::linker::{Definition, DefinitionType};
 use crate::prelude::*;
 use crate::runtime::vm::{
-    Imports, ModuleRuntimeInfo, VMFuncRef, VMFunctionImport, VMGlobalImport, VMMemoryImport,
-    VMOpaqueContext, VMTableImport, VMTagImport,
+    self, Imports, ModuleRuntimeInfo, VMFuncRef, VMFunctionImport, VMGlobalImport, VMMemoryImport,
+    VMTableImport, VMTagImport,
 };
 use crate::store::{AllocateInstanceKind, InstanceId, StoreInstanceId, StoreOpaque};
 use crate::types::matching;
@@ -197,7 +197,10 @@ impl Instance {
             !store.0.async_support(),
             "must use async instantiation when async support is enabled",
         );
-        Self::new_started_impl(store, module, imports)
+
+        // SAFETY: the safety contract of `new_started_impl` is the same as this
+        // function.
+        unsafe { Self::new_started_impl(store, module, imports) }
     }
 
     /// Internal function to create an instance and run the start function.
@@ -209,7 +212,9 @@ impl Instance {
         module: &Module,
         imports: Imports<'_>,
     ) -> Result<Instance> {
-        let (instance, start) = Instance::new_raw(store.0, module, imports)?;
+        // SAFETY: the safety contract of `new_raw` is the same as this
+        // function.
+        let (instance, start) = unsafe { Instance::new_raw(store.0, module, imports)? };
         if let Some(start) = start {
             instance.start_raw(store, start)?;
         }
@@ -234,7 +239,11 @@ impl Instance {
         );
 
         store
-            .on_fiber(|store| Self::new_started_impl(store, module, imports))
+            .on_fiber(|store| {
+                // SAFETY: the unsafe contract of `new_started_impl` is the same
+                // as this function.
+                unsafe { Self::new_started_impl(store, module, imports) }
+            })
             .await?
     }
 
@@ -277,11 +286,16 @@ impl Instance {
         // The first thing we do is issue an instance allocation request
         // to the instance allocator. This, on success, will give us an
         // instance handle.
-        let id = store.allocate_instance(
-            AllocateInstanceKind::Module(module_id),
-            &ModuleRuntimeInfo::Module(module.clone()),
-            imports,
-        )?;
+        //
+        // SAFETY: this module, by construction, was already validated within
+        // the store.
+        let id = unsafe {
+            store.allocate_instance(
+                AllocateInstanceKind::Module(module_id),
+                &ModuleRuntimeInfo::Module(module.clone()),
+                imports,
+            )?
+        };
 
         // Additionally, before we start doing fallible instantiation, we
         // do one more step which is to insert an `InstanceData`
@@ -311,10 +325,8 @@ impl Instance {
             .engine()
             .features()
             .contains(WasmFeatures::BULK_MEMORY);
-        store
-            .instance(id)
-            .clone()
-            .initialize(store, compiled_module.module(), bulk_memory)?;
+
+        vm::initialize_instance(store, id, compiled_module.module(), bulk_memory)?;
 
         Ok((instance, compiled_module.module().start_func))
     }
@@ -328,16 +340,16 @@ impl Instance {
     fn start_raw<T>(&self, store: &mut StoreContextMut<'_, T>, start: FuncIndex) -> Result<()> {
         // If a start function is present, invoke it. Make sure we use all the
         // trap-handling configuration in `store` as well.
-        let instance = &mut store.0[self.id];
-        let f = instance.get_exported_func(start);
+        let store_id = store.0.id();
+        let mut instance = self.id.get_mut(store.0);
+        // SAFETY: the `store_id` is the id of the store that owns this
+        // instance and any function stored within the instance.
+        let f = unsafe { instance.as_mut().get_exported_func(store_id, start) };
         let caller_vmctx = instance.vmctx();
         unsafe {
+            let funcref = f.vm_func_ref(store.0);
             super::func::invoke_wasm_and_catch_traps(store, |_default_caller, vm| {
-                f.func_ref.as_ref().array_call(
-                    vm,
-                    VMOpaqueContext::from_vmcontext(caller_vmctx),
-                    NonNull::from(&mut []),
-                )
+                VMFuncRef::array_call(funcref, vm, caller_vmctx, NonNull::from(&mut []))
             })?;
         }
         Ok(())
@@ -368,11 +380,17 @@ impl Instance {
         &'a self,
         store: &'a mut StoreOpaque,
     ) -> impl ExactSizeIterator<Item = Export<'a>> + 'a {
+        let module = store[self.id].env_module().clone();
+        let mut items = Vec::new();
+        for (_name, entity) in module.exports.iter() {
+            items.push(self._get_export(store, *entity));
+        }
         store[self.id]
             .env_module()
             .exports
             .iter()
-            .map(|(name, entity)| Export::new(name, self._get_export(store, *entity)))
+            .zip(items)
+            .map(|((name, _), item)| Export::new(name, item))
     }
 
     /// Looks up an exported [`Extern`] value by name.
@@ -408,6 +426,7 @@ impl Instance {
     /// the value, if found.
     ///
     /// Returns `None` if there was no export with a matching entity index.
+    ///
     /// # Panics
     ///
     /// Panics if `store` does not own this instance.
@@ -426,8 +445,11 @@ impl Instance {
         Some(self._get_export(store, export.entity))
     }
 
-    fn _get_export(&self, store: &StoreOpaque, entity: EntityIndex) -> Extern {
-        let export = store[self.id].get_export_by_index(entity);
+    fn _get_export(&self, store: &mut StoreOpaque, entity: EntityIndex) -> Extern {
+        let id = store.id();
+        // SAFETY: the store `id` owns this instance and all exports contained
+        // within.
+        let export = unsafe { self.id.get_mut(store).get_export_by_index_mut(id, entity) };
         unsafe { Extern::from_wasmtime_export(export, store) }
     }
 
@@ -536,7 +558,10 @@ impl Instance {
         self.get_export(store, name)?.into_tag()
     }
 
-    #[cfg(feature = "component-model")]
+    #[allow(
+        dead_code,
+        reason = "c-api crate does not yet support exnrefs and causes this method to be dead."
+    )]
     pub(crate) fn id(&self) -> InstanceId {
         self.id.instance()
     }
@@ -553,11 +578,8 @@ impl Instance {
         &'a self,
         store: &'a mut StoreOpaque,
     ) -> impl ExactSizeIterator<Item = (GlobalIndex, Global)> + 'a {
-        store[self.id]
-            .all_globals()
-            .collect::<Vec<_>>()
-            .into_iter()
-            .map(|(i, g)| (i, unsafe { Global::from_wasmtime_global(g, store) }))
+        let store_id = store.id();
+        store[self.id].all_globals(store_id)
     }
 
     /// Get all memories within this instance.
@@ -570,13 +592,10 @@ impl Instance {
     #[cfg(feature = "coredump")]
     pub(crate) fn all_memories<'a>(
         &'a self,
-        store: &'a mut StoreOpaque,
+        store: &'a StoreOpaque,
     ) -> impl ExactSizeIterator<Item = (MemoryIndex, Memory)> + 'a {
-        store[self.id]
-            .all_memories()
-            .collect::<Vec<_>>()
-            .into_iter()
-            .map(|(i, m)| (i, unsafe { Memory::from_wasmtime_memory(m, store) }))
+        let store_id = store.id();
+        store[self.id].all_memories(store_id)
     }
 }
 
@@ -649,10 +668,12 @@ impl OwnedImports {
     /// Note that this is unsafe as the validity of `item` is not verified and
     /// it contains a bunch of raw pointers.
     #[cfg(feature = "component-model")]
-    pub(crate) unsafe fn push_export(&mut self, item: &crate::runtime::vm::Export) {
+    pub(crate) fn push_export(&mut self, store: &StoreOpaque, item: &crate::runtime::vm::Export) {
         match item {
             crate::runtime::vm::Export::Function(f) => {
-                let f = f.func_ref.as_ref();
+                // SAFETY: the funcref associated with a `Func` is valid to use
+                // under the `store` that owns the function.
+                let f = unsafe { f.vm_func_ref(store).as_ref() };
                 self.functions.push(VMFunctionImport {
                     wasm_call: f.wasm_call.unwrap(),
                     array_call: f.array_call,
@@ -660,28 +681,16 @@ impl OwnedImports {
                 });
             }
             crate::runtime::vm::Export::Global(g) => {
-                self.globals.push(g.vmimport());
+                self.globals.push(g.vmimport(store));
             }
             crate::runtime::vm::Export::Table(t) => {
-                self.tables.push(VMTableImport {
-                    from: t.definition.into(),
-                    vmctx: t.vmctx.into(),
-                    index: t.index,
-                });
+                self.tables.push(t.vmimport(store));
             }
-            crate::runtime::vm::Export::Memory(m) => {
-                self.memories.push(VMMemoryImport {
-                    from: m.definition.into(),
-                    vmctx: m.vmctx.into(),
-                    index: m.index,
-                });
+            crate::runtime::vm::Export::Memory { memory, .. } => {
+                self.memories.push(memory.vmimport(store));
             }
             crate::runtime::vm::Export::Tag(t) => {
-                self.tags.push(VMTagImport {
-                    from: t.definition.into(),
-                    vmctx: t.vmctx.into(),
-                    index: t.index,
-                });
+                self.tags.push(t.vmimport(store));
             }
         }
     }
@@ -844,8 +853,11 @@ impl<T: 'static> InstancePre<T> {
     #[cfg(feature = "async")]
     pub async fn instantiate_async(
         &self,
-        mut store: impl AsContextMut<Data: Send>,
-    ) -> Result<Instance> {
+        mut store: impl AsContextMut<Data = T>,
+    ) -> Result<Instance>
+    where
+        T: Send,
+    {
         let mut store = store.as_context_mut();
         let imports = pre_instantiate_raw(
             &mut store.0,
