@@ -9,9 +9,9 @@ use crate::{
 use crate::{IntoFunc, prelude::*};
 use alloc::sync::Arc;
 use core::fmt::{self, Debug};
-use core::marker;
 #[cfg(feature = "async")]
-use core::{future::Future, pin::Pin};
+use core::future::Future;
+use core::marker;
 use log::warn;
 
 /// Structure used to link wasm modules/instances together.
@@ -416,6 +416,10 @@ impl<T> Linker<T> {
     ///
     /// Panics if the given function type is not associated with the same engine
     /// as this linker.
+    ///
+    /// # Safety
+    ///
+    /// See [`Func::new_unchecked`] for more safety information.
     pub unsafe fn func_new_unchecked(
         &mut self,
         module: &str,
@@ -427,7 +431,8 @@ impl<T> Linker<T> {
         T: 'static,
     {
         assert!(ty.comes_from_same_engine(self.engine()));
-        let func = HostFunc::new_unchecked(&self.engine, ty, func);
+        // SAFETY: the contract of this function is the same as `new_unchecked`.
+        let func = unsafe { HostFunc::new_unchecked(&self.engine, ty, func) };
         let key = self.import_key(module, Some(name));
         self.insert(key, Definition::HostFunc(Arc::new(func)))?;
         Ok(self)
@@ -470,18 +475,12 @@ impl<T> Linker<T> {
             "cannot use `func_new_async` without enabling async support in the config"
         );
         assert!(ty.comes_from_same_engine(self.engine()));
-        self.func_new(module, name, ty, move |mut caller, params, results| {
-            let async_cx = caller
-                .store
-                .as_context_mut()
-                .0
-                .async_cx()
-                .expect("Attempt to spawn new function on dying fiber");
-            let future = func(caller, params, results);
-            match unsafe { async_cx.block_on(Pin::from(future)) } {
-                Ok(Ok(())) => Ok(()),
-                Ok(Err(trap)) | Err(trap) => Err(trap),
-            }
+        self.func_new(module, name, ty, move |caller, params, results| {
+            let instance = caller.caller();
+            caller.store.with_blocking(|store, cx| {
+                let caller = Caller::new(store, instance);
+                cx.block_on(core::pin::Pin::from(func(caller, params, results)))
+            })?
         })
     }
 
@@ -575,22 +574,18 @@ impl<T> Linker<T> {
             self.engine.config().async_support,
             "cannot use `func_wrap_async` without enabling async support on the config",
         );
-        let func = HostFunc::wrap_inner(
-            &self.engine,
-            move |mut caller: Caller<'_, T>, args: Params| {
-                let async_cx = caller
-                    .store
-                    .as_context_mut()
-                    .0
-                    .async_cx()
-                    .expect("Attempt to start async function on dying fiber");
-                let future = func(caller, args);
-                match unsafe { async_cx.block_on(Pin::from(future)) } {
+        let func =
+            HostFunc::wrap_inner(&self.engine, move |caller: Caller<'_, T>, args: Params| {
+                let instance = caller.caller();
+                let result = caller.store.block_on(|store| {
+                    let caller = Caller::new(store, instance);
+                    func(caller, args).into()
+                });
+                match result {
                     Ok(ret) => ret.into_fallible(),
                     Err(e) => Args::fallible_from_error(e),
                 }
-            },
-        );
+            });
         let key = self.import_key(module, Some(name));
         self.insert(key, Definition::HostFunc(Arc::new(func)))?;
         Ok(self)
@@ -1384,13 +1379,20 @@ impl Definition {
         }
     }
 
+    /// Inserts this definition into the `store` provided.
+    ///
+    /// # Safety
+    ///
     /// Note the unsafety here is due to calling `HostFunc::to_func`. The
     /// requirement here is that the `T` that was originally used to create the
     /// `HostFunc` matches the `T` on the store.
     pub(crate) unsafe fn to_extern(&self, store: &mut StoreOpaque) -> Extern {
         match self {
             Definition::Extern(e, _) => e.clone(),
-            Definition::HostFunc(func) => func.to_func(store).into(),
+            // SAFETY: the contract of this function is the same as what's
+            // required of `to_func`, that `T` of the store matches the `T` of
+            // this original definition.
+            Definition::HostFunc(func) => unsafe { func.to_func(store).into() },
         }
     }
 
