@@ -1,7 +1,7 @@
 use crate::runtime::vm::TableElement;
 use crate::store::{AutoAssertNoGc, StoreOpaque};
 use crate::{
-    AnyRef, ArrayRef, AsContext, AsContextMut, ExternRef, Func, HeapType, RefType, Rooted,
+    AnyRef, ArrayRef, AsContext, AsContextMut, ExnRef, ExternRef, Func, HeapType, RefType, Rooted,
     RootedGcRefImpl, StructRef, V128, ValType, prelude::*,
 };
 use core::ptr;
@@ -47,6 +47,9 @@ pub enum Val {
 
     /// An internal reference.
     AnyRef(Option<Rooted<AnyRef>>),
+
+    /// An exception reference.
+    ExnRef(Option<Rooted<ExnRef>>),
 }
 
 macro_rules! accessors {
@@ -162,6 +165,8 @@ impl Val {
             )),
             Val::AnyRef(None) => ValType::NULLREF,
             Val::AnyRef(Some(a)) => ValType::Ref(RefType::new(false, a._ty(store)?)),
+            Val::ExnRef(None) => ValType::NULLEXNREF,
+            Val::ExnRef(Some(e)) => ValType::Ref(RefType::new(false, e._ty(store)?.into())),
         })
     }
 
@@ -191,6 +196,7 @@ impl Val {
                 Ref::from(*e)._matches_ty(store, ref_ty)?
             }
             (Val::AnyRef(a), ValType::Ref(ref_ty)) => Ref::from(*a)._matches_ty(store, ref_ty)?,
+            (Val::ExnRef(e), ValType::Ref(ref_ty)) => Ref::from(*e)._matches_ty(store, ref_ty)?,
 
             (Val::I32(_), _)
             | (Val::I64(_), _)
@@ -199,7 +205,8 @@ impl Val {
             | (Val::V128(_), _)
             | (Val::FuncRef(_), _)
             | (Val::ExternRef(_), _)
-            | (Val::AnyRef(_), _) => false,
+            | (Val::AnyRef(_), _)
+            | (Val::ExnRef(_), _) => false,
         })
     }
 
@@ -223,11 +230,12 @@ impl Val {
     /// Returns an error if this value is a GC reference and the GC reference
     /// has been unrooted.
     ///
-    /// # Unsafety
+    /// # Safety
     ///
-    /// This method is unsafe for the reasons that [`ExternRef::to_raw`] and
-    /// [`Func::to_raw`] are unsafe.
-    pub unsafe fn to_raw(&self, store: impl AsContextMut) -> Result<ValRaw> {
+    /// The returned [`ValRaw`] does not carry type information and is only safe
+    /// to use within the context of this store itself. For more information see
+    /// [`ExternRef::to_raw`] and [`Func::to_raw`].
+    pub fn to_raw(&self, store: impl AsContextMut) -> Result<ValRaw> {
         match self {
             Val::I32(i) => Ok(ValRaw::i32(*i)),
             Val::I64(i) => Ok(ValRaw::i64(*i)),
@@ -242,6 +250,10 @@ impl Val {
                 None => 0,
                 Some(e) => e.to_raw(store)?,
             })),
+            Val::ExnRef(e) => Ok(ValRaw::exnref(match e {
+                None => 0,
+                Some(e) => e.to_raw(store)?,
+            })),
             Val::FuncRef(f) => Ok(ValRaw::funcref(match f {
                 Some(f) => f.to_raw(store),
                 None => ptr::null_mut(),
@@ -258,9 +270,11 @@ impl Val {
     /// otherwise that `raw` should have the type `ty` specified.
     pub unsafe fn from_raw(mut store: impl AsContextMut, raw: ValRaw, ty: ValType) -> Val {
         let mut store = AutoAssertNoGc::new(store.as_context_mut().0);
-        Self::_from_raw(&mut store, raw, &ty)
+        // SAFETY: `_from_raw` has the same contract as this function.
+        unsafe { Self::_from_raw(&mut store, raw, &ty) }
     }
 
+    /// Same as [`Self::from_raw`], but with a monomorphic store.
     pub(crate) unsafe fn _from_raw(
         store: &mut AutoAssertNoGc<'_>,
         raw: ValRaw,
@@ -274,12 +288,19 @@ impl Val {
             ValType::V128 => Val::V128(raw.get_v128().into()),
             ValType::Ref(ref_ty) => {
                 let ref_ = match ref_ty.heap_type() {
-                    HeapType::Func | HeapType::ConcreteFunc(_) => {
+                    // SAFETY: it's a safety contract of this function that the
+                    // funcref is valid and owned by the provided store.
+                    HeapType::Func | HeapType::ConcreteFunc(_) => unsafe {
                         Func::_from_raw(store, raw.get_funcref()).into()
-                    }
+                    },
 
                     HeapType::NoFunc => Ref::Func(None),
 
+                    HeapType::NoCont | HeapType::ConcreteCont(_) | HeapType::Cont => {
+                        // TODO(#10248): Required to support stack switching in the embedder API.
+                        unimplemented!()
+                    }
+
                     HeapType::Extern => ExternRef::_from_raw(store, raw.get_externref()).into(),
 
                     HeapType::NoExtern => Ref::Extern(None),
@@ -294,6 +315,11 @@ impl Val {
                         AnyRef::_from_raw(store, raw.get_anyref()).into()
                     }
 
+                    HeapType::Exn | HeapType::ConcreteExn(_) => {
+                        ExnRef::_from_raw(store, raw.get_exnref()).into()
+                    }
+                    HeapType::NoExn => Ref::Exn(None),
+
                     HeapType::None => Ref::Any(None),
                 };
                 assert!(
@@ -325,6 +351,7 @@ impl Val {
             Val::FuncRef(f) => Some(Ref::Func(f)),
             Val::ExternRef(e) => Some(Ref::Extern(e)),
             Val::AnyRef(a) => Some(Ref::Any(a)),
+            Val::ExnRef(e) => Some(Ref::Exn(e)),
             Val::I32(_) | Val::I64(_) | Val::F32(_) | Val::F64(_) | Val::V128(_) => None,
         }
     }
@@ -391,6 +418,37 @@ impl Val {
         self.anyref().expect("expected anyref")
     }
 
+    /// Attempt to access the underlying `exnref` value of this `Val`.
+    ///
+    /// If this is not an `exnref`, then `None` is returned.
+    ///
+    /// If this is a null `exnref`, then `Some(None)` is returned.
+    ///
+    /// If this is a non-null `exnref`, then `Some(Some(..))` is returned.
+    #[inline]
+    pub fn exnref(&self) -> Option<Option<&Rooted<ExnRef>>> {
+        match self {
+            Val::ExnRef(None) => Some(None),
+            Val::ExnRef(Some(e)) => Some(Some(e)),
+            _ => None,
+        }
+    }
+
+    /// Returns the underlying `exnref` value of this `Val`, panicking if it's the
+    /// wrong type.
+    ///
+    /// If this is a null `exnref`, then `None` is returned.
+    ///
+    /// If this is a non-null `exnref`, then `Some(..)` is returned.
+    ///
+    /// # Panics
+    ///
+    /// Panics if `self` is not a (nullable) `exnref`.
+    #[inline]
+    pub fn unwrap_exnref(&self) -> Option<&Rooted<ExnRef>> {
+        self.exnref().expect("expected exnref")
+    }
+
     /// Attempt to access the underlying `funcref` value of this `Val`.
     ///
     /// If this is not an `funcref`, then `None` is returned.
@@ -434,6 +492,9 @@ impl Val {
             Val::AnyRef(Some(a)) => a.comes_from_same_store(store),
             Val::AnyRef(None) => true,
 
+            Val::ExnRef(Some(e)) => e.comes_from_same_store(store),
+            Val::ExnRef(None) => true,
+
             // Integers, floats, and vectors have no association with any
             // particular store, so they're always considered as "yes I came
             // from that store",
@@ -477,6 +538,7 @@ impl From<Ref> for Val {
             Ref::Extern(e) => Val::ExternRef(e),
             Ref::Func(f) => Val::FuncRef(f),
             Ref::Any(a) => Val::AnyRef(a),
+            Ref::Exn(e) => Val::ExnRef(e),
         }
     }
 }
@@ -537,6 +599,20 @@ impl From<Option<Rooted<ArrayRef>>> for Val {
     }
 }
 
+impl From<Rooted<ExnRef>> for Val {
+    #[inline]
+    fn from(val: Rooted<ExnRef>) -> Val {
+        Val::ExnRef(Some(val))
+    }
+}
+
+impl From<Option<Rooted<ExnRef>>> for Val {
+    #[inline]
+    fn from(val: Option<Rooted<ExnRef>>) -> Val {
+        Val::ExnRef(val)
+    }
+}
+
 impl From<Func> for Val {
     #[inline]
     fn from(val: Func) -> Val {
@@ -649,6 +725,14 @@ pub enum Ref {
     /// Unlike `externref`, Wasm guests can directly allocate `anyref`s, and
     /// does not need to rely on the host to do that.
     Any(Option<Rooted<AnyRef>>),
+
+    /// An exception-object reference.
+    ///
+    /// The `ExnRef` type represents WebAssembly `exnref`
+    /// values. These are references to exception objects as caught by
+    /// `catch_ref` clauses on `try_table` instructions, or as
+    /// allocated via the host API.
+    Exn(Option<Rooted<ExnRef>>),
 }
 
 impl From<Func> for Ref {
@@ -721,6 +805,20 @@ impl From<Option<Rooted<ArrayRef>>> for Ref {
     }
 }
 
+impl From<Rooted<ExnRef>> for Ref {
+    #[inline]
+    fn from(e: Rooted<ExnRef>) -> Ref {
+        Ref::Exn(Some(e))
+    }
+}
+
+impl From<Option<Rooted<ExnRef>>> for Ref {
+    #[inline]
+    fn from(e: Option<Rooted<ExnRef>>) -> Ref {
+        Ref::Exn(e)
+    }
+}
+
 impl Ref {
     /// Create a null reference to the given heap type.
     #[inline]
@@ -737,8 +835,10 @@ impl Ref {
     #[inline]
     pub fn is_null(&self) -> bool {
         match self {
-            Ref::Any(None) | Ref::Extern(None) | Ref::Func(None) => true,
-            Ref::Any(Some(_)) | Ref::Extern(Some(_)) | Ref::Func(Some(_)) => false,
+            Ref::Any(None) | Ref::Extern(None) | Ref::Func(None) | Ref::Exn(None) => true,
+            Ref::Any(Some(_)) | Ref::Extern(Some(_)) | Ref::Func(Some(_)) | Ref::Exn(Some(_)) => {
+                false
+            }
         }
     }
 
@@ -815,6 +915,39 @@ impl Ref {
         self.as_any().expect("Ref::unwrap_any on non-any reference")
     }
 
+    /// Is this an `exn` reference?
+    #[inline]
+    pub fn is_exn(&self) -> bool {
+        matches!(self, Ref::Exn(_))
+    }
+
+    /// Get the underlying `exn` reference, if any.
+    ///
+    /// Returns `None` if this `Ref` is not an `exn` reference, eg it is a
+    /// `func` reference.
+    ///
+    /// Returns `Some(None)` if this `Ref` is a null `exn` reference.
+    ///
+    /// Returns `Some(Some(_))` if this `Ref` is a non-null `exn` reference.
+    #[inline]
+    pub fn as_exn(&self) -> Option<Option<&Rooted<ExnRef>>> {
+        match self {
+            Ref::Exn(e) => Some(e.as_ref()),
+            _ => None,
+        }
+    }
+
+    /// Get the underlying `exn` reference, panicking if this is a different
+    /// kind of reference.
+    ///
+    /// Returns `None` if this `Ref` is a null `exn` reference.
+    ///
+    /// Returns `Some(_)` if this `Ref` is a non-null `exn` reference.
+    #[inline]
+    pub fn unwrap_exn(&self) -> Option<&Rooted<ExnRef>> {
+        self.as_exn().expect("Ref::unwrap_exn on non-exn reference")
+    }
+
     /// Is this a `func` reference?
     #[inline]
     pub fn is_func(&self) -> bool {
@@ -878,6 +1011,9 @@ impl Ref {
 
                 Ref::Any(None) => HeapType::None,
                 Ref::Any(Some(a)) => a._ty(store)?,
+
+                Ref::Exn(None) => HeapType::None,
+                Ref::Exn(Some(e)) => e._ty(store)?.into(),
             },
         ))
     }
@@ -914,14 +1050,12 @@ impl Ref {
             (Ref::Any(Some(a)), HeapType::Struct) => a._is_struct(store)?,
             (Ref::Any(Some(a)), HeapType::ConcreteStruct(_ty)) => match a._as_struct(store)? {
                 None => false,
-                #[cfg_attr(not(feature = "gc"), allow(unreachable_patterns))]
                 Some(s) => s._matches_ty(store, _ty)?,
             },
             (Ref::Any(Some(a)), HeapType::Eq) => a._is_eqref(store)?,
             (Ref::Any(Some(a)), HeapType::Array) => a._is_array(store)?,
             (Ref::Any(Some(a)), HeapType::ConcreteArray(_ty)) => match a._as_array(store)? {
                 None => false,
-                #[cfg_attr(not(feature = "gc"), allow(unreachable_patterns))]
                 Some(a) => a._matches_ty(store, _ty)?,
             },
             (
@@ -935,6 +1069,13 @@ impl Ref {
                 | HeapType::Eq,
             ) => true,
             (Ref::Any(_), _) => false,
+
+            (Ref::Exn(_), HeapType::Exn) => true,
+            (Ref::Exn(None), HeapType::NoExn | HeapType::ConcreteExn(_)) => true,
+            (Ref::Exn(Some(e)), HeapType::ConcreteExn(_)) => {
+                e._matches_ty(store, &ty.heap_type())?
+            }
+            (Ref::Exn(_), _) => false,
         })
     }
 
@@ -961,6 +1102,8 @@ impl Ref {
             Ref::Extern(None) => true,
             Ref::Any(Some(a)) => a.comes_from_same_store(store),
             Ref::Any(None) => true,
+            Ref::Exn(Some(e)) => e.comes_from_same_store(store),
+            Ref::Exn(None) => true,
         }
     }
 
@@ -991,7 +1134,6 @@ impl Ref {
                     assert!(ty.is_nullable());
                     Ok(TableElement::GcRef(None))
                 }
-                #[cfg_attr(not(feature = "gc"), allow(unreachable_patterns))]
                 Some(e) => {
                     let gc_ref = e.try_clone_gc_ref(&mut store)?;
                     Ok(TableElement::GcRef(Some(gc_ref)))
@@ -1003,13 +1145,23 @@ impl Ref {
                     assert!(ty.is_nullable());
                     Ok(TableElement::GcRef(None))
                 }
-                #[cfg_attr(not(feature = "gc"), allow(unreachable_patterns))]
                 Some(a) => {
                     let gc_ref = a.try_clone_gc_ref(&mut store)?;
                     Ok(TableElement::GcRef(Some(gc_ref)))
                 }
             },
 
+            (Ref::Exn(e), HeapType::Exn) => match e {
+                None => {
+                    assert!(ty.is_nullable());
+                    Ok(TableElement::GcRef(None))
+                }
+                Some(e) => {
+                    let gc_ref = e.try_clone_gc_ref(&mut store)?;
+                    Ok(TableElement::GcRef(Some(gc_ref)))
+                }
+            },
+
             _ => unreachable!("checked that the value matches the type above"),
         }
     }
