@@ -7,13 +7,13 @@
 
 // Polyfill `std::simd::i8x16` etc. until they're stable.
 #[cfg(all(target_arch = "x86_64", target_feature = "sse"))]
-#[allow(non_camel_case_types)]
+#[expect(non_camel_case_types, reason = "matching wasm conventions")]
 pub(crate) type i8x16 = core::arch::x86_64::__m128i;
 #[cfg(all(target_arch = "x86_64", target_feature = "sse"))]
-#[allow(non_camel_case_types)]
+#[expect(non_camel_case_types, reason = "matching wasm conventions")]
 pub(crate) type f32x4 = core::arch::x86_64::__m128;
 #[cfg(all(target_arch = "x86_64", target_feature = "sse"))]
-#[allow(non_camel_case_types)]
+#[expect(non_camel_case_types, reason = "matching wasm conventions")]
 pub(crate) type f64x2 = core::arch::x86_64::__m128d;
 
 // On platforms other than x86_64, define i8x16 to a non-constructible type;
@@ -21,20 +21,23 @@ pub(crate) type f64x2 = core::arch::x86_64::__m128d;
 // functions that are awkward to make conditional on the target, but it
 // doesn't need to actually be constructible unless we're on x86_64.
 #[cfg(not(all(target_arch = "x86_64", target_feature = "sse")))]
-#[allow(non_camel_case_types)]
+#[expect(non_camel_case_types, reason = "matching wasm conventions")]
 #[derive(Copy, Clone)]
 pub(crate) struct i8x16(crate::uninhabited::Uninhabited);
 #[cfg(not(all(target_arch = "x86_64", target_feature = "sse")))]
-#[allow(non_camel_case_types)]
+#[expect(non_camel_case_types, reason = "matching wasm conventions")]
 #[derive(Copy, Clone)]
 pub(crate) struct f32x4(crate::uninhabited::Uninhabited);
 #[cfg(not(all(target_arch = "x86_64", target_feature = "sse")))]
-#[allow(non_camel_case_types)]
+#[expect(non_camel_case_types, reason = "matching wasm conventions")]
 #[derive(Copy, Clone)]
 pub(crate) struct f64x2(crate::uninhabited::Uninhabited);
 
+use crate::StoreContextMut;
 use crate::prelude::*;
+use crate::store::StoreInner;
 use crate::store::StoreOpaque;
+use crate::type_registry::RegisteredType;
 use alloc::sync::Arc;
 use core::fmt;
 use core::ops::Deref;
@@ -48,8 +51,6 @@ use wasmtime_environ::{
 #[cfg(feature = "gc")]
 use wasmtime_environ::ModuleInternedTypeIndex;
 
-#[cfg(has_host_compiler_backend)]
-mod arch;
 #[cfg(feature = "component-model")]
 pub mod component;
 mod const_expr;
@@ -61,11 +62,11 @@ mod memory;
 mod mmap_vec;
 mod provenance;
 mod send_sync_ptr;
+mod stack_switching;
 mod store_box;
 mod sys;
 mod table;
 mod traphandlers;
-mod unwind;
 mod vmcontext;
 
 #[cfg(feature = "threads")]
@@ -89,15 +90,13 @@ pub(crate) use interpreter_disabled as interpreter;
 #[cfg(feature = "debug-builtins")]
 pub use wasmtime_jit_debug::gdb_jit_int::GdbJitImageRegistration;
 
-#[cfg(has_host_compiler_backend)]
-pub use crate::runtime::vm::arch::get_stack_pointer;
 pub use crate::runtime::vm::export::*;
 pub use crate::runtime::vm::gc::*;
 pub use crate::runtime::vm::imports::Imports;
 pub use crate::runtime::vm::instance::{
     GcHeapAllocationIndex, Instance, InstanceAllocationRequest, InstanceAllocator,
     InstanceAllocatorImpl, InstanceAndStore, InstanceHandle, MemoryAllocationIndex,
-    OnDemandInstanceAllocator, StorePtr, TableAllocationIndex,
+    OnDemandInstanceAllocator, StorePtr, TableAllocationIndex, initialize_instance,
 };
 #[cfg(feature = "pooling-allocator")]
 pub use crate::runtime::vm::instance::{
@@ -110,6 +109,7 @@ pub use crate::runtime::vm::memory::{
 };
 pub use crate::runtime::vm::mmap_vec::MmapVec;
 pub use crate::runtime::vm::provenance::*;
+pub use crate::runtime::vm::stack_switching::*;
 pub use crate::runtime::vm::store_box::*;
 #[cfg(feature = "std")]
 pub use crate::runtime::vm::sys::mmap::open_file_for_mmap;
@@ -117,16 +117,17 @@ pub use crate::runtime::vm::sys::mmap::open_file_for_mmap;
 pub use crate::runtime::vm::sys::unwind::UnwindRegistration;
 pub use crate::runtime::vm::table::{Table, TableElement};
 pub use crate::runtime::vm::traphandlers::*;
-pub use crate::runtime::vm::unwind::*;
-#[cfg(feature = "component-model")]
-pub use crate::runtime::vm::vmcontext::VMTableDefinition;
 pub use crate::runtime::vm::vmcontext::{
-    VMArrayCallFunction, VMArrayCallHostFuncContext, VMContext, VMFuncRef, VMFunctionBody,
-    VMFunctionImport, VMGlobalDefinition, VMGlobalImport, VMGlobalKind, VMMemoryDefinition,
-    VMMemoryImport, VMOpaqueContext, VMStoreContext, VMTableImport, VMTagImport,
-    VMWasmCallFunction, ValRaw,
+    VMArrayCallFunction, VMArrayCallHostFuncContext, VMContext, VMFuncRef, VMFunctionImport,
+    VMGlobalDefinition, VMGlobalImport, VMGlobalKind, VMMemoryDefinition, VMMemoryImport,
+    VMOpaqueContext, VMStoreContext, VMTableImport, VMTagImport, VMWasmCallFunction, ValRaw,
 };
+
 pub use send_sync_ptr::SendSyncPtr;
+pub use wasmtime_unwinder::Unwind;
+
+#[cfg(has_host_compiler_backend)]
+pub use wasmtime_unwinder::{UnwindHost, get_stack_pointer};
 
 mod module_id;
 pub use module_id::CompiledModuleId;
@@ -165,14 +166,22 @@ cfg_if::cfg_if! {
 ///
 /// This trait is used to store a raw pointer trait object within each
 /// `VMContext`. This raw pointer trait object points back to the
-/// `wasmtime::Store` internally but is type-erased so this `wasmtime-runtime`
-/// crate doesn't need the entire `wasmtime` crate to build.
+/// `wasmtime::Store` internally but is type-erased to avoid needing to
+/// monomorphize the entire runtime on the `T` in `Store<T>`
 ///
-/// Note that this is an extra-unsafe trait because no heed is paid to the
-/// lifetime of this store or the Send/Sync-ness of this store. All of that must
-/// be respected by embedders (e.g. the `wasmtime::Store` structure). The theory
-/// is that `wasmtime::Store` handles all this correctly.
-pub unsafe trait VMStore {
+/// # Safety
+///
+/// This trait should be implemented by nothing other than `StoreInner<T>` in
+/// this crate. It's not sound to implement it for anything else due to
+/// `unchecked_context_mut` below.
+///
+/// It's also worth nothing that there are various locations where a `*mut dyn
+/// VMStore` is asserted to be both `Send` and `Sync` which disregards the `T`
+/// that's actually stored in the store itself. It's assume that the high-level
+/// APIs using `Store<T>` are correctly inferring send/sync on the returned
+/// values (e.g. futures) and that internally in the runtime we aren't doing
+/// anything "weird" with threads for example.
+pub unsafe trait VMStore: 'static {
     /// Get a shared borrow of this store's `StoreOpaque`.
     fn store_opaque(&self) -> &StoreOpaque;
 
@@ -264,6 +273,19 @@ impl DerefMut for dyn VMStore + '_ {
     }
 }
 
+impl dyn VMStore + '_ {
+    /// Asserts that this `VMStore` was originally paired with `StoreInner<T>`
+    /// and then casts to the `StoreContextMut` type.
+    ///
+    /// # Unsafety
+    ///
+    /// This method is not safe as there's no static guarantee that `T` is
+    /// correct for this store.
+    pub(crate) unsafe fn unchecked_context_mut<T>(&mut self) -> StoreContextMut<'_, T> {
+        unsafe { StoreContextMut(&mut *(self as *mut dyn VMStore as *mut StoreInner<T>)) }
+    }
+}
+
 /// A newtype wrapper around `NonNull<dyn VMStore>` intended to be a
 /// self-pointer back to the `Store<T>` within raw data structures like
 /// `VMContext`.
@@ -313,23 +335,23 @@ pub enum ModuleRuntimeInfo {
 #[derive(Clone)]
 pub struct BareModuleInfo {
     module: Arc<wasmtime_environ::Module>,
-    one_signature: Option<VMSharedTypeIndex>,
     offsets: VMOffsets<HostPtr>,
+    _registered_type: Option<RegisteredType>,
 }
 
 impl ModuleRuntimeInfo {
     pub(crate) fn bare(module: Arc<wasmtime_environ::Module>) -> Self {
-        ModuleRuntimeInfo::bare_maybe_imported_func(module, None)
+        ModuleRuntimeInfo::bare_with_registered_type(module, None)
     }
 
-    pub(crate) fn bare_maybe_imported_func(
+    pub(crate) fn bare_with_registered_type(
         module: Arc<wasmtime_environ::Module>,
-        one_signature: Option<VMSharedTypeIndex>,
+        registered_type: Option<RegisteredType>,
     ) -> Self {
         ModuleRuntimeInfo::Bare(Box::new(BareModuleInfo {
             offsets: VMOffsets::new(HostPtr, &module),
             module,
-            one_signature,
+            _registered_type: registered_type,
         }))
     }
 
@@ -431,10 +453,7 @@ impl ModuleRuntimeInfo {
                 .as_module_map()
                 .values()
                 .as_slice(),
-            ModuleRuntimeInfo::Bare(b) => match &b.one_signature {
-                Some(s) => core::slice::from_ref(s),
-                None => &[],
-            },
+            ModuleRuntimeInfo::Bare(_) => &[],
         }
     }
 
