@@ -215,8 +215,19 @@ pub struct CallContexts {
     scopes: Vec<CallContext>,
 }
 
+impl CallContexts {
+    pub fn push(&mut self, cx: CallContext) {
+        self.scopes.push(cx);
+    }
+
+    pub fn pop(&mut self) -> Option<CallContext> {
+        self.scopes.pop()
+    }
+}
+
+/// State related to borrows for a specific call.
 #[derive(Default)]
-struct CallContext {
+pub struct CallContext {
     lenders: Vec<TypedResourceIndex>,
     borrow_count: u32,
 }
