@@ -70,7 +70,7 @@ pub struct InstanceAllocationRequest<'a> {
     pub store: StorePtr,
 
     /// Indicates '--wmemcheck' flag.
-    #[cfg_attr(not(feature = "wmemcheck"), allow(dead_code))]
+    #[cfg(feature = "wmemcheck")]
     pub wmemcheck: bool,
 
     /// Request that the instance's memories be protected by a specific
@@ -122,7 +122,7 @@ impl StorePtr {
     ///
     /// Safety: must not be used outside the original lifetime of the borrow.
     pub(crate) unsafe fn get(&mut self) -> Option<&mut dyn VMStore> {
-        let ptr = self.0?.as_mut();
+        let ptr = unsafe { self.0?.as_mut() };
         Some(ptr)
     }
 }
@@ -253,13 +253,7 @@ pub unsafe trait InstanceAllocatorImpl {
     fn decrement_core_instance_count(&self);
 
     /// Allocate a memory for an instance.
-    ///
-    /// # Unsafety
-    ///
-    /// The memory and its associated module must have already been validated by
-    /// `Self::validate_memory` (or transtively via
-    /// `Self::validate_{module,component}`) and passed that validation.
-    unsafe fn allocate_memory(
+    fn allocate_memory(
         &self,
         request: &mut InstanceAllocationRequest,
         ty: &wasmtime_environ::Memory,
@@ -282,12 +276,7 @@ pub unsafe trait InstanceAllocatorImpl {
     );
 
     /// Allocate a table for an instance.
-    ///
-    /// # Unsafety
-    ///
-    /// The table and its associated module must have already been validated by
-    /// `Self::validate_module` and passed that validation.
-    unsafe fn allocate_table(
+    fn allocate_table(
         &self,
         req: &mut InstanceAllocationRequest,
         table: &wasmtime_environ::Table,
@@ -409,10 +398,10 @@ pub trait InstanceAllocator: InstanceAllocatorImpl {
     /// Note that the returned instance must still have `.initialize(..)` called
     /// on it to complete the instantiation process.
     ///
-    /// # Unsafety
+    /// # Safety
     ///
-    /// The request's associated module, memories, tables, and vmctx must have
-    /// already have been validated by `Self::validate_module`.
+    /// The `request` provided must be valid, e.g. the imports within are
+    /// correctly sized/typed for the instance being created.
     unsafe fn allocate_module(
         &self,
         mut request: InstanceAllocationRequest,
@@ -436,10 +425,16 @@ pub trait InstanceAllocator: InstanceAllocatorImpl {
             self.allocate_tables(&mut request, &mut tables)?;
             Ok(())
         })() {
-            Ok(_) => Ok(Instance::new(request, memories, tables, &module.memories)),
+            // SAFETY: memories/tables were just allocated from the store within
+            // `request` and this function's own contract requires that the
+            // imports are valid.
+            Ok(_) => unsafe { Ok(Instance::new(request, memories, tables, &module.memories)) },
             Err(e) => {
-                self.deallocate_memories(&mut memories);
-                self.deallocate_tables(&mut tables);
+                // SAFETY: these were previously allocated by this allocator
+                unsafe {
+                    self.deallocate_memories(&mut memories);
+                    self.deallocate_tables(&mut tables);
+                }
                 self.decrement_core_instance_count();
                 Err(e)
             }
@@ -455,25 +450,20 @@ pub trait InstanceAllocator: InstanceAllocatorImpl {
     ///
     /// The instance must have previously been allocated by `Self::allocate`.
     unsafe fn deallocate_module(&self, handle: &mut InstanceHandle) {
-        self.deallocate_memories(&mut handle.instance_mut().memories);
-        self.deallocate_tables(&mut handle.instance_mut().tables);
-
-        let layout = Instance::alloc_layout(handle.instance().offsets());
-        let ptr = handle.instance.take().unwrap();
-        ptr::drop_in_place(ptr.as_ptr());
-        alloc::alloc::dealloc(ptr.as_ptr().cast(), layout);
+        // SAFETY: the contract of `deallocate_*` is itself a contract of this
+        // function, that the memories/tables were previously allocated from
+        // here.
+        unsafe {
+            self.deallocate_memories(handle.get_mut().memories_mut());
+            self.deallocate_tables(handle.get_mut().tables_mut());
+        }
 
         self.decrement_core_instance_count();
     }
 
     /// Allocate the memories for the given instance allocation request, pushing
     /// them into `memories`.
-    ///
-    /// # Unsafety
-    ///
-    /// The request's associated module and memories must have previously been
-    /// validated by `Self::validate_module`.
-    unsafe fn allocate_memories(
+    fn allocate_memories(
         &self,
         request: &mut InstanceAllocationRequest,
         memories: &mut PrimaryMap<DefinedMemoryIndex, (MemoryAllocationIndex, Memory)>,
@@ -489,12 +479,8 @@ pub trait InstanceAllocator: InstanceAllocatorImpl {
                 .defined_memory_index(memory_index)
                 .expect("should be a defined memory since we skipped imported ones");
 
-            memories.push(self.allocate_memory(
-                request,
-                ty,
-                request.tunables,
-                Some(memory_index),
-            )?);
+            let memory = self.allocate_memory(request, ty, request.tunables, Some(memory_index))?;
+            memories.push(memory);
         }
 
         Ok(())
@@ -515,18 +501,19 @@ pub trait InstanceAllocator: InstanceAllocatorImpl {
             // about leaking subsequent memories if the first memory failed to
             // deallocate. If deallocating memory ever becomes fallible, we will
             // need to be careful here!
-            self.deallocate_memory(Some(memory_index), allocation_index, memory);
+            //
+            // SAFETY: the unsafe contract here is the same as the unsafe
+            // contract of this function, that the memories were previously
+            // allocated by this allocator.
+            unsafe {
+                self.deallocate_memory(Some(memory_index), allocation_index, memory);
+            }
         }
     }
 
     /// Allocate tables for the given instance allocation request, pushing them
     /// into `tables`.
-    ///
-    /// # Unsafety
-    ///
-    /// The request's associated module and tables must have previously been
-    /// validated by `Self::validate_module`.
-    unsafe fn allocate_tables(
+    fn allocate_tables(
         &self,
         request: &mut InstanceAllocationRequest,
         tables: &mut PrimaryMap<DefinedTableIndex, (TableAllocationIndex, Table)>,
@@ -542,7 +529,8 @@ pub trait InstanceAllocator: InstanceAllocatorImpl {
                 .defined_table_index(index)
                 .expect("should be a defined table since we skipped imported ones");
 
-            tables.push(self.allocate_table(request, table, request.tunables, def_index)?);
+            let table = self.allocate_table(request, table, request.tunables, def_index)?;
+            tables.push(table);
         }
 
         Ok(())
@@ -559,7 +547,11 @@ pub trait InstanceAllocator: InstanceAllocatorImpl {
         tables: &mut PrimaryMap<DefinedTableIndex, (TableAllocationIndex, Table)>,
     ) {
         for (table_index, (allocation_index, table)) in mem::take(tables) {
-            self.deallocate_table(table_index, allocation_index, table);
+            // SAFETY: the tables here were allocated from this allocator per
+            // the contract on this function itself.
+            unsafe {
+                self.deallocate_table(table_index, allocation_index, table);
+            }
         }
     }
 }
@@ -571,13 +563,12 @@ impl<T: InstanceAllocatorImpl> InstanceAllocator for T {}
 
 fn check_table_init_bounds(
     store: &mut StoreOpaque,
-    instance: &mut Instance,
+    instance: InstanceId,
     module: &Module,
 ) -> Result<()> {
     let mut const_evaluator = ConstExprEvaluator::default();
 
     for segment in module.table_initialization.segments.iter() {
-        let table = unsafe { &*instance.get_table(segment.table_index) };
         let mut context = ConstEvalContext::new(instance);
         let start = unsafe {
             const_evaluator
@@ -587,6 +578,7 @@ fn check_table_init_bounds(
         let start = usize::try_from(start.get_u32()).unwrap();
         let end = start.checked_add(usize::try_from(segment.elements.len()).unwrap());
 
+        let table = store.instance_mut(instance).get_table(segment.table_index);
         match end {
             Some(end) if end <= table.size() => {
                 // Initializer is in bounds
@@ -602,7 +594,7 @@ fn check_table_init_bounds(
 
 fn initialize_tables(
     store: &mut StoreOpaque,
-    context: &mut ConstEvalContext<'_>,
+    context: &mut ConstEvalContext,
     const_evaluator: &mut ConstExprEvaluator,
     module: &Module,
 ) -> Result<()> {
@@ -618,25 +610,47 @@ fn initialize_tables(
                         .expect("const expression should be valid")
                 };
                 let idx = module.table_index(table);
-                let table = unsafe { context.instance.get_defined_table(table).as_mut().unwrap() };
                 match module.tables[idx].ref_type.heap_type.top() {
                     WasmHeapTopType::Extern => {
+                        store.gc_store_mut()?;
+                        let (gc_store, instance) =
+                            store.optional_gc_store_and_instance_mut(context.instance);
+                        let gc_store = gc_store.unwrap();
+                        let table = instance.get_defined_table(table);
                         let gc_ref = VMGcRef::from_raw_u32(raw.get_externref());
-                        let gc_store = store.gc_store_mut()?;
                         let items = (0..table.size())
                             .map(|_| gc_ref.as_ref().map(|r| gc_store.clone_gc_ref(r)));
                         table.init_gc_refs(0, items)?;
                     }
 
                     WasmHeapTopType::Any => {
+                        store.gc_store_mut()?;
+                        let (gc_store, instance) =
+                            store.optional_gc_store_and_instance_mut(context.instance);
+                        let gc_store = gc_store.unwrap();
+                        let table = instance.get_defined_table(table);
                         let gc_ref = VMGcRef::from_raw_u32(raw.get_anyref());
-                        let gc_store = store.gc_store_mut()?;
+                        let items = (0..table.size())
+                            .map(|_| gc_ref.as_ref().map(|r| gc_store.clone_gc_ref(r)));
+                        table.init_gc_refs(0, items)?;
+                    }
+
+                    WasmHeapTopType::Exn => {
+                        store.gc_store_mut()?;
+                        let (gc_store, instance) =
+                            store.optional_gc_store_and_instance_mut(context.instance);
+                        let gc_store = gc_store.unwrap();
+                        let table = instance.get_defined_table(table);
+                        let gc_ref = VMGcRef::from_raw_u32(raw.get_exnref());
                         let items = (0..table.size())
                             .map(|_| gc_ref.as_ref().map(|r| gc_store.clone_gc_ref(r)));
                         table.init_gc_refs(0, items)?;
                     }
 
                     WasmHeapTopType::Func => {
+                        let table = store
+                            .instance_mut(context.instance)
+                            .get_defined_table(table);
                         let funcref = NonNull::new(raw.get_funcref().cast::<VMFuncRef>());
                         let items = (0..table.size()).map(|_| funcref);
                         table.init_func(0, items)?;
@@ -661,8 +675,9 @@ fn initialize_tables(
                 .eval(store, context, &segment.offset)
                 .expect("const expression should be valid")
         };
-        context.instance.table_init_segment(
+        Instance::table_init_segment(
             store,
+            context.instance,
             const_evaluator,
             segment.table_index,
             &segment.elements,
@@ -678,12 +693,12 @@ fn initialize_tables(
 fn get_memory_init_start(
     store: &mut StoreOpaque,
     init: &MemoryInitializer,
-    instance: &mut Instance,
+    instance: InstanceId,
 ) -> Result<u64> {
     let mut context = ConstEvalContext::new(instance);
     let mut const_evaluator = ConstExprEvaluator::default();
     unsafe { const_evaluator.eval(store, &mut context, &init.offset) }.map(|v| {
-        match instance.env_module().memories[init.memory_index].idx_type {
+        match store.instance(instance).env_module().memories[init.memory_index].idx_type {
             wasmtime_environ::IndexType::I32 => v.get_u32().into(),
             wasmtime_environ::IndexType::I64 => v.get_u64(),
         }
@@ -692,11 +707,11 @@ fn get_memory_init_start(
 
 fn check_memory_init_bounds(
     store: &mut StoreOpaque,
-    instance: &mut Instance,
+    instance: InstanceId,
     initializers: &[MemoryInitializer],
 ) -> Result<()> {
     for init in initializers {
-        let memory = instance.get_memory(init.memory_index);
+        let memory = store.instance_mut(instance).get_memory(init.memory_index);
         let start = get_memory_init_start(store, init, instance)?;
         let end = usize::try_from(start)
             .ok()
@@ -717,7 +732,7 @@ fn check_memory_init_bounds(
 
 fn initialize_memories(
     store: &mut StoreOpaque,
-    context: &mut ConstEvalContext<'_>,
+    context: &mut ConstEvalContext,
     const_evaluator: &mut ConstExprEvaluator,
     module: &Module,
 ) -> Result<()> {
@@ -730,19 +745,23 @@ fn initialize_memories(
     // so errors only happen if an out-of-bounds segment is found, in which case
     // a trap is returned.
 
-    struct InitMemoryAtInstantiation<'a, 'b> {
+    struct InitMemoryAtInstantiation<'a> {
         module: &'a Module,
         store: &'a mut StoreOpaque,
-        context: &'a mut ConstEvalContext<'b>,
+        context: &'a mut ConstEvalContext,
         const_evaluator: &'a mut ConstExprEvaluator,
     }
 
-    impl InitMemory for InitMemoryAtInstantiation<'_, '_> {
+    impl InitMemory for InitMemoryAtInstantiation<'_> {
         fn memory_size_in_bytes(
             &mut self,
             memory: wasmtime_environ::MemoryIndex,
         ) -> Result<u64, SizeOverflow> {
-            let len = self.context.instance.get_memory(memory).current_length();
+            let len = self
+                .store
+                .instance(self.context.instance)
+                .get_memory(memory)
+                .current_length();
             let len = u64::try_from(len).unwrap();
             Ok(len)
         }
@@ -755,7 +774,13 @@ fn initialize_memories(
             let val = unsafe { self.const_evaluator.eval(self.store, self.context, expr) }
                 .expect("const expression should be valid");
             Some(
-                match self.context.instance.env_module().memories[memory].idx_type {
+                match self
+                    .store
+                    .instance(self.context.instance)
+                    .env_module()
+                    .memories[memory]
+                    .idx_type
+                {
                     wasmtime_environ::IndexType::I32 => val.get_u32().into(),
                     wasmtime_environ::IndexType::I64 => val.get_u64(),
                 },
@@ -771,15 +796,16 @@ fn initialize_memories(
             // doesn't need initialization, due to something like copy-on-write
             // pre-initializing it via mmap magic, then this initializer can be
             // skipped entirely.
+            let instance = self.store.instance_mut(self.context.instance);
             if let Some(memory_index) = self.module.defined_memory_index(memory_index) {
-                if !self.context.instance.memories[memory_index].1.needs_init() {
+                if !instance.memories[memory_index].1.needs_init() {
                     return true;
                 }
             }
-            let memory = self.context.instance.get_memory(memory_index);
+            let memory = instance.get_memory(memory_index);
 
             unsafe {
-                let src = self.context.instance.wasm_data(init.data.clone());
+                let src = instance.wasm_data(init.data.clone());
                 let offset = usize::try_from(init.offset).unwrap();
                 let dst = memory.base.as_ptr().add(offset);
 
@@ -809,11 +835,7 @@ fn initialize_memories(
     Ok(())
 }
 
-fn check_init_bounds(
-    store: &mut StoreOpaque,
-    instance: &mut Instance,
-    module: &Module,
-) -> Result<()> {
+fn check_init_bounds(store: &mut StoreOpaque, instance: InstanceId, module: &Module) -> Result<()> {
     check_table_init_bounds(store, instance, module)?;
 
     match &module.memory_initialization {
@@ -829,11 +851,14 @@ fn check_init_bounds(
 
 fn initialize_globals(
     store: &mut StoreOpaque,
-    context: &mut ConstEvalContext<'_>,
+    context: &mut ConstEvalContext,
     const_evaluator: &mut ConstExprEvaluator,
     module: &Module,
 ) -> Result<()> {
-    assert!(core::ptr::eq(&**context.instance.env_module(), module));
+    assert!(core::ptr::eq(
+        &**store.instance(context.instance).env_module(),
+        module
+    ));
 
     let mut store = AutoAssertNoGc::new(store);
 
@@ -844,12 +869,13 @@ fn initialize_globals(
                 .expect("should be a valid const expr")
         };
 
-        let to = context.instance.global_ptr(index);
+        let instance = store.instance_mut(context.instance);
+        let to = instance.global_ptr(index);
         let wasm_ty = module.globals[module.global_index(index)].wasm_ty;
 
         #[cfg(feature = "wmemcheck")]
         if index.as_u32() == 0 && wasm_ty == wasmtime_environ::WasmValType::I32 {
-            if let Some(wmemcheck) = &mut context.instance.wmemcheck_state {
+            if let Some(wmemcheck) = instance.wmemcheck_state_mut() {
                 let size = usize::try_from(raw.get_i32()).unwrap();
                 wmemcheck.set_stack_size(size);
             }
@@ -864,9 +890,9 @@ fn initialize_globals(
     Ok(())
 }
 
-pub(super) fn initialize_instance(
+pub fn initialize_instance(
     store: &mut StoreOpaque,
-    instance: &mut Instance,
+    instance: InstanceId,
     module: &Module,
     is_bulk_memory: bool,
 ) -> Result<()> {
