@@ -8,6 +8,7 @@ use core::ptr::NonNull;
 use pulley_interpreter::interp::{DoneReason, RegType, TrapKind, Val, Vm, XRegVal};
 use pulley_interpreter::{FReg, Reg, XReg};
 use wasmtime_environ::{BuiltinFunctionIndex, HostCall, Trap};
+use wasmtime_unwinder::Unwind;
 
 /// Interpreter state stored within a `Store<T>`.
 #[repr(transparent)]
@@ -70,6 +71,11 @@ impl Interpreter {
     pub fn pulley(&self) -> &Vm {
         unsafe { self.pulley.get().as_ref() }
     }
+
+    /// Get an implementation of `Unwind` used to walk the Pulley stack.
+    pub fn unwinder(&self) -> &'static dyn Unwind {
+        &UnwindPulley
+    }
 }
 
 /// Wrapper around `&mut pulley_interpreter::Vm` to enable compiling this to a
@@ -80,6 +86,36 @@ pub struct InterpreterRef<'a> {
     _phantom: marker::PhantomData<&'a mut Vm>,
 }
 
+/// An implementation of stack-walking details specifically designed
+/// for unwinding Pulley's runtime stack.
+pub struct UnwindPulley;
+
+unsafe impl Unwind for UnwindPulley {
+    fn next_older_fp_from_fp_offset(&self) -> usize {
+        0
+    }
+    fn next_older_sp_from_fp_offset(&self) -> usize {
+        if cfg!(target_pointer_width = "32") {
+            8
+        } else {
+            16
+        }
+    }
+    unsafe fn get_next_older_pc_from_fp(&self, fp: usize) -> usize {
+        // The calling convention always pushes the return pointer (aka the PC
+        // of the next older frame) just before this frame.
+        unsafe { *(fp as *mut usize).offset(1) }
+    }
+    fn assert_fp_is_aligned(&self, fp: usize) {
+        let expected = if cfg!(target_pointer_width = "32") {
+            8
+        } else {
+            16
+        };
+        assert_eq!(fp % expected, 0, "stack should always be aligned");
+    }
+}
+
 /// Equivalent of a native platform's `jmp_buf` (sort of).
 ///
 /// This structure ensures that all callee-save state in Pulley is saved at wasm
@@ -123,7 +159,7 @@ impl InterpreterRef<'_> {
         // that pointer originates from a `StoreBox<VM>` in the store itself.
         // One level of safety here relies on that never being deallocated or
         // overwritten, which this file upholds as it's a private field only
-        // this module can acccess.
+        // this module can access.
         //
         // Another aspect upheld by `InterpreterRef` is that it transfers, to
         // the compiler, a mutable borrow of the store (e.g `struct Interpreter`
@@ -152,7 +188,7 @@ impl InterpreterRef<'_> {
         mut self,
         mut bytecode: NonNull<u8>,
         callee: NonNull<VMOpaqueContext>,
-        caller: NonNull<VMOpaqueContext>,
+        caller: NonNull<VMContext>,
         args_and_results: NonNull<[ValRaw]>,
     ) -> bool {
         // Initialize argument registers with the ABI arguments.
@@ -173,16 +209,16 @@ impl InterpreterRef<'_> {
         // correct as it's not saving all callee-save state.
         let setjmp = setjmp(vm);
 
-        let old_lr = vm.call_start(&args);
+        let old_lr = unsafe { vm.call_start(&args) };
 
         // Run the interpreter as much as possible until it finishes, and then
         // handle each finish condition differently.
         let ret = loop {
-            match vm.call_run(bytecode) {
+            match unsafe { vm.call_run(bytecode) } {
                 // If the VM returned entirely then read the return value and
                 // return that (it indicates whether a trap happened or not.
                 DoneReason::ReturnToHost(()) => {
-                    match vm.call_end(old_lr, [RegType::XReg]).next().unwrap() {
+                    match unsafe { vm.call_end(old_lr, [RegType::XReg]).next().unwrap() } {
                         #[allow(
                             clippy::cast_possible_truncation,
                             reason = "intentionally reading the lower bits only"
@@ -202,7 +238,7 @@ impl InterpreterRef<'_> {
                         longjmp(vm, setjmp);
                         break false;
                     } else {
-                        vm = self.call_indirect_host(id);
+                        vm = unsafe { self.call_indirect_host(id) };
                         bytecode = resume;
                     }
                 }
@@ -254,11 +290,9 @@ impl InterpreterRef<'_> {
         /// `call(@host Ty(ty1, ty2, ...) -> retty)` - invoke a host function
         /// with the type `Ty`. The other types in the macro are checked by
         /// rustc to match the actual `Ty` definition in Rust.
-        ///
-        /// Ignore improper ctypes to permit `__m128i` on x86_64.
         macro_rules! call {
             (@builtin($($param:ident),*) $(-> $result:ident)?) => {{
-                #[allow(improper_ctypes_definitions)]
+                #[allow(improper_ctypes_definitions, reason = "__m128i known not FFI-safe")]
                 type T = unsafe extern "C" fn($(call!(@ty $param)),*) $(-> call!(@ty $result))?;
                 call!(@host T($($param),*) $(-> $result)?);
             }};
@@ -266,7 +300,7 @@ impl InterpreterRef<'_> {
 
                 // Decode each argument according to this macro, pulling
                 // arguments from successive registers.
-                let ret = {
+                let ret = unsafe {
                     let mut vm = self.vm();
                     // Convert the pointer from pulley to a native function pointer.
                     union GetNative {
@@ -389,7 +423,7 @@ impl InterpreterRef<'_> {
             use wasmtime_environ::component::ComponentBuiltinFunctionIndex;
 
             if id == const { HostCall::ComponentLowerImport.index() } {
-                call!(@host VMLoweringCallee(nonnull, nonnull, u32, u32, nonnull, ptr, ptr, u8, u8, nonnull, size) -> bool);
+                call!(@host VMLoweringCallee(nonnull, nonnull, u32, u32, nonnull, size) -> bool);
             }
 
             macro_rules! component {
@@ -437,6 +471,7 @@ fn trap(vm: &mut Vm, pc: NonNull<u8>, kind: Option<TrapKind>, setjmp: Setjmp) {
                     TrapKind::BadConversionToInteger => Trap::BadConversionToInteger,
                     TrapKind::MemoryOutOfBounds => Trap::MemoryOutOfBounds,
                     TrapKind::DisabledOpcode => Trap::DisabledOpcode,
+                    TrapKind::StackOverflow => Trap::StackOverflow,
                 };
                 s.set_jit_trap(regs, None, trap);
             }
