@@ -54,19 +54,26 @@
 //! }
 //! ```
 
+#[cfg(feature = "stack-switching")]
+use super::stack_switching::VMContObj;
 use crate::prelude::*;
+use crate::runtime::store::StoreInstanceId;
 #[cfg(feature = "gc")]
 use crate::runtime::vm::VMGcRef;
-use crate::runtime::vm::table::{Table, TableElementType};
+use crate::runtime::vm::table::TableElementType;
 use crate::runtime::vm::vmcontext::VMFuncRef;
 use crate::runtime::vm::{
     HostResultHasUnwindSentinel, Instance, TrapReason, VMStore, f32x4, f64x2, i8x16,
 };
 use core::convert::Infallible;
+use core::pin::Pin;
 use core::ptr::NonNull;
 #[cfg(feature = "threads")]
 use core::time::Duration;
-use wasmtime_environ::{DataIndex, ElemIndex, FuncIndex, MemoryIndex, TableIndex, Trap};
+use wasmtime_environ::{
+    DataIndex, DefinedMemoryIndex, DefinedTableIndex, ElemIndex, FuncIndex, MemoryIndex,
+    TableIndex, Trap,
+};
 #[cfg(feature = "wmemcheck")]
 use wasmtime_wmemcheck::AccessError::{
     DoubleMalloc, InvalidFree, InvalidRead, InvalidWrite, OutOfBounds,
@@ -90,10 +97,6 @@ use wasmtime_wmemcheck::AccessError::{
 /// For more information on converting from host-defined values to Cranelift ABI
 /// values see the `catch_unwind_and_record_trap` function.
 pub mod raw {
-    // Allow these things because of the macro and how we can't differentiate
-    // between doc comments and `cfg`s.
-    #![allow(unused_doc_comments, unused_attributes)]
-
     use crate::runtime::vm::{InstanceAndStore, VMContext, f32x4, f64x2, i8x16};
     use core::ptr::NonNull;
 
@@ -111,17 +114,14 @@ pub mod raw {
                 // This will delegate to the outer module to the actual
                 // implementation and automatically perform `catch_unwind` along
                 // with conversion of the return value in the face of traps.
-                //
-                // Ignore improper ctypes to permit `__m128i` on x86_64.
-                #[allow(unused_variables, missing_docs)]
-                #[allow(improper_ctypes_definitions)]
+                #[allow(improper_ctypes_definitions, reason = "__m128i known not FFI-safe")]
                 pub unsafe extern "C" fn $name(
                     vmctx: NonNull<VMContext>,
                     $( $pname : libcall!(@ty $param), )*
                 ) $(-> libcall!(@ty $result))? {
                     $(#[cfg($attr)])?
                     {
-                        crate::runtime::vm::traphandlers::catch_unwind_and_record_trap(|| {
+                        crate::runtime::vm::traphandlers::catch_unwind_and_record_trap(|| unsafe {
                             InstanceAndStore::from_vmctx(vmctx, |pair| {
                                 let (instance, store) = pair.unpack_mut();
                                 super::$name(store, instance, $($pname),*)
@@ -130,19 +130,19 @@ pub mod raw {
                     }
                     $(
                         #[cfg(not($attr))]
-                        unreachable!();
+                        {
+                            let _ = vmctx;
+                            unreachable!();
+                        }
                     )?
                 }
 
                 // This works around a `rustc` bug where compiling with LTO
                 // will sometimes strip out some of these symbols resulting
                 // in a linking failure.
-                //
-                // Ignore improper ctypes to permit `__m128i` on x86_64.
-                #[allow(non_upper_case_globals)]
+                #[allow(improper_ctypes_definitions, reason = "__m128i known not FFI-safe")]
                 const _: () = {
                     #[used]
-                    #[allow(improper_ctypes_definitions)]
                     static I_AM_USED: unsafe extern "C" fn(
                         NonNull<VMContext>,
                         $( $pname : libcall!(@ty $param), )*
@@ -166,18 +166,20 @@ pub mod raw {
     wasmtime_environ::foreach_builtin_function!(libcall);
 }
 
-fn memory32_grow(
+fn memory_grow(
     store: &mut dyn VMStore,
-    instance: &mut Instance,
+    mut instance: Pin<&mut Instance>,
     delta: u64,
     memory_index: u32,
 ) -> Result<Option<AllocationSize>, TrapReason> {
-    let memory_index = MemoryIndex::from_u32(memory_index);
+    let memory_index = DefinedMemoryIndex::from_u32(memory_index);
+    let module = instance.env_module();
+    let page_size_log2 = module.memories[module.memory_index(memory_index)].page_size_log2;
+
     let result = instance
+        .as_mut()
         .memory_grow(store, memory_index, delta)?
-        .map(|size_in_bytes| {
-            AllocationSize(size_in_bytes / instance.memory_page_size(memory_index))
-        });
+        .map(|size_in_bytes| AllocationSize(size_in_bytes >> page_size_log2));
 
     Ok(result)
 }
@@ -222,20 +224,20 @@ unsafe impl HostResultHasUnwindSentinel for Option<AllocationSize> {
 /// Implementation of `table.grow` for `funcref` tables.
 unsafe fn table_grow_func_ref(
     store: &mut dyn VMStore,
-    instance: &mut Instance,
-    table_index: u32,
+    mut instance: Pin<&mut Instance>,
+    defined_table_index: u32,
     delta: u64,
     init_value: *mut u8,
 ) -> Result<Option<AllocationSize>> {
-    let table_index = TableIndex::from_u32(table_index);
-
-    let element = match instance.table_element_type(table_index) {
-        TableElementType::Func => NonNull::new(init_value.cast::<VMFuncRef>()).into(),
-        TableElementType::GcRef => unreachable!(),
-    };
-
+    let defined_table_index = DefinedTableIndex::from_u32(defined_table_index);
+    let table_index = instance.env_module().table_index(defined_table_index);
+    debug_assert!(matches!(
+        instance.as_mut().table_element_type(table_index),
+        TableElementType::Func,
+    ));
+    let element = NonNull::new(init_value.cast::<VMFuncRef>()).into();
     let result = instance
-        .table_grow(store, table_index, delta, element)?
+        .defined_table_grow(store, defined_table_index, delta, element)?
         .map(AllocationSize);
     Ok(result)
 }
@@ -244,27 +246,54 @@ unsafe fn table_grow_func_ref(
 #[cfg(feature = "gc")]
 unsafe fn table_grow_gc_ref(
     store: &mut dyn VMStore,
-    instance: &mut Instance,
-    table_index: u32,
+    mut instance: Pin<&mut Instance>,
+    defined_table_index: u32,
     delta: u64,
     init_value: u32,
 ) -> Result<Option<AllocationSize>> {
-    let table_index = TableIndex::from_u32(table_index);
+    let defined_table_index = DefinedTableIndex::from_u32(defined_table_index);
+    let table_index = instance.env_module().table_index(defined_table_index);
+    debug_assert!(matches!(
+        instance.as_mut().table_element_type(table_index),
+        TableElementType::GcRef,
+    ));
 
-    let element = match instance.table_element_type(table_index) {
-        TableElementType::Func => unreachable!(),
-        TableElementType::GcRef => VMGcRef::from_raw_u32(init_value)
-            .map(|r| {
-                store
-                    .store_opaque_mut()
-                    .unwrap_gc_store_mut()
-                    .clone_gc_ref(&r)
-            })
-            .into(),
-    };
+    let element = VMGcRef::from_raw_u32(init_value)
+        .map(|r| {
+            store
+                .store_opaque_mut()
+                .unwrap_gc_store_mut()
+                .clone_gc_ref(&r)
+        })
+        .into();
 
     let result = instance
-        .table_grow(store, table_index, delta, element)?
+        .defined_table_grow(store, defined_table_index, delta, element)?
+        .map(AllocationSize);
+    Ok(result)
+}
+
+#[cfg(feature = "stack-switching")]
+unsafe fn table_grow_cont_obj(
+    store: &mut dyn VMStore,
+    mut instance: Pin<&mut Instance>,
+    defined_table_index: u32,
+    delta: u64,
+    // The following two values together form the initial Option<VMContObj>.
+    // A None value is indicated by the pointer being null.
+    init_value_contref: *mut u8,
+    init_value_revision: u64,
+) -> Result<Option<AllocationSize>> {
+    let defined_table_index = DefinedTableIndex::from_u32(defined_table_index);
+    let table_index = instance.env_module().table_index(defined_table_index);
+    debug_assert!(matches!(
+        instance.as_mut().table_element_type(table_index),
+        TableElementType::Cont,
+    ));
+    let element =
+        unsafe { VMContObj::from_raw_parts(init_value_contref, init_value_revision).into() };
+    let result = instance
+        .defined_table_grow(store, defined_table_index, delta, element)?
         .map(AllocationSize);
     Ok(result)
 }
@@ -272,14 +301,14 @@ unsafe fn table_grow_gc_ref(
 /// Implementation of `table.fill` for `funcref`s.
 unsafe fn table_fill_func_ref(
     store: &mut dyn VMStore,
-    instance: &mut Instance,
+    instance: Pin<&mut Instance>,
     table_index: u32,
     dst: u64,
     val: *mut u8,
     len: u64,
 ) -> Result<()> {
-    let table_index = TableIndex::from_u32(table_index);
-    let table = &mut *instance.get_table(table_index);
+    let table_index = DefinedTableIndex::from_u32(table_index);
+    let table = instance.get_defined_table(table_index);
     match table.element_type() {
         TableElementType::Func => {
             let val = NonNull::new(val.cast::<VMFuncRef>());
@@ -287,20 +316,21 @@ unsafe fn table_fill_func_ref(
             Ok(())
         }
         TableElementType::GcRef => unreachable!(),
+        TableElementType::Cont => unreachable!(),
     }
 }
 
 #[cfg(feature = "gc")]
 unsafe fn table_fill_gc_ref(
     store: &mut dyn VMStore,
-    instance: &mut Instance,
+    instance: Pin<&mut Instance>,
     table_index: u32,
     dst: u64,
     val: u32,
     len: u64,
 ) -> Result<()> {
-    let table_index = TableIndex::from_u32(table_index);
-    let table = &mut *instance.get_table(table_index);
+    let table_index = DefinedTableIndex::from_u32(table_index);
+    let table = instance.get_defined_table(table_index);
     match table.element_type() {
         TableElementType::Func => unreachable!(),
         TableElementType::GcRef => {
@@ -310,35 +340,76 @@ unsafe fn table_fill_gc_ref(
             table.fill(Some(gc_store), dst, gc_ref.into(), len)?;
             Ok(())
         }
+
+        TableElementType::Cont => unreachable!(),
+    }
+}
+
+#[cfg(feature = "stack-switching")]
+unsafe fn table_fill_cont_obj(
+    store: &mut dyn VMStore,
+    instance: Pin<&mut Instance>,
+    table_index: u32,
+    dst: u64,
+    value_contref: *mut u8,
+    value_revision: u64,
+    len: u64,
+) -> Result<()> {
+    let table_index = DefinedTableIndex::from_u32(table_index);
+    let table = instance.get_defined_table(table_index);
+    match table.element_type() {
+        TableElementType::Cont => {
+            let contobj = unsafe { VMContObj::from_raw_parts(value_contref, value_revision) };
+            table.fill(store.optional_gc_store_mut(), dst, contobj.into(), len)?;
+            Ok(())
+        }
+        _ => panic!("Wrong table filling function"),
     }
 }
 
 // Implementation of `table.copy`.
-unsafe fn table_copy(
+fn table_copy(
     store: &mut dyn VMStore,
-    instance: &mut Instance,
+    mut instance: Pin<&mut Instance>,
     dst_table_index: u32,
     src_table_index: u32,
     dst: u64,
     src: u64,
     len: u64,
-) -> Result<()> {
+) -> Result<(), Trap> {
     let dst_table_index = TableIndex::from_u32(dst_table_index);
     let src_table_index = TableIndex::from_u32(src_table_index);
     let store = store.store_opaque_mut();
-    let dst_table = instance.get_table(dst_table_index);
-    // Lazy-initialize the whole range in the source table first.
-    let src_range = src..(src.checked_add(len).unwrap_or(u64::MAX));
-    let src_table = instance.get_table_with_lazy_init(src_table_index, src_range);
-    let gc_store = store.optional_gc_store_mut();
-    Table::copy(gc_store, dst_table, src_table, dst, src, len)?;
-    Ok(())
+
+    // Convert the two table indices relative to `instance` into two
+    // defining instances and the defined table index within that instance.
+    let (dst_def_index, dst_instance) = instance
+        .as_mut()
+        .defined_table_index_and_instance(dst_table_index);
+    let dst_instance_id = dst_instance.id();
+    let (src_def_index, src_instance) = instance
+        .as_mut()
+        .defined_table_index_and_instance(src_table_index);
+    let src_instance_id = src_instance.id();
+
+    let src_table = crate::Table::from_raw(
+        StoreInstanceId::new(store.id(), src_instance_id),
+        src_def_index,
+    );
+    let dst_table = crate::Table::from_raw(
+        StoreInstanceId::new(store.id(), dst_instance_id),
+        dst_def_index,
+    );
+
+    // SAFETY: this is only safe if the two tables have the same type, and that
+    // was validated during wasm-validation time.
+    unsafe { crate::Table::copy_raw(store, &dst_table, dst, &src_table, src, len) }
 }
 
 // Implementation of `table.init`.
 fn table_init(
     store: &mut dyn VMStore,
-    instance: &mut Instance,
+    instance: Pin<&mut Instance>,
     table_index: u32,
     elem_index: u32,
     dst: u64,
@@ -358,7 +429,7 @@ fn table_init(
 }
 
 // Implementation of `elem.drop`.
-fn elem_drop(_store: &mut dyn VMStore, instance: &mut Instance, elem_index: u32) {
+fn elem_drop(_store: &mut dyn VMStore, instance: Pin<&mut Instance>, elem_index: u32) {
     let elem_index = ElemIndex::from_u32(elem_index);
     instance.elem_drop(elem_index)
 }
@@ -366,7 +437,7 @@ fn elem_drop(_store: &mut dyn VMStore, instance: &mut Instance, elem_index: u32)
 // Implementation of `memory.copy`.
 fn memory_copy(
     _store: &mut dyn VMStore,
-    instance: &mut Instance,
+    instance: Pin<&mut Instance>,
     dst_index: u32,
     dst: u64,
     src_index: u32,
@@ -381,21 +452,21 @@ fn memory_copy(
 // Implementation of `memory.fill` for locally defined memories.
 fn memory_fill(
     _store: &mut dyn VMStore,
-    instance: &mut Instance,
+    instance: Pin<&mut Instance>,
     memory_index: u32,
     dst: u64,
     val: u32,
     len: u64,
 ) -> Result<(), Trap> {
-    let memory_index = MemoryIndex::from_u32(memory_index);
-    #[allow(clippy::cast_possible_truncation)]
+    let memory_index = DefinedMemoryIndex::from_u32(memory_index);
+    #[expect(clippy::cast_possible_truncation, reason = "known to truncate here")]
     instance.memory_fill(memory_index, dst, val as u8, len)
 }
 
 // Implementation of `memory.init`.
 fn memory_init(
     _store: &mut dyn VMStore,
-    instance: &mut Instance,
+    instance: Pin<&mut Instance>,
     memory_index: u32,
     data_index: u32,
     dst: u64,
@@ -408,7 +479,11 @@ fn memory_init(
 }
 
 // Implementation of `ref.func`.
-fn ref_func(_store: &mut dyn VMStore, instance: &mut Instance, func_index: u32) -> NonNull<u8> {
+fn ref_func(
+    _store: &mut dyn VMStore,
+    instance: Pin<&mut Instance>,
+    func_index: u32,
+) -> NonNull<u8> {
     instance
         .get_func_ref(FuncIndex::from_u32(func_index))
         .expect("ref_func: funcref should always be available for given func index")
@@ -416,7 +491,7 @@ fn ref_func(_store: &mut dyn VMStore, instance: &mut Instance, func_index: u32)
 }
 
 // Implementation of `data.drop`.
-fn data_drop(_store: &mut dyn VMStore, instance: &mut Instance, data_index: u32) {
+fn data_drop(_store: &mut dyn VMStore, instance: Pin<&mut Instance>, data_index: u32) {
     let data_index = DataIndex::from_u32(data_index);
     instance.data_drop(data_index)
 }
@@ -424,7 +499,7 @@ fn data_drop(_store: &mut dyn VMStore, instance: &mut Instance, data_index: u32)
 // Returns a table entry after lazily initializing it.
 unsafe fn table_get_lazy_init_func_ref(
     _store: &mut dyn VMStore,
-    instance: &mut Instance,
+    instance: Pin<&mut Instance>,
     table_index: u32,
     index: u64,
 ) -> *mut u8 {
@@ -434,7 +509,7 @@ unsafe fn table_get_lazy_init_func_ref(
         .get(None, index)
         .expect("table access already bounds-checked");
 
-    match elem.into_func_ref_asserting_initialized() {
+    match unsafe { elem.into_func_ref_asserting_initialized() } {
         Some(ptr) => ptr.as_ptr().cast(),
         None => core::ptr::null_mut(),
     }
@@ -442,7 +517,7 @@ unsafe fn table_get_lazy_init_func_ref(
 
 /// Drop a GC reference.
 #[cfg(feature = "gc-drc")]
-unsafe fn drop_gc_ref(store: &mut dyn VMStore, _instance: &mut Instance, gc_ref: u32) {
+unsafe fn drop_gc_ref(store: &mut dyn VMStore, _instance: Pin<&mut Instance>, gc_ref: u32) {
     log::trace!("libcalls::drop_gc_ref({gc_ref:#x})");
     let gc_ref = VMGcRef::from_raw_u32(gc_ref).expect("non-null VMGcRef");
     store
@@ -455,15 +530,17 @@ unsafe fn drop_gc_ref(store: &mut dyn VMStore, _instance: &mut Instance, gc_ref:
 #[cfg(feature = "gc-null")]
 unsafe fn grow_gc_heap(
     store: &mut dyn VMStore,
-    _instance: &mut Instance,
+    _instance: Pin<&mut Instance>,
     bytes_needed: u64,
 ) -> Result<()> {
     let orig_len = u64::try_from(store.gc_store()?.gc_heap.vmmemory().current_length()).unwrap();
 
-    store
-        .maybe_async_gc(None, Some(bytes_needed))
-        .context("failed to grow the GC heap")
-        .context(crate::Trap::AllocationTooLarge)?;
+    unsafe {
+        store
+            .maybe_async_gc(None, Some(bytes_needed))
+            .context("failed to grow the GC heap")
+            .context(crate::Trap::AllocationTooLarge)?;
+    }
 
     // JIT code relies on the memory having grown by `bytes_needed` bytes if
     // this libcall returns successfully, so trap if we didn't grow that much.
@@ -478,50 +555,13 @@ unsafe fn grow_gc_heap(
     Ok(())
 }
 
-/// Do a GC, keeping `gc_ref` rooted and returning the updated `gc_ref`
-/// reference.
-#[cfg(feature = "gc-drc")]
-unsafe fn gc(store: &mut dyn VMStore, _instance: &mut Instance, gc_ref: u32) -> Result<u32> {
-    let gc_ref = VMGcRef::from_raw_u32(gc_ref);
-    let gc_ref = gc_ref.map(|r| {
-        store
-            .store_opaque_mut()
-            .unwrap_gc_store_mut()
-            .clone_gc_ref(&r)
-    });
-
-    if let Some(gc_ref) = &gc_ref {
-        // It is possible that we are GC'ing because the DRC's activation
-        // table's bump region is full, and we failed to insert `gc_ref` into
-        // the bump region. But it is an invariant for DRC collection that all
-        // GC references on the stack are in the DRC's activations table at the
-        // time of a GC. So make sure to "expose" this GC reference to Wasm (aka
-        // insert it into the DRC's activation table) before we do the actual
-        // GC.
-        let gc_store = store.store_opaque_mut().unwrap_gc_store_mut();
-        let gc_ref = gc_store.clone_gc_ref(gc_ref);
-        let _ = gc_store.expose_gc_ref_to_wasm(gc_ref);
-    }
-
-    match store.maybe_async_grow_or_collect_gc_heap(gc_ref, None)? {
-        None => Ok(0),
-        Some(r) => {
-            let raw = store
-                .store_opaque_mut()
-                .unwrap_gc_store_mut()
-                .expose_gc_ref_to_wasm(r);
-            Ok(raw.get())
-        }
-    }
-}
-
 /// Allocate a raw, unininitialized GC object for Wasm code.
 ///
 /// The Wasm code is responsible for initializing the object.
 #[cfg(feature = "gc-drc")]
 unsafe fn gc_alloc_raw(
     store: &mut dyn VMStore,
-    instance: &mut Instance,
+    instance: Pin<&mut Instance>,
     kind_and_reserved: u32,
     module_interned_type_index: u32,
     size: u32,
@@ -545,7 +585,7 @@ unsafe fn gc_alloc_raw(
         .expect("should have engine type index for module type index");
 
     let mut header = VMGcHeader::from_kind_and_index(kind, shared_type_index);
-    header.set_reserved_u27(kind_and_reserved & VMGcKind::UNUSED_MASK);
+    header.set_reserved_u26(kind_and_reserved & VMGcKind::UNUSED_MASK);
 
     let size = usize::try_from(size).unwrap();
     let align = usize::try_from(align).unwrap();
@@ -575,7 +615,7 @@ unsafe fn gc_alloc_raw(
 #[cfg(feature = "gc")]
 unsafe fn intern_func_ref_for_gc_heap(
     store: &mut dyn VMStore,
-    _instance: &mut Instance,
+    _instance: Pin<&mut Instance>,
     func_ref: *mut u8,
 ) -> Result<u32> {
     use crate::{store::AutoAssertNoGc, vm::SendSyncPtr};
@@ -586,7 +626,7 @@ unsafe fn intern_func_ref_for_gc_heap(
     let func_ref = func_ref.cast::<VMFuncRef>();
     let func_ref = NonNull::new(func_ref).map(SendSyncPtr::new);
 
-    let func_ref_id = store.gc_store_mut()?.func_ref_table.intern(func_ref);
+    let func_ref_id = unsafe { store.gc_store_mut()?.func_ref_table.intern(func_ref) };
     Ok(func_ref_id.into_raw())
 }
 
@@ -597,7 +637,7 @@ unsafe fn intern_func_ref_for_gc_heap(
 #[cfg(feature = "gc")]
 unsafe fn get_interned_func_ref(
     store: &mut dyn VMStore,
-    instance: &mut Instance,
+    instance: Pin<&mut Instance>,
     func_ref_id: u32,
     module_interned_type_index: u32,
 ) -> *mut u8 {
@@ -631,7 +671,7 @@ unsafe fn get_interned_func_ref(
 #[cfg(feature = "gc")]
 unsafe fn array_new_data(
     store: &mut dyn VMStore,
-    instance: &mut Instance,
+    instance: Pin<&mut Instance>,
     array_type_index: u32,
     data_index: u32,
     src: u32,
@@ -673,12 +713,14 @@ unsafe fn array_new_data(
         .layout(shared_ty)
         .expect("array types have GC layouts");
     let array_layout = gc_layout.unwrap_array();
-    let array_ref = store.retry_after_gc_maybe_async((), |store, ()| {
-        store
-            .unwrap_gc_store_mut()
-            .alloc_uninit_array(shared_ty, len, &array_layout)?
-            .map_err(|bytes_needed| crate::GcHeapOutOfMemory::new((), bytes_needed).into())
-    })?;
+    let array_ref = unsafe {
+        store.retry_after_gc_maybe_async((), |store, ()| {
+            store
+                .unwrap_gc_store_mut()
+                .alloc_uninit_array(shared_ty, len, &array_layout)?
+                .map_err(|bytes_needed| crate::GcHeapOutOfMemory::new((), bytes_needed).into())
+        })?
+    };
 
     // Copy the data into the array, initializing it.
     store
@@ -697,7 +739,7 @@ unsafe fn array_new_data(
 #[cfg(feature = "gc")]
 unsafe fn array_init_data(
     store: &mut dyn VMStore,
-    instance: &mut Instance,
+    instance: Pin<&mut Instance>,
     array_type_index: u32,
     array: u32,
     dst: u32,
@@ -778,7 +820,7 @@ unsafe fn array_init_data(
 #[cfg(feature = "gc")]
 unsafe fn array_new_elem(
     store: &mut dyn VMStore,
-    instance: &mut Instance,
+    mut instance: Pin<&mut Instance>,
     array_type_index: u32,
     elem_index: u32,
     src: u32,
@@ -817,8 +859,10 @@ unsafe fn array_new_elem(
                         .ok_or_else(|| Trap::TableOutOfBounds)?
                         .iter()
                         .map(|f| {
-                            let raw_func_ref = instance.get_func_ref(*f);
-                            let func = raw_func_ref.map(|p| Func::from_vm_func_ref(store, p));
+                            let raw_func_ref = instance.as_mut().get_func_ref(*f);
+                            let func = unsafe {
+                                raw_func_ref.map(|p| Func::from_vm_func_ref(store.id(), p))
+                            };
                             Val::FuncRef(func)
                         }),
                 );
@@ -829,7 +873,7 @@ unsafe fn array_new_elem(
                     .and_then(|s| s.get(..len))
                     .ok_or_else(|| Trap::TableOutOfBounds)?;
 
-                let mut const_context = ConstEvalContext::new(instance);
+                let mut const_context = ConstEvalContext::new(instance.id());
                 let mut const_evaluator = ConstExprEvaluator::default();
 
                 vals.extend(xs.iter().map(|x| unsafe {
@@ -854,7 +898,7 @@ unsafe fn array_new_elem(
 #[cfg(feature = "gc")]
 unsafe fn array_init_elem(
     store: &mut dyn VMStore,
-    instance: &mut Instance,
+    mut instance: Pin<&mut Instance>,
     array_type_index: u32,
     array: u32,
     dst: u32,
@@ -910,8 +954,8 @@ unsafe fn array_init_elem(
             .ok_or_else(|| Trap::TableOutOfBounds)?
             .iter()
             .map(|f| {
-                let raw_func_ref = instance.get_func_ref(*f);
-                let func = raw_func_ref.map(|p| Func::from_vm_func_ref(&store, p));
+                let raw_func_ref = instance.as_mut().get_func_ref(*f);
+                let func = unsafe { raw_func_ref.map(|p| Func::from_vm_func_ref(store.id(), p)) };
                 Val::FuncRef(func)
             })
             .collect::<Vec<_>>(),
@@ -919,7 +963,7 @@ unsafe fn array_init_elem(
             let elem_ty = array._ty(&store)?.element_type();
             let elem_ty = elem_ty.unwrap_val_type();
 
-            let mut const_context = ConstEvalContext::new(instance);
+            let mut const_context = ConstEvalContext::new(instance.id());
             let mut const_evaluator = ConstExprEvaluator::default();
 
             xs.get(src..)
@@ -954,7 +998,7 @@ unsafe fn array_init_elem(
 #[cfg(feature = "gc")]
 unsafe fn array_copy(
     store: &mut dyn VMStore,
-    _instance: &mut Instance,
+    _instance: Pin<&mut Instance>,
     dst_array: u32,
     dst: u32,
     src_array: u32,
@@ -1015,7 +1059,7 @@ unsafe fn array_copy(
 #[cfg(feature = "gc")]
 unsafe fn is_subtype(
     store: &mut dyn VMStore,
-    _instance: &mut Instance,
+    _instance: Pin<&mut Instance>,
     actual_engine_type: u32,
     expected_engine_type: u32,
 ) -> u32 {
@@ -1034,14 +1078,14 @@ unsafe fn is_subtype(
 #[cfg(feature = "threads")]
 fn memory_atomic_notify(
     _store: &mut dyn VMStore,
-    instance: &mut Instance,
+    instance: Pin<&mut Instance>,
     memory_index: u32,
     addr_index: u64,
     count: u32,
 ) -> Result<u32, Trap> {
-    let memory = MemoryIndex::from_u32(memory_index);
+    let memory = DefinedMemoryIndex::from_u32(memory_index);
     instance
-        .get_runtime_memory(memory)
+        .get_defined_memory_mut(memory)
         .atomic_notify(addr_index, count)
 }
 
@@ -1049,16 +1093,16 @@ fn memory_atomic_notify(
 #[cfg(feature = "threads")]
 fn memory_atomic_wait32(
     _store: &mut dyn VMStore,
-    instance: &mut Instance,
+    instance: Pin<&mut Instance>,
     memory_index: u32,
     addr_index: u64,
     expected: u32,
     timeout: u64,
 ) -> Result<u32, Trap> {
     let timeout = (timeout as i64 >= 0).then(|| Duration::from_nanos(timeout));
-    let memory = MemoryIndex::from_u32(memory_index);
+    let memory = DefinedMemoryIndex::from_u32(memory_index);
     Ok(instance
-        .get_runtime_memory(memory)
+        .get_defined_memory_mut(memory)
         .atomic_wait32(addr_index, expected, timeout)? as u32)
 }
 
@@ -1066,27 +1110,27 @@ fn memory_atomic_wait32(
 #[cfg(feature = "threads")]
 fn memory_atomic_wait64(
     _store: &mut dyn VMStore,
-    instance: &mut Instance,
+    instance: Pin<&mut Instance>,
     memory_index: u32,
     addr_index: u64,
     expected: u64,
     timeout: u64,
 ) -> Result<u32, Trap> {
     let timeout = (timeout as i64 >= 0).then(|| Duration::from_nanos(timeout));
-    let memory = MemoryIndex::from_u32(memory_index);
+    let memory = DefinedMemoryIndex::from_u32(memory_index);
     Ok(instance
-        .get_runtime_memory(memory)
+        .get_defined_memory_mut(memory)
         .atomic_wait64(addr_index, expected, timeout)? as u32)
 }
 
 // Hook for when an instance runs out of fuel.
-fn out_of_gas(store: &mut dyn VMStore, _instance: &mut Instance) -> Result<()> {
+fn out_of_gas(store: &mut dyn VMStore, _instance: Pin<&mut Instance>) -> Result<()> {
     store.out_of_gas()
 }
 
 // Hook for when an instance observes that the epoch has changed.
 #[cfg(target_has_atomic = "64")]
-fn new_epoch(store: &mut dyn VMStore, _instance: &mut Instance) -> Result<NextEpoch> {
+fn new_epoch(store: &mut dyn VMStore, _instance: Pin<&mut Instance>) -> Result<NextEpoch> {
     store.new_epoch().map(NextEpoch)
 }
 
@@ -1104,11 +1148,11 @@ unsafe impl HostResultHasUnwindSentinel for NextEpoch {
 #[cfg(feature = "wmemcheck")]
 unsafe fn check_malloc(
     _store: &mut dyn VMStore,
-    instance: &mut Instance,
+    instance: Pin<&mut Instance>,
     addr: u32,
     len: u32,
 ) -> Result<()> {
-    if let Some(wmemcheck_state) = &mut instance.wmemcheck_state {
+    if let Some(wmemcheck_state) = instance.wmemcheck_state_mut() {
         let result = wmemcheck_state.malloc(addr as usize, len as usize);
         wmemcheck_state.memcheck_on();
         match result {
@@ -1129,8 +1173,12 @@ unsafe fn check_malloc(
 
 // Hook for validating free using wmemcheck_state.
 #[cfg(feature = "wmemcheck")]
-unsafe fn check_free(_store: &mut dyn VMStore, instance: &mut Instance, addr: u32) -> Result<()> {
-    if let Some(wmemcheck_state) = &mut instance.wmemcheck_state {
+unsafe fn check_free(
+    _store: &mut dyn VMStore,
+    instance: Pin<&mut Instance>,
+    addr: u32,
+) -> Result<()> {
+    if let Some(wmemcheck_state) = instance.wmemcheck_state_mut() {
         let result = wmemcheck_state.free(addr as usize);
         wmemcheck_state.memcheck_on();
         match result {
@@ -1150,12 +1198,12 @@ unsafe fn check_free(_store: &mut dyn VMStore, instance: &mut Instance, addr: u3
 #[cfg(feature = "wmemcheck")]
 fn check_load(
     _store: &mut dyn VMStore,
-    instance: &mut Instance,
+    instance: Pin<&mut Instance>,
     num_bytes: u32,
     addr: u32,
     offset: u32,
 ) -> Result<()> {
-    if let Some(wmemcheck_state) = &mut instance.wmemcheck_state {
+    if let Some(wmemcheck_state) = instance.wmemcheck_state_mut() {
         let result = wmemcheck_state.read(addr as usize + offset as usize, num_bytes as usize);
         match result {
             Ok(()) => {}
@@ -1177,12 +1225,12 @@ fn check_load(
 #[cfg(feature = "wmemcheck")]
 fn check_store(
     _store: &mut dyn VMStore,
-    instance: &mut Instance,
+    instance: Pin<&mut Instance>,
     num_bytes: u32,
     addr: u32,
     offset: u32,
 ) -> Result<()> {
-    if let Some(wmemcheck_state) = &mut instance.wmemcheck_state {
+    if let Some(wmemcheck_state) = instance.wmemcheck_state_mut() {
         let result = wmemcheck_state.write(addr as usize + offset as usize, num_bytes as usize);
         match result {
             Ok(()) => {}
@@ -1202,23 +1250,23 @@ fn check_store(
 
 // Hook for turning wmemcheck load/store validation off when entering a malloc function.
 #[cfg(feature = "wmemcheck")]
-fn malloc_start(_store: &mut dyn VMStore, instance: &mut Instance) {
-    if let Some(wmemcheck_state) = &mut instance.wmemcheck_state {
+fn malloc_start(_store: &mut dyn VMStore, instance: Pin<&mut Instance>) {
+    if let Some(wmemcheck_state) = instance.wmemcheck_state_mut() {
         wmemcheck_state.memcheck_off();
     }
 }
 
 // Hook for turning wmemcheck load/store validation off when entering a free function.
 #[cfg(feature = "wmemcheck")]
-fn free_start(_store: &mut dyn VMStore, instance: &mut Instance) {
-    if let Some(wmemcheck_state) = &mut instance.wmemcheck_state {
+fn free_start(_store: &mut dyn VMStore, instance: Pin<&mut Instance>) {
+    if let Some(wmemcheck_state) = instance.wmemcheck_state_mut() {
         wmemcheck_state.memcheck_off();
     }
 }
 
 // Hook for tracking wasm stack updates using wmemcheck_state.
 #[cfg(feature = "wmemcheck")]
-fn update_stack_pointer(_store: &mut dyn VMStore, _instance: &mut Instance, _value: u32) {
+fn update_stack_pointer(_store: &mut dyn VMStore, _instance: Pin<&mut Instance>, _value: u32) {
     // TODO: stack-tracing has yet to be finalized. All memory below
     // the address of the top of the stack is marked as valid for
     // loads and stores.
@@ -1229,50 +1277,55 @@ fn update_stack_pointer(_store: &mut dyn VMStore, _instance: &mut Instance, _val
 
 // Hook updating wmemcheck_state memory state vector every time memory.grow is called.
 #[cfg(feature = "wmemcheck")]
-fn update_mem_size(_store: &mut dyn VMStore, instance: &mut Instance, num_pages: u32) {
-    if let Some(wmemcheck_state) = &mut instance.wmemcheck_state {
+fn update_mem_size(_store: &mut dyn VMStore, instance: Pin<&mut Instance>, num_pages: u32) {
+    if let Some(wmemcheck_state) = instance.wmemcheck_state_mut() {
         const KIB: usize = 1024;
         let num_bytes = num_pages as usize * 64 * KIB;
         wmemcheck_state.update_mem_size(num_bytes);
     }
 }
 
-fn floor_f32(_store: &mut dyn VMStore, _instance: &mut Instance, val: f32) -> f32 {
+fn floor_f32(_store: &mut dyn VMStore, _instance: Pin<&mut Instance>, val: f32) -> f32 {
     wasmtime_math::WasmFloat::wasm_floor(val)
 }
 
-fn floor_f64(_store: &mut dyn VMStore, _instance: &mut Instance, val: f64) -> f64 {
+fn floor_f64(_store: &mut dyn VMStore, _instance: Pin<&mut Instance>, val: f64) -> f64 {
     wasmtime_math::WasmFloat::wasm_floor(val)
 }
 
-fn ceil_f32(_store: &mut dyn VMStore, _instance: &mut Instance, val: f32) -> f32 {
+fn ceil_f32(_store: &mut dyn VMStore, _instance: Pin<&mut Instance>, val: f32) -> f32 {
     wasmtime_math::WasmFloat::wasm_ceil(val)
 }
 
-fn ceil_f64(_store: &mut dyn VMStore, _instance: &mut Instance, val: f64) -> f64 {
+fn ceil_f64(_store: &mut dyn VMStore, _instance: Pin<&mut Instance>, val: f64) -> f64 {
     wasmtime_math::WasmFloat::wasm_ceil(val)
 }
 
-fn trunc_f32(_store: &mut dyn VMStore, _instance: &mut Instance, val: f32) -> f32 {
+fn trunc_f32(_store: &mut dyn VMStore, _instance: Pin<&mut Instance>, val: f32) -> f32 {
     wasmtime_math::WasmFloat::wasm_trunc(val)
 }
 
-fn trunc_f64(_store: &mut dyn VMStore, _instance: &mut Instance, val: f64) -> f64 {
+fn trunc_f64(_store: &mut dyn VMStore, _instance: Pin<&mut Instance>, val: f64) -> f64 {
     wasmtime_math::WasmFloat::wasm_trunc(val)
 }
 
-fn nearest_f32(_store: &mut dyn VMStore, _instance: &mut Instance, val: f32) -> f32 {
+fn nearest_f32(_store: &mut dyn VMStore, _instance: Pin<&mut Instance>, val: f32) -> f32 {
     wasmtime_math::WasmFloat::wasm_nearest(val)
 }
 
-fn nearest_f64(_store: &mut dyn VMStore, _instance: &mut Instance, val: f64) -> f64 {
+fn nearest_f64(_store: &mut dyn VMStore, _instance: Pin<&mut Instance>, val: f64) -> f64 {
     wasmtime_math::WasmFloat::wasm_nearest(val)
 }
 
 // This intrinsic is only used on x86_64 platforms as an implementation of
 // the `i8x16.swizzle` instruction when `pshufb` in SSSE3 is not available.
 #[cfg(all(target_arch = "x86_64", target_feature = "sse"))]
-fn i8x16_swizzle(_store: &mut dyn VMStore, _instance: &mut Instance, a: i8x16, b: i8x16) -> i8x16 {
+fn i8x16_swizzle(
+    _store: &mut dyn VMStore,
+    _instance: Pin<&mut Instance>,
+    a: i8x16,
+    b: i8x16,
+) -> i8x16 {
     union U {
         reg: i8x16,
         mem: [u8; 16],
@@ -1316,7 +1369,7 @@ fn i8x16_swizzle(_store: &mut dyn VMStore, _instance: &mut Instance, a: i8x16, b
 #[cfg(not(all(target_arch = "x86_64", target_feature = "sse")))]
 fn i8x16_swizzle(
     _store: &mut dyn VMStore,
-    _instance: &mut Instance,
+    _instance: Pin<&mut Instance>,
     _a: i8x16,
     _b: i8x16,
 ) -> i8x16 {
@@ -1328,7 +1381,7 @@ fn i8x16_swizzle(
 #[cfg(all(target_arch = "x86_64", target_feature = "sse"))]
 fn i8x16_shuffle(
     _store: &mut dyn VMStore,
-    _instance: &mut Instance,
+    _instance: Pin<&mut Instance>,
     a: i8x16,
     b: i8x16,
     c: i8x16,
@@ -1382,7 +1435,7 @@ fn i8x16_shuffle(
 #[cfg(not(all(target_arch = "x86_64", target_feature = "sse")))]
 fn i8x16_shuffle(
     _store: &mut dyn VMStore,
-    _instance: &mut Instance,
+    _instance: Pin<&mut Instance>,
     _a: i8x16,
     _b: i8x16,
     _c: i8x16,
@@ -1392,7 +1445,7 @@ fn i8x16_shuffle(
 
 fn fma_f32x4(
     _store: &mut dyn VMStore,
-    _instance: &mut Instance,
+    _instance: Pin<&mut Instance>,
     x: f32x4,
     y: f32x4,
     z: f32x4,
@@ -1421,7 +1474,7 @@ fn fma_f32x4(
 
 fn fma_f64x2(
     _store: &mut dyn VMStore,
-    _instance: &mut Instance,
+    _instance: Pin<&mut Instance>,
     x: f64x2,
     y: f64x2,
     z: f64x2,
@@ -1453,7 +1506,7 @@ fn fma_f64x2(
 /// `Result` values to record such trap information.
 fn trap(
     _store: &mut dyn VMStore,
-    _instance: &mut Instance,
+    _instance: Pin<&mut Instance>,
     code: u8,
 ) -> Result<Infallible, TrapReason> {
     Err(TrapReason::Wasm(
@@ -1461,7 +1514,7 @@ fn trap(
     ))
 }
 
-fn raise(_store: &mut dyn VMStore, _instance: &mut Instance) {
+fn raise(_store: &mut dyn VMStore, _instance: Pin<&mut Instance>) {
     // SAFETY: this is only called from compiled wasm so we know that wasm has
     // already been entered. It's a dynamic safety precondition that the trap
     // information has already been arranged to be present.
@@ -1475,3 +1528,18 @@ fn raise(_store: &mut dyn VMStore, _instance: &mut Instance) {
     #[cfg(not(has_host_compiler_backend))]
     unreachable!()
 }
+
+// Builtins for continuations. These are thin wrappers around the
+// respective definitions in stack_switching.rs.
+#[cfg(feature = "stack-switching")]
+fn cont_new(
+    store: &mut dyn VMStore,
+    instance: Pin<&mut Instance>,
+    func: *mut u8,
+    param_count: u32,
+    result_count: u32,
+) -> Result<Option<AllocationSize>, TrapReason> {
+    let ans =
+        crate::vm::stack_switching::cont_new(store, instance, func, param_count, result_count)?;
+    Ok(Some(AllocationSize(ans.cast::<u8>() as usize)))
+}
