@@ -169,6 +169,9 @@ pub trait RuntimeLinearMemory: Send + Sync {
     /// offset within it.
     fn base(&self) -> MemoryBase;
 
+    /// Get a `VMMemoryDefinition` for this linear memory.
+    fn vmmemory(&self) -> VMMemoryDefinition;
+
     /// Internal method for Wasmtime when used in conjunction with CoW images.
     /// This is used to inform the underlying memory that the size of memory has
     /// changed.
@@ -429,9 +432,14 @@ impl Memory {
         }
     }
 
+    /// Is this a shared memory?
+    pub fn is_shared_memory(&self) -> bool {
+        matches!(self, Memory::Shared(_))
+    }
+
     /// If the [Memory] is a [SharedMemory], unwrap it and return a clone to
     /// that shared memory.
-    pub fn as_shared_memory(&mut self) -> Option<&mut SharedMemory> {
+    pub fn as_shared_memory(&self) -> Option<&SharedMemory> {
         match self {
             Memory::Local(_) => None,
             Memory::Shared(mem) => Some(mem),
@@ -539,29 +547,6 @@ impl LocalMemory {
 
                     let mut slot =
                         MemoryImageSlot::create(mmap_base, byte_size, alloc.byte_capacity());
-                    // On drop, we will unmap our mmap'd range that this slot
-                    // was mapped on top of, so there is no need for the slot to
-                    // wipe it with an anonymous mapping first.
-                    //
-                    // Note that this code would be incorrect if clear-on-drop
-                    // were enabled. That's because:
-                    //
-                    // * In the struct definition, `memory_image` above is listed
-                    //   after `alloc`.
-                    // * Rust drops fields in the order they're defined, so
-                    //   `memory_image` would be dropped after `alloc`.
-                    // * `alloc` can represent either owned memory (i.e. the mmap is
-                    //   freed on drop) or logically borrowed memory (something else
-                    //   manages the mmap).
-                    // * If `alloc` is borrowed memory, then this isn't an issue.
-                    // * But if `alloc` is owned memory, then it would first drop
-                    //   the mmap, and then `memory_image` would try to remap
-                    //   part of that same memory as part of clear-on-drop.
-                    //
-                    // A lot of this really suggests representing the ownership
-                    // via Rust lifetimes -- that would be a major refactor,
-                    // though.
-                    slot.no_clear_on_drop();
                     slot.instantiate(alloc.byte_size(), Some(image), ty, tunables)?;
                     Some(slot)
                 } else {
@@ -702,10 +687,7 @@ impl LocalMemory {
     }
 
     pub fn vmmemory(&self) -> VMMemoryDefinition {
-        VMMemoryDefinition {
-            base: self.alloc.base().as_non_null().into(),
-            current_length: self.alloc.byte_size().into(),
-        }
+        self.alloc.vmmemory()
     }
 
     pub fn byte_size(&self) -> usize {
