@@ -23,10 +23,12 @@ use crate::{EntryStoreContext, prelude::*};
 use crate::{StoreContextMut, WasmBacktrace};
 use core::cell::Cell;
 use core::num::NonZeroU32;
-use core::ops::Range;
 use core::ptr::{self, NonNull};
 
 pub use self::backtrace::Backtrace;
+#[cfg(feature = "gc")]
+pub use wasmtime_unwinder::Frame;
+
 pub use self::coredump::CoreDumpStack;
 pub use self::tls::tls_eager_initialize;
 #[cfg(feature = "async")]
@@ -82,7 +84,7 @@ fn lazy_per_thread_init() {
 /// stack. They will be skipped and not executed.
 #[cfg(has_host_compiler_backend)]
 pub(super) unsafe fn raise_preexisting_trap() -> ! {
-    tls::with(|info| info.unwrap().unwind())
+    tls::with(|info| unsafe { info.unwrap().unwind() })
 }
 
 /// Invokes the closure `f` and returns a `bool` if it succeeded.
@@ -305,6 +307,14 @@ unsafe impl HostResultHasUnwindSentinel for core::convert::Infallible {
     }
 }
 
+unsafe impl HostResultHasUnwindSentinel for bool {
+    type Abi = u32;
+    const SENTINEL: Self::Abi = u32::MAX;
+    fn into_abi(self) -> Self::Abi {
+        u32::from(self)
+    }
+}
+
 /// Stores trace message with backtrace.
 #[derive(Debug)]
 pub struct Trap {
@@ -370,7 +380,7 @@ impl From<wasmtime_environ::Trap> for TrapReason {
 /// longjmp'd over and none of its destructors on the stack may be run.
 pub unsafe fn catch_traps<T, F>(
     store: &mut StoreContextMut<'_, T>,
-    old_state: &EntryStoreContext,
+    old_state: &mut EntryStoreContext,
     mut closure: F,
 ) -> Result<(), Box<Trap>>
 where
@@ -396,21 +406,26 @@ where
         // interpreter since this branch is only ever taken if the interpreter
         // isn't present.
         #[cfg(has_host_compiler_backend)]
-        ExecutorRef::Native => traphandlers::wasmtime_setjmp(
-            cx.jmp_buf.as_ptr(),
-            {
-                extern "C" fn call_closure<F>(payload: *mut u8, caller: NonNull<VMContext>) -> bool
-                where
-                    F: FnMut(NonNull<VMContext>, Option<InterpreterRef<'_>>) -> bool,
+        ExecutorRef::Native => unsafe {
+            traphandlers::wasmtime_setjmp(
+                cx.jmp_buf.as_ptr(),
                 {
-                    unsafe { (*(payload as *mut F))(caller, None) }
-                }
+                    extern "C" fn call_closure<F>(
+                        payload: *mut u8,
+                        caller: NonNull<VMContext>,
+                    ) -> bool
+                    where
+                        F: FnMut(NonNull<VMContext>, Option<InterpreterRef<'_>>) -> bool,
+                    {
+                        unsafe { (*(payload as *mut F))(caller, None) }
+                    }
 
-                call_closure::<F>
-            },
-            &mut closure as *mut F as *mut u8,
-            caller,
-        ),
+                    call_closure::<F>
+                },
+                &mut closure as *mut F as *mut u8,
+                caller,
+            )
+        },
     });
 
     return match result {
@@ -430,7 +445,7 @@ where
 mod call_thread_state {
     use super::*;
     use crate::EntryStoreContext;
-    use crate::runtime::vm::Unwind;
+    use crate::runtime::vm::{Unwind, VMStackChain};
 
     /// Temporary state stored on the stack which is registered in the `tls`
     /// module below for calls into wasm.
@@ -466,8 +481,6 @@ mod call_thread_state {
         pub(crate) unwinder: &'static dyn Unwind,
 
         pub(super) prev: Cell<tls::Ptr>,
-        #[cfg(all(has_native_signals, unix))]
-        pub(crate) async_guard_range: Range<*mut u8>,
 
         // The state of the runtime for the *previous* `CallThreadState` for
         // this same store. Our *current* state is saved in `self.vm_store_context`,
@@ -477,7 +490,7 @@ mod call_thread_state {
         // same store and `self.vm_store_context == self.prev.vm_store_context`) and we must to
         // maintain the list of contiguous-Wasm-frames stack regions for
         // backtracing purposes.
-        old_state: *const EntryStoreContext,
+        old_state: *mut EntryStoreContext,
     }
 
     impl Drop for CallThreadState {
@@ -494,12 +507,8 @@ mod call_thread_state {
         #[inline]
         pub(super) fn new(
             store: &mut StoreOpaque,
-            old_state: *const EntryStoreContext,
+            old_state: *mut EntryStoreContext,
         ) -> CallThreadState {
-            // Don't try to plumb #[cfg] everywhere for this field, just pretend
-            // we're using it on miri/windows to silence compiler warnings.
-            let _: Range<_> = store.async_guard_range();
-
             CallThreadState {
                 unwind: Cell::new(None),
                 unwinder: store.unwinder(),
@@ -510,8 +519,6 @@ mod call_thread_state {
                 #[cfg(feature = "coredump")]
                 capture_coredump: store.engine().config().coredump_on_trap,
                 vm_store_context: store.vm_store_context_ptr(),
-                #[cfg(all(has_native_signals, unix))]
-                async_guard_range: store.async_guard_range(),
                 prev: Cell::new(ptr::null()),
                 old_state,
             }
@@ -519,17 +526,22 @@ mod call_thread_state {
 
         /// Get the saved FP upon exit from Wasm for the previous `CallThreadState`.
         pub unsafe fn old_last_wasm_exit_fp(&self) -> usize {
-            (&*self.old_state).last_wasm_exit_fp
+            unsafe { (&*self.old_state).last_wasm_exit_fp }
         }
 
         /// Get the saved PC upon exit from Wasm for the previous `CallThreadState`.
         pub unsafe fn old_last_wasm_exit_pc(&self) -> usize {
-            (&*self.old_state).last_wasm_exit_pc
+            unsafe { (&*self.old_state).last_wasm_exit_pc }
         }
 
         /// Get the saved FP upon entry into Wasm for the previous `CallThreadState`.
         pub unsafe fn old_last_wasm_entry_fp(&self) -> usize {
-            (&*self.old_state).last_wasm_entry_fp
+            unsafe { (&*self.old_state).last_wasm_entry_fp }
+        }
+
+        /// Get the saved `VMStackChain` for the previous `CallThreadState`.
+        pub unsafe fn old_stack_chain(&self) -> VMStackChain {
+            unsafe { (&*self.old_state).stack_chain.clone() }
         }
 
         /// Get the previous `CallThreadState`.
@@ -568,6 +580,40 @@ mod call_thread_state {
             let head = tls::raw::replace(prev);
             assert!(core::ptr::eq(head, self));
         }
+
+        /// Swaps the state in this `CallThreadState`'s `VMStoreContext` with
+        /// the state in `EntryStoreContext` that was saved when this
+        /// activation was created.
+        ///
+        /// This method is using during suspension of a fiber to restore the
+        /// store back to what it originally was and prepare it to be resumed
+        /// later on. This takes various fields of `VMStoreContext` and swaps
+        /// them with what was saved in `EntryStoreContext`. That restores
+        /// a store to just before this activation was called but saves off the
+        /// fields of this activation to get restored/resumed at a later time.
+        #[cfg(feature = "async")]
+        pub(super) unsafe fn swap(&self) {
+            unsafe fn swap<T>(a: &core::cell::UnsafeCell<T>, b: &mut T) {
+                unsafe { core::mem::swap(&mut *a.get(), b) }
+            }
+
+            unsafe {
+                let cx = self.vm_store_context.as_ref();
+                swap(
+                    &cx.last_wasm_exit_fp,
+                    &mut (*self.old_state).last_wasm_exit_fp,
+                );
+                swap(
+                    &cx.last_wasm_exit_pc,
+                    &mut (*self.old_state).last_wasm_exit_pc,
+                );
+                swap(
+                    &cx.last_wasm_entry_fp,
+                    &mut (*self.old_state).last_wasm_entry_fp,
+                );
+                swap(&cx.stack_chain, &mut (*self.old_state).stack_chain);
+            }
+        }
     }
 }
 pub use call_thread_state::*;
@@ -656,7 +702,9 @@ impl CallThreadState {
     unsafe fn unwind(&self) -> ! {
         debug_assert!(!self.jmp_buf.get().is_null());
         debug_assert!(self.jmp_buf.get() != CallThreadState::JMP_BUF_INTERPRETER_SENTINEL);
-        traphandlers::wasmtime_longjmp(self.jmp_buf.get());
+        unsafe {
+            traphandlers::wasmtime_longjmp(self.jmp_buf.get());
+        }
     }
 
     fn capture_backtrace(
@@ -970,14 +1018,36 @@ pub(crate) mod tls {
         /// that this doesn't push stale data and the data is popped
         /// appropriately.
         pub unsafe fn push(self) -> PreviousAsyncWasmCallState {
+            // First save the state of TLS as-is so when this state is popped
+            // off later on we know where to stop.
+            let ret = PreviousAsyncWasmCallState { state: raw::get() };
+
+            // The oldest activation, if present, has various `VMStoreContext`
+            // fields saved within it. These fields were the state for the
+            // *youngest* activation when a suspension previously happened. By
+            // swapping them back into the store this is an O(1) way of
+            // restoring the state of a store's metadata fields at the time of
+            // the suspension.
+            //
+            // The store's previous values before this function will all get
+            // saved in the oldest activation's state on the stack. The store's
+            // current state then describes the youngest activation which is
+            // restored via the loop below.
+            unsafe {
+                if let Some(state) = self.state.as_ref() {
+                    state.swap();
+                }
+            }
+
             // Our `state` pointer is a linked list of oldest-to-youngest so by
             // pushing in order of the list we restore the youngest-to-oldest
             // list as stored in the state of this current thread.
-            let ret = PreviousAsyncWasmCallState { state: raw::get() };
             let mut ptr = self.state;
-            while let Some(state) = ptr.as_ref() {
-                ptr = state.prev.replace(core::ptr::null_mut());
-                state.push();
+            unsafe {
+                while let Some(state) = ptr.as_ref() {
+                    ptr = state.prev.replace(core::ptr::null_mut());
+                    state.push();
+                }
             }
             ret
         }
@@ -1033,19 +1103,44 @@ pub(crate) mod tls {
             loop {
                 // If the current TLS state is as we originally found it, then
                 // this loop is finished.
+                //
+                // Note, though, that before exiting, if the oldest
+                // `CallThreadState` is present, the current state of
+                // `VMStoreContext` is saved off within it. This will save the
+                // current state, before this function, of `VMStoreContext`
+                // into the `EntryStoreContext` stored with the oldest
+                // activation. This is a bit counter-intuitive where the state
+                // for the youngest activation is stored in the "old" state
+                // of the oldest activation.
+                //
+                // What this does is restores the state of the store to just
+                // before this async fiber was started. The fiber's state will
+                // be entirely self-contained in the fiber itself and the
+                // returned `AsyncWasmCallState`. Resumption above in
+                // `AsyncWasmCallState::push` will perform the swap back into
+                // the store to hook things up again.
                 let ptr = raw::get();
                 if ptr == thread_head {
+                    unsafe {
+                        if let Some(state) = ret.state.as_ref() {
+                            state.swap();
+                        }
+                    }
+
                     break ret;
                 }
 
                 // Pop this activation from the current thread's TLS state, and
                 // then afterwards push it onto our own linked list within this
-                // `AsyncWasmCallState`. Note that the linked list in `AsyncWasmCallState` is stored
-                // in reverse order so a subsequent `push` later on pushes
-                // everything in the right order.
-                (*ptr).pop();
-                if let Some(state) = ret.state.as_ref() {
-                    (*ptr).prev.set(state);
+                // `AsyncWasmCallState`. Note that the linked list in
+                // `AsyncWasmCallState` is stored in reverse order so a
+                // subsequent `push` later on pushes everything in the right
+                // order.
+                unsafe {
+                    (*ptr).pop();
+                    if let Some(state) = ret.state.as_ref() {
+                        (*ptr).prev.set(state);
+                    }
                 }
                 ret.state = ptr;
             }
