@@ -62,7 +62,7 @@ impl From<WasmError> for CompileError {
 }
 
 impl core::error::Error for CompileError {
-    fn source(&self) -> Option<&(dyn std::error::Error + 'static)> {
+    fn source(&self) -> Option<&(dyn core::error::Error + 'static)> {
         match self {
             CompileError::Wasm(e) => Some(e),
             _ => None,
@@ -91,7 +91,7 @@ pub enum RelocationTarget {
 pub trait CacheStore: Send + Sync + std::fmt::Debug {
     /// Try to retrieve an arbitrary cache key entry, and returns a reference to bytes that were
     /// inserted via `Self::insert` before.
-    fn get(&self, key: &[u8]) -> Option<Cow<[u8]>>;
+    fn get(&self, key: &[u8]) -> Option<Cow<'_, [u8]>>;
 
     /// Given an arbitrary key and bytes, stores them in the cache.
     ///
@@ -180,7 +180,7 @@ pub enum SettingKind {
 pub struct CompiledFunctionBody {
     /// The code. This is whatever type the `Compiler` implementation wants it
     /// to be, we just shepherd it around.
-    pub code: Box<dyn Any + Send>,
+    pub code: Box<dyn Any + Send + Sync>,
     /// Whether the compiled function needs a GC heap to run; that is, whether
     /// it reads a struct field, allocates, an array, or etc...
     pub needs_gc_heap: bool,
@@ -188,7 +188,69 @@ pub struct CompiledFunctionBody {
 
 /// An implementation of a compiler which can compile WebAssembly functions to
 /// machine code and perform other miscellaneous tasks needed by the JIT runtime.
+///
+/// The diagram below depicts typical usage of this trait:
+///
+/// ```ignore
+///                     +------+
+///                     | Wasm |
+///                     +------+
+///                        |
+///                        |
+///           Compiler::compile_function()
+///                        |
+///                        |
+///                        V
+///             +----------------------+
+///             | CompiledFunctionBody |
+///             +----------------------+
+///               |                  |
+///               |                  |
+///               |                When
+///               |       Compiler::inlining_compiler()
+///               |               is some
+///               |                  |
+///             When                 |
+/// Compiler::inlining_compiler()    |-----------------.
+///             is none              |                 |
+///               |                  |                 |
+///               |           Optionally call          |
+///               |        InliningCompiler::inline()  |
+///               |                  |                 |
+///               |                  |                 |
+///               |                  |-----------------'
+///               |                  |
+///               |                  |
+///               |                  V
+///               |     InliningCompiler::finish_compiling()
+///               |                  |
+///               |                  |
+///               |------------------'
+///               |
+///               |
+///   Compiler::append_code()
+///               |
+///               |
+///               V
+///           +--------+
+///           | Object |
+///           +--------+
+/// ```
 pub trait Compiler: Send + Sync {
+    /// Get this compiler's inliner.
+    ///
+    /// Consumers of this trait **must** check for when when this method returns
+    /// `Some(_)`, and **must** call `InliningCompiler::finish_compiling` on all
+    /// `CompiledFunctionBody`s produced by this compiler in that case before
+    /// passing the the compiled functions to `Compiler::append_code`, even if
+    /// the consumer does not actually intend to do any inlining. This allows
+    /// implementations of the trait to only translate to an internal
+    /// representation in `Compiler::compile_*` methods so that they can then
+    /// perform inlining afterwards if the consumer desires, and then finally
+    /// proceed with compilng that internal representation to native code in
+    /// `InliningCompiler::finish_compiling`.
+    fn inlining_compiler(&self) -> Option<&dyn InliningCompiler>;
+
     /// Compiles the function `index` within `translation`.
     ///
     /// The body of the function is available in `data` and configuration
@@ -200,6 +262,7 @@ pub trait Compiler: Send + Sync {
         index: DefinedFuncIndex,
         data: FunctionBodyData<'_>,
         types: &ModuleTypesBuilder,
+        symbol: &str,
     ) -> Result<CompiledFunctionBody, CompileError>;
 
     /// Compile a trampoline for an array-call host function caller calling the
@@ -212,6 +275,7 @@ pub trait Compiler: Send + Sync {
         translation: &ModuleTranslation<'_>,
         types: &ModuleTypesBuilder,
         index: DefinedFuncIndex,
+        symbol: &str,
     ) -> Result<CompiledFunctionBody, CompileError>;
 
     /// Compile a trampoline for a Wasm caller calling a array callee with the
@@ -222,6 +286,7 @@ pub trait Compiler: Send + Sync {
     fn compile_wasm_to_array_trampoline(
         &self,
         wasm_func_ty: &WasmFuncType,
+        symbol: &str,
     ) -> Result<CompiledFunctionBody, CompileError>;
 
     /// Creates a trampoline that can be used to call Wasmtime's implementation
@@ -236,6 +301,7 @@ pub trait Compiler: Send + Sync {
     fn compile_wasm_to_builtin(
         &self,
         index: BuiltinFunctionIndex,
+        symbol: &str,
     ) -> Result<CompiledFunctionBody, CompileError>;
 
     /// Returns the list of relocations required for a function from one of the
@@ -275,7 +341,7 @@ pub trait Compiler: Send + Sync {
     fn append_code(
         &self,
         obj: &mut Object<'static>,
-        funcs: &[(String, Box<dyn Any + Send>)],
+        funcs: &[(String, Box<dyn Any + Send + Sync>)],
         resolve_reloc: &dyn Fn(usize, RelocationTarget) -> usize,
     ) -> Result<Vec<(SymbolId, FunctionLoc)>>;
 
@@ -386,7 +452,7 @@ pub trait Compiler: Send + Sync {
         get_func: &'a dyn Fn(
             StaticModuleIndex,
             DefinedFuncIndex,
-        ) -> (SymbolId, &'a (dyn Any + Send)),
+        ) -> (SymbolId, &'a (dyn Any + Send + Sync)),
         dwarf_package_bytes: Option<&'a [u8]>,
         tunables: &'a Tunables,
     ) -> Result<()>;
@@ -399,3 +465,36 @@ pub trait Compiler: Send + Sync {
         None
     }
 }
+
+/// An inlining compiler.
+pub trait InliningCompiler: Sync + Send {
+    /// Enumerate the function calls that the given `func` makes.
+    fn calls(&self, func: &CompiledFunctionBody, calls: &mut IndexSet<FuncIndex>) -> Result<()>;
+
+    /// Get the abstract size of the given function, for the purposes of
+    /// inlining heuristics.
+    fn size(&self, func: &CompiledFunctionBody) -> u32;
+
+    /// Process this function for inlining.
+    ///
+    /// Implementations should call `get_callee` for each of their direct
+    /// function call sites and if `get_callee` returns `Some(_)`, they should
+    /// inline the given function body into that call site.
+    fn inline<'a>(
+        &self,
+        func: &mut CompiledFunctionBody,
+        get_callee: &'a mut dyn FnMut(FuncIndex) -> Option<&'a CompiledFunctionBody>,
+    ) -> Result<()>;
+
+    /// Finish compiling the given function.
+    ///
+    /// This method **must** be called before passing the
+    /// `CompiledFunctionBody`'s contents to `Compiler::append_code`, even if no
+    /// inlining was performed.
+    fn finish_compiling(
+        &self,
+        func: &mut CompiledFunctionBody,
+        input: Option<wasmparser::FunctionBody<'_>>,
+        symbol: &str,
+    ) -> Result<()>;
+}
