@@ -1,4 +1,5 @@
 use crate::ScopeVec;
+use crate::component::dfg::AbstractInstantiations;
 use crate::component::*;
 use crate::prelude::*;
 use crate::{
@@ -8,6 +9,8 @@ use crate::{
 };
 use anyhow::anyhow;
 use anyhow::{Result, bail};
+use cranelift_entity::SecondaryMap;
+use cranelift_entity::packed_option::PackedOption;
 use indexmap::IndexMap;
 use std::collections::HashMap;
 use std::mem;
@@ -177,7 +180,6 @@ enum LocalInitializer<'data> {
     Lower {
         func: ComponentFuncIndex,
         lower_ty: ComponentFuncTypeId,
-        canonical_abi: ModuleInternedTypeIndex,
         options: LocalCanonicalOptions,
     },
     Lift(ComponentFuncTypeId, FuncIndex, LocalCanonicalOptions),
@@ -192,22 +194,20 @@ enum LocalInitializer<'data> {
         func: ModuleInternedTypeIndex,
     },
     TaskReturn {
-        func: ModuleInternedTypeIndex,
         result: Option<ComponentValType>,
         options: LocalCanonicalOptions,
     },
+    TaskCancel {
+        func: ModuleInternedTypeIndex,
+    },
     WaitableSetNew {
         func: ModuleInternedTypeIndex,
     },
     WaitableSetWait {
-        func: ModuleInternedTypeIndex,
-        async_: bool,
-        memory: MemoryIndex,
+        options: LocalCanonicalOptions,
     },
     WaitableSetPoll {
-        func: ModuleInternedTypeIndex,
-        async_: bool,
-        memory: MemoryIndex,
+        options: LocalCanonicalOptions,
     },
     WaitableSetDrop {
         func: ModuleInternedTypeIndex,
@@ -222,18 +222,20 @@ enum LocalInitializer<'data> {
     SubtaskDrop {
         func: ModuleInternedTypeIndex,
     },
+    SubtaskCancel {
+        func: ModuleInternedTypeIndex,
+        async_: bool,
+    },
     StreamNew {
         ty: ComponentDefinedTypeId,
         func: ModuleInternedTypeIndex,
     },
     StreamRead {
         ty: ComponentDefinedTypeId,
-        func: ModuleInternedTypeIndex,
         options: LocalCanonicalOptions,
     },
     StreamWrite {
         ty: ComponentDefinedTypeId,
-        func: ModuleInternedTypeIndex,
         options: LocalCanonicalOptions,
     },
     StreamCancelRead {
@@ -246,11 +248,11 @@ enum LocalInitializer<'data> {
         func: ModuleInternedTypeIndex,
         async_: bool,
     },
-    StreamCloseReadable {
+    StreamDropReadable {
         ty: ComponentDefinedTypeId,
         func: ModuleInternedTypeIndex,
     },
-    StreamCloseWritable {
+    StreamDropWritable {
         ty: ComponentDefinedTypeId,
         func: ModuleInternedTypeIndex,
     },
@@ -260,12 +262,10 @@ enum LocalInitializer<'data> {
     },
     FutureRead {
         ty: ComponentDefinedTypeId,
-        func: ModuleInternedTypeIndex,
         options: LocalCanonicalOptions,
     },
     FutureWrite {
         ty: ComponentDefinedTypeId,
-        func: ModuleInternedTypeIndex,
         options: LocalCanonicalOptions,
     },
     FutureCancelRead {
@@ -278,25 +278,31 @@ enum LocalInitializer<'data> {
         func: ModuleInternedTypeIndex,
         async_: bool,
     },
-    FutureCloseReadable {
+    FutureDropReadable {
         ty: ComponentDefinedTypeId,
         func: ModuleInternedTypeIndex,
     },
-    FutureCloseWritable {
+    FutureDropWritable {
         ty: ComponentDefinedTypeId,
         func: ModuleInternedTypeIndex,
     },
     ErrorContextNew {
-        func: ModuleInternedTypeIndex,
         options: LocalCanonicalOptions,
     },
     ErrorContextDebugMessage {
-        func: ModuleInternedTypeIndex,
         options: LocalCanonicalOptions,
     },
     ErrorContextDrop {
         func: ModuleInternedTypeIndex,
     },
+    ContextGet {
+        func: ModuleInternedTypeIndex,
+        i: u32,
+    },
+    ContextSet {
+        func: ModuleInternedTypeIndex,
+        i: u32,
+    },
 
     // core wasm modules
     ModuleStatic(StaticModuleIndex, ComponentCoreModuleTypeId),
@@ -360,14 +366,30 @@ enum ClosedOverModule {
     Upvar(ModuleUpvarIndex),
 }
 
+/// The data model for objects that are not unboxed in locals.
+#[derive(Debug, Clone, Hash, Eq, PartialEq)]
+pub enum LocalDataModel {
+    /// Data is stored in GC objects.
+    Gc {},
+
+    /// Data is stored in a linear memory.
+    LinearMemory {
+        /// An optional memory definition supplied.
+        memory: Option<MemoryIndex>,
+        /// An optional definition of `realloc` to used.
+        realloc: Option<FuncIndex>,
+    },
+}
+
 /// Representation of canonical ABI options.
 struct LocalCanonicalOptions {
     string_encoding: StringEncoding,
-    memory: Option<MemoryIndex>,
-    realloc: Option<FuncIndex>,
     post_return: Option<FuncIndex>,
     async_: bool,
     callback: Option<FuncIndex>,
+    /// The type index of the core GC types signature.
+    core_type: ModuleInternedTypeIndex,
+    data_model: LocalDataModel,
 }
 
 enum Action {
@@ -469,12 +491,133 @@ impl<'a, 'data> Translator<'a, 'data> {
             &self.static_modules,
             &self.static_components,
         )?;
+
         self.partition_adapter_modules(&mut component);
+
         let translation =
             component.finish(self.types.types_mut_for_inlining(), self.result.types_ref())?;
+
+        self.analyze_function_imports(&translation);
+
         Ok((translation, self.static_modules))
     }
 
+    fn analyze_function_imports(&mut self, translation: &ComponentTranslation) {
+        // First, abstract interpret the initializers to create a map from each
+        // static module to its abstract set of instantiations.
+        let mut instantiations = SecondaryMap::<StaticModuleIndex, AbstractInstantiations>::new();
+        let mut instance_to_module =
+            PrimaryMap::<RuntimeInstanceIndex, PackedOption<StaticModuleIndex>>::new();
+        for init in &translation.component.initializers {
+            match init {
+                GlobalInitializer::InstantiateModule(instantiation) => match instantiation {
+                    InstantiateModule::Static(module, args) => {
+                        instantiations[*module].join(AbstractInstantiations::One(&*args));
+                        instance_to_module.push(Some(*module).into());
+                    }
+                    _ => {
+                        instance_to_module.push(None.into());
+                    }
+                },
+                _ => continue,
+            }
+        }
+
+        // Second, make sure to mark exported modules as instantiated many
+        // times, since they could be linked with who-knows-what at runtime.
+        for item in translation.component.export_items.values() {
+            if let Export::ModuleStatic { index, .. } = item {
+                instantiations[*index].join(AbstractInstantiations::Many)
+            }
+        }
+
+        // Finally, iterate over our instantiations and record statically-known
+        // function imports so that they can get translated into direct calls
+        // (and eventually get inlined) rather than indirect calls through the
+        // imports table.
+        for (module, instantiations) in instantiations.iter() {
+            let args = match instantiations {
+                dfg::AbstractInstantiations::Many | dfg::AbstractInstantiations::None => continue,
+                dfg::AbstractInstantiations::One(args) => args,
+            };
+
+            let mut imported_func_counter = 0_u32;
+            for (i, arg) in args.iter().enumerate() {
+                // Only consider function imports.
+                let (_, _, crate::types::EntityType::Function(_)) =
+                    self.static_modules[module].module.import(i).unwrap()
+                else {
+                    continue;
+                };
+
+                let imported_func = FuncIndex::from_u32(imported_func_counter);
+                imported_func_counter += 1;
+                debug_assert!(
+                    self.static_modules[module]
+                        .module
+                        .defined_func_index(imported_func)
+                        .is_none()
+                );
+
+                match arg {
+                    CoreDef::InstanceFlags(_) => unreachable!("instance flags are not a function"),
+
+                    // We could in theory inline these trampolines, so it could
+                    // potentially make sense to record that we know this
+                    // imported function is this particular trampoline. However,
+                    // everything else is based around (module,
+                    // defined-function) pairs and these trampolines don't fit
+                    // that paradigm. Also, inlining trampolines gets really
+                    // tricky when we consider the stack pointer, frame pointer,
+                    // and return address note-taking that they do for the
+                    // purposes of stack walking. We could, with enough effort,
+                    // turn them into direct calls even though we probably
+                    // wouldn't ever inline them, but it just doesn't seem worth
+                    // the effort.
+                    CoreDef::Trampoline(_) => continue,
+
+                    // This imported function is an export from another
+                    // instance, a perfect candidate for becoming an inlinable
+                    // direct call!
+                    CoreDef::Export(export) => {
+                        let Some(arg_module) = &instance_to_module[export.instance].expand() else {
+                            // Instance of a dynamic module that is not part of
+                            // this component, not a statically-known module
+                            // inside this component. We have to do an indirect
+                            // call.
+                            continue;
+                        };
+
+                        let ExportItem::Index(EntityIndex::Function(arg_func)) = &export.item
+                        else {
+                            unreachable!("function imports must be functions")
+                        };
+
+                        let Some(arg_module_def_func) = self.static_modules[*arg_module]
+                            .module
+                            .defined_func_index(*arg_func)
+                        else {
+                            // TODO: we should ideally follow re-export chains
+                            // to bottom out the instantiation argument in
+                            // either a definition or an import at the root
+                            // component boundary. In practice, this pattern is
+                            // rare, so following these chains is left for the
+                            // Future.
+                            continue;
+                        };
+
+                        assert!(
+                            self.static_modules[module].known_imported_functions[imported_func]
+                                .is_none()
+                        );
+                        self.static_modules[module].known_imported_functions[imported_func] =
+                            Some((*arg_module, arg_module_def_func));
+                    }
+                }
+            }
+        }
+    }
+
     fn translate_payload(
         &mut self,
         payload: Payload<'data>,
@@ -588,43 +731,59 @@ impl<'a, 'data> Translator<'a, 'data> {
                 let mut core_func_index = types.function_count();
                 self.validator.component_canonical_section(&s)?;
                 for func in s {
-                    let types = self.validator.types(0).unwrap();
                     let init = match func? {
                         wasmparser::CanonicalFunction::Lift {
                             type_index,
                             core_func_index,
                             options,
                         } => {
-                            let ty = types.component_any_type_at(type_index).unwrap_func();
+                            let ty = self
+                                .validator
+                                .types(0)
+                                .unwrap()
+                                .component_any_type_at(type_index)
+                                .unwrap_func();
+
                             let func = FuncIndex::from_u32(core_func_index);
-                            let options = self.canonical_options(&options);
+                            let options = self.canonical_options(&options, core_func_index)?;
                             LocalInitializer::Lift(ty, func, options)
                         }
                         wasmparser::CanonicalFunction::Lower {
                             func_index,
                             options,
                         } => {
-                            let lower_ty = types.component_function_at(func_index);
+                            let lower_ty = self
+                                .validator
+                                .types(0)
+                                .unwrap()
+                                .component_function_at(func_index);
                             let func = ComponentFuncIndex::from_u32(func_index);
-                            let options = self.canonical_options(&options);
-                            let canonical_abi = self.core_func_signature(core_func_index)?;
-
+                            let options = self.canonical_options(&options, core_func_index)?;
                             core_func_index += 1;
                             LocalInitializer::Lower {
                                 func,
                                 options,
-                                canonical_abi,
                                 lower_ty,
                             }
                         }
                         wasmparser::CanonicalFunction::ResourceNew { resource } => {
-                            let resource = types.component_any_type_at(resource).unwrap_resource();
+                            let resource = self
+                                .validator
+                                .types(0)
+                                .unwrap()
+                                .component_any_type_at(resource)
+                                .unwrap_resource();
                             let ty = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
                             LocalInitializer::ResourceNew(resource, ty)
                         }
                         wasmparser::CanonicalFunction::ResourceDrop { resource } => {
-                            let resource = types.component_any_type_at(resource).unwrap_resource();
+                            let resource = self
+                                .validator
+                                .types(0)
+                                .unwrap()
+                                .component_any_type_at(resource)
+                                .unwrap_resource();
                             let ty = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
                             LocalInitializer::ResourceDrop(resource, ty)
@@ -634,11 +793,21 @@ impl<'a, 'data> Translator<'a, 'data> {
                             bail!("support for `resource.drop async` not implemented yet")
                         }
                         wasmparser::CanonicalFunction::ResourceRep { resource } => {
-                            let resource = types.component_any_type_at(resource).unwrap_resource();
+                            let resource = self
+                                .validator
+                                .types(0)
+                                .unwrap()
+                                .component_any_type_at(resource)
+                                .unwrap_resource();
                             let ty = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
                             LocalInitializer::ResourceRep(resource, ty)
                         }
+                        wasmparser::CanonicalFunction::ThreadSpawnRef { .. }
+                        | wasmparser::CanonicalFunction::ThreadSpawnIndirect { .. }
+                        | wasmparser::CanonicalFunction::ThreadAvailableParallelism => {
+                            bail!("unsupported intrinsic")
+                        }
                         wasmparser::CanonicalFunction::BackpressureSet => {
                             let core_type = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
@@ -649,18 +818,21 @@ impl<'a, 'data> Translator<'a, 'data> {
                                 wasmparser::ComponentValType::Primitive(ty) => {
                                     ComponentValType::Primitive(ty)
                                 }
-                                wasmparser::ComponentValType::Type(ty) => {
-                                    ComponentValType::Type(types.component_defined_type_at(ty))
-                                }
+                                wasmparser::ComponentValType::Type(ty) => ComponentValType::Type(
+                                    self.validator
+                                        .types(0)
+                                        .unwrap()
+                                        .component_defined_type_at(ty),
+                                ),
                             });
-                            let options = self.canonical_options(&options);
+                            let options = self.canonical_options(&options, core_func_index)?;
+                            core_func_index += 1;
+                            LocalInitializer::TaskReturn { result, options }
+                        }
+                        wasmparser::CanonicalFunction::TaskCancel => {
                             let func = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
-                            LocalInitializer::TaskReturn {
-                                func,
-                                result,
-                                options,
-                            }
+                            LocalInitializer::TaskCancel { func }
                         }
                         wasmparser::CanonicalFunction::WaitableSetNew => {
                             let func = self.core_func_signature(core_func_index)?;
@@ -668,21 +840,37 @@ impl<'a, 'data> Translator<'a, 'data> {
                             LocalInitializer::WaitableSetNew { func }
                         }
                         wasmparser::CanonicalFunction::WaitableSetWait { async_, memory } => {
-                            let func = self.core_func_signature(core_func_index)?;
+                            let core_type = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
                             LocalInitializer::WaitableSetWait {
-                                func,
-                                async_,
-                                memory: MemoryIndex::from_u32(memory),
+                                options: LocalCanonicalOptions {
+                                    core_type,
+                                    async_,
+                                    data_model: LocalDataModel::LinearMemory {
+                                        memory: Some(MemoryIndex::from_u32(memory)),
+                                        realloc: None,
+                                    },
+                                    post_return: None,
+                                    callback: None,
+                                    string_encoding: StringEncoding::Utf8,
+                                },
                             }
                         }
                         wasmparser::CanonicalFunction::WaitableSetPoll { async_, memory } => {
-                            let func = self.core_func_signature(core_func_index)?;
+                            let core_type = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
                             LocalInitializer::WaitableSetPoll {
-                                func,
-                                async_,
-                                memory: MemoryIndex::from_u32(memory),
+                                options: LocalCanonicalOptions {
+                                    core_type,
+                                    async_,
+                                    data_model: LocalDataModel::LinearMemory {
+                                        memory: Some(MemoryIndex::from_u32(memory)),
+                                        realloc: None,
+                                    },
+                                    post_return: None,
+                                    callback: None,
+                                    string_encoding: StringEncoding::Utf8,
+                                },
                             }
                         }
                         wasmparser::CanonicalFunction::WaitableSetDrop => {
@@ -705,119 +893,175 @@ impl<'a, 'data> Translator<'a, 'data> {
                             core_func_index += 1;
                             LocalInitializer::SubtaskDrop { func }
                         }
+                        wasmparser::CanonicalFunction::SubtaskCancel { async_ } => {
+                            let func = self.core_func_signature(core_func_index)?;
+                            core_func_index += 1;
+                            LocalInitializer::SubtaskCancel { func, async_ }
+                        }
                         wasmparser::CanonicalFunction::StreamNew { ty } => {
-                            let ty = types.component_defined_type_at(ty);
+                            let ty = self
+                                .validator
+                                .types(0)
+                                .unwrap()
+                                .component_defined_type_at(ty);
                             let func = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
                             LocalInitializer::StreamNew { ty, func }
                         }
                         wasmparser::CanonicalFunction::StreamRead { ty, options } => {
-                            let ty = types.component_defined_type_at(ty);
-                            let options = self.canonical_options(&options);
-                            let func = self.core_func_signature(core_func_index)?;
+                            let ty = self
+                                .validator
+                                .types(0)
+                                .unwrap()
+                                .component_defined_type_at(ty);
+                            let options = self.canonical_options(&options, core_func_index)?;
                             core_func_index += 1;
-                            LocalInitializer::StreamRead { ty, func, options }
+                            LocalInitializer::StreamRead { ty, options }
                         }
                         wasmparser::CanonicalFunction::StreamWrite { ty, options } => {
-                            let ty = types.component_defined_type_at(ty);
-                            let options = self.canonical_options(&options);
-                            let func = self.core_func_signature(core_func_index)?;
+                            let ty = self
+                                .validator
+                                .types(0)
+                                .unwrap()
+                                .component_defined_type_at(ty);
+                            let options = self.canonical_options(&options, core_func_index)?;
                             core_func_index += 1;
-                            LocalInitializer::StreamWrite { ty, func, options }
+                            LocalInitializer::StreamWrite { ty, options }
                         }
                         wasmparser::CanonicalFunction::StreamCancelRead { ty, async_ } => {
-                            let ty = types.component_defined_type_at(ty);
+                            let ty = self
+                                .validator
+                                .types(0)
+                                .unwrap()
+                                .component_defined_type_at(ty);
                             let func = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
                             LocalInitializer::StreamCancelRead { ty, func, async_ }
                         }
                         wasmparser::CanonicalFunction::StreamCancelWrite { ty, async_ } => {
-                            let ty = types.component_defined_type_at(ty);
+                            let ty = self
+                                .validator
+                                .types(0)
+                                .unwrap()
+                                .component_defined_type_at(ty);
                             let func = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
                             LocalInitializer::StreamCancelWrite { ty, func, async_ }
                         }
-                        wasmparser::CanonicalFunction::StreamCloseReadable { ty } => {
-                            let ty = types.component_defined_type_at(ty);
+                        wasmparser::CanonicalFunction::StreamDropReadable { ty } => {
+                            let ty = self
+                                .validator
+                                .types(0)
+                                .unwrap()
+                                .component_defined_type_at(ty);
                             let func = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
-                            LocalInitializer::StreamCloseReadable { ty, func }
+                            LocalInitializer::StreamDropReadable { ty, func }
                         }
-                        wasmparser::CanonicalFunction::StreamCloseWritable { ty } => {
-                            let ty = types.component_defined_type_at(ty);
+                        wasmparser::CanonicalFunction::StreamDropWritable { ty } => {
+                            let ty = self
+                                .validator
+                                .types(0)
+                                .unwrap()
+                                .component_defined_type_at(ty);
                             let func = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
-                            LocalInitializer::StreamCloseWritable { ty, func }
+                            LocalInitializer::StreamDropWritable { ty, func }
                         }
                         wasmparser::CanonicalFunction::FutureNew { ty } => {
-                            let ty = types.component_defined_type_at(ty);
+                            let ty = self
+                                .validator
+                                .types(0)
+                                .unwrap()
+                                .component_defined_type_at(ty);
                             let func = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
                             LocalInitializer::FutureNew { ty, func }
                         }
                         wasmparser::CanonicalFunction::FutureRead { ty, options } => {
-                            let ty = types.component_defined_type_at(ty);
-                            let options = self.canonical_options(&options);
-                            let func = self.core_func_signature(core_func_index)?;
+                            let ty = self
+                                .validator
+                                .types(0)
+                                .unwrap()
+                                .component_defined_type_at(ty);
+                            let options = self.canonical_options(&options, core_func_index)?;
                             core_func_index += 1;
-                            LocalInitializer::FutureRead { ty, func, options }
+                            LocalInitializer::FutureRead { ty, options }
                         }
                         wasmparser::CanonicalFunction::FutureWrite { ty, options } => {
-                            let ty = types.component_defined_type_at(ty);
-                            let options = self.canonical_options(&options);
-                            let func = self.core_func_signature(core_func_index)?;
+                            let ty = self
+                                .validator
+                                .types(0)
+                                .unwrap()
+                                .component_defined_type_at(ty);
+                            let options = self.canonical_options(&options, core_func_index)?;
                             core_func_index += 1;
-                            LocalInitializer::FutureWrite { ty, func, options }
+                            LocalInitializer::FutureWrite { ty, options }
                         }
                         wasmparser::CanonicalFunction::FutureCancelRead { ty, async_ } => {
-                            let ty = types.component_defined_type_at(ty);
+                            let ty = self
+                                .validator
+                                .types(0)
+                                .unwrap()
+                                .component_defined_type_at(ty);
                             let func = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
                             LocalInitializer::FutureCancelRead { ty, func, async_ }
                         }
                         wasmparser::CanonicalFunction::FutureCancelWrite { ty, async_ } => {
-                            let ty = types.component_defined_type_at(ty);
+                            let ty = self
+                                .validator
+                                .types(0)
+                                .unwrap()
+                                .component_defined_type_at(ty);
                             let func = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
                             LocalInitializer::FutureCancelWrite { ty, func, async_ }
                         }
-                        wasmparser::CanonicalFunction::FutureCloseReadable { ty } => {
-                            let ty = types.component_defined_type_at(ty);
+                        wasmparser::CanonicalFunction::FutureDropReadable { ty } => {
+                            let ty = self
+                                .validator
+                                .types(0)
+                                .unwrap()
+                                .component_defined_type_at(ty);
                             let func = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
-                            LocalInitializer::FutureCloseReadable { ty, func }
+                            LocalInitializer::FutureDropReadable { ty, func }
                         }
-                        wasmparser::CanonicalFunction::FutureCloseWritable { ty } => {
-                            let ty = types.component_defined_type_at(ty);
+                        wasmparser::CanonicalFunction::FutureDropWritable { ty } => {
+                            let ty = self
+                                .validator
+                                .types(0)
+                                .unwrap()
+                                .component_defined_type_at(ty);
                             let func = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
-                            LocalInitializer::FutureCloseWritable { ty, func }
+                            LocalInitializer::FutureDropWritable { ty, func }
                         }
                         wasmparser::CanonicalFunction::ErrorContextNew { options } => {
-                            let options = self.canonical_options(&options);
-                            let func = self.core_func_signature(core_func_index)?;
+                            let options = self.canonical_options(&options, core_func_index)?;
                             core_func_index += 1;
-                            LocalInitializer::ErrorContextNew { func, options }
+                            LocalInitializer::ErrorContextNew { options }
                         }
                         wasmparser::CanonicalFunction::ErrorContextDebugMessage { options } => {
-                            let options = self.canonical_options(&options);
-                            let func = self.core_func_signature(core_func_index)?;
+                            let options = self.canonical_options(&options, core_func_index)?;
                             core_func_index += 1;
-                            LocalInitializer::ErrorContextDebugMessage { func, options }
+                            LocalInitializer::ErrorContextDebugMessage { options }
                         }
                         wasmparser::CanonicalFunction::ErrorContextDrop => {
                             let func = self.core_func_signature(core_func_index)?;
                             core_func_index += 1;
                             LocalInitializer::ErrorContextDrop { func }
                         }
-                        wasmparser::CanonicalFunction::ContextGet(..)
-                        | wasmparser::CanonicalFunction::ContextSet(..)
-                        | wasmparser::CanonicalFunction::TaskCancel
-                        | wasmparser::CanonicalFunction::SubtaskCancel { .. }
-                        | wasmparser::CanonicalFunction::ThreadSpawnRef { .. }
-                        | wasmparser::CanonicalFunction::ThreadSpawnIndirect { .. }
-                        | wasmparser::CanonicalFunction::ThreadAvailableParallelism => {
-                            bail!("unsupported intrinsic")
+                        wasmparser::CanonicalFunction::ContextGet(i) => {
+                            let func = self.core_func_signature(core_func_index)?;
+                            core_func_index += 1;
+                            LocalInitializer::ContextGet { i, func }
+                        }
+                        wasmparser::CanonicalFunction::ContextSet(i) => {
+                            let func = self.core_func_signature(core_func_index)?;
+                            core_func_index += 1;
+                            LocalInitializer::ContextSet { i, func }
                         }
                     };
                     self.result.initializers.push(init);
@@ -1182,49 +1426,78 @@ impl<'a, 'data> Translator<'a, 'data> {
         }
     }
 
-    fn canonical_options(&self, opts: &[wasmparser::CanonicalOption]) -> LocalCanonicalOptions {
-        let mut ret = LocalCanonicalOptions {
-            string_encoding: StringEncoding::Utf8,
-            memory: None,
-            realloc: None,
-            post_return: None,
-            async_: false,
-            callback: None,
-        };
+    fn canonical_options(
+        &mut self,
+        opts: &[wasmparser::CanonicalOption],
+        core_func_index: u32,
+    ) -> WasmResult<LocalCanonicalOptions> {
+        let core_type = self.core_func_signature(core_func_index)?;
+
+        let mut string_encoding = StringEncoding::Utf8;
+        let mut post_return = None;
+        let mut async_ = false;
+        let mut callback = None;
+        let mut memory = None;
+        let mut realloc = None;
+        let mut gc = false;
+
         for opt in opts {
             match opt {
                 wasmparser::CanonicalOption::UTF8 => {
-                    ret.string_encoding = StringEncoding::Utf8;
+                    string_encoding = StringEncoding::Utf8;
                 }
                 wasmparser::CanonicalOption::UTF16 => {
-                    ret.string_encoding = StringEncoding::Utf16;
+                    string_encoding = StringEncoding::Utf16;
                 }
                 wasmparser::CanonicalOption::CompactUTF16 => {
-                    ret.string_encoding = StringEncoding::CompactUtf16;
+                    string_encoding = StringEncoding::CompactUtf16;
                 }
                 wasmparser::CanonicalOption::Memory(idx) => {
                     let idx = MemoryIndex::from_u32(*idx);
-                    ret.memory = Some(idx);
+                    memory = Some(idx);
                 }
                 wasmparser::CanonicalOption::Realloc(idx) => {
                     let idx = FuncIndex::from_u32(*idx);
-                    ret.realloc = Some(idx);
+                    realloc = Some(idx);
                 }
                 wasmparser::CanonicalOption::PostReturn(idx) => {
                     let idx = FuncIndex::from_u32(*idx);
-                    ret.post_return = Some(idx);
+                    post_return = Some(idx);
                 }
-                wasmparser::CanonicalOption::Async => ret.async_ = true,
+                wasmparser::CanonicalOption::Async => async_ = true,
                 wasmparser::CanonicalOption::Callback(idx) => {
                     let idx = FuncIndex::from_u32(*idx);
-                    ret.callback = Some(idx);
+                    callback = Some(idx);
                 }
-                wasmparser::CanonicalOption::CoreType(_) | wasmparser::CanonicalOption::Gc => {
-                    todo!("component model and GC support")
+                wasmparser::CanonicalOption::CoreType(idx) => {
+                    if cfg!(debug_assertions) {
+                        let types = self.validator.types(0).unwrap();
+                        let core_ty_id = types.core_type_at_in_component(*idx).unwrap_sub();
+                        let interned = self
+                            .types
+                            .module_types_builder()
+                            .intern_type(types, core_ty_id)?;
+                        debug_assert_eq!(interned, core_type);
+                    }
+                }
+                wasmparser::CanonicalOption::Gc => {
+                    gc = true;
                 }
             }
         }
-        return ret;
+
+        Ok(LocalCanonicalOptions {
+            string_encoding,
+            post_return,
+            async_,
+            callback,
+            core_type,
+            data_model: if gc {
+                LocalDataModel::Gc {}
+            } else {
+                LocalDataModel::LinearMemory { memory, realloc }
+            },
+        })
     }
 
     /// Get the interned type index for the `index`th core function.
