@@ -142,6 +142,23 @@ pub struct Adapter {
     pub func: dfg::CoreDef,
 }
 
+/// The data model for objects that are not unboxed in locals.
+#[derive(Debug, Clone, Hash, Eq, PartialEq)]
+pub enum DataModel {
+    /// Data is stored in GC objects.
+    Gc {},
+
+    /// Data is stored in a linear memory.
+    LinearMemory {
+        /// An optional memory definition supplied.
+        memory: Option<dfg::CoreExport<MemoryIndex>>,
+        /// If `memory` is specified, whether it's a 64-bit memory.
+        memory64: bool,
+        /// An optional definition of `realloc` to used.
+        realloc: Option<dfg::CoreDef>,
+    },
+}
+
 /// Configuration options which can be specified as part of the canonical ABI
 /// in the component model.
 #[derive(Debug, Clone, Hash, Eq, PartialEq)]
@@ -151,18 +168,16 @@ pub struct AdapterOptions {
     pub instance: RuntimeComponentInstanceIndex,
     /// How strings are encoded.
     pub string_encoding: StringEncoding,
-    /// An optional memory definition supplied.
-    pub memory: Option<dfg::CoreExport<MemoryIndex>>,
-    /// If `memory` is specified, whether it's a 64-bit memory.
-    pub memory64: bool,
-    /// An optional definition of `realloc` to used.
-    pub realloc: Option<dfg::CoreDef>,
     /// The async callback function used by these options, if specified.
     pub callback: Option<dfg::CoreDef>,
     /// An optional definition of a `post-return` to use.
     pub post_return: Option<dfg::CoreDef>,
     /// Whether to use the async ABI for lifting or lowering.
     pub async_: bool,
+    /// The core function type that is being lifted from / lowered to.
+    pub core_type: ModuleInternedTypeIndex,
+    /// The data model used by this adapter: linear memory or GC objects.
+    pub data_model: DataModel,
 }
 
 impl<'data> Translator<'_, 'data> {
@@ -205,8 +220,8 @@ impl<'data> Translator<'_, 'data> {
             let wasm = &*self.scope_vec.push(wasm);
             if log::log_enabled!(log::Level::Trace) {
                 match wasmprinter::print_bytes(wasm) {
-                    Ok(s) => log::trace!("generated adapter module:\n{}", s),
-                    Err(e) => log::trace!("failed to print adapter module: {}", e),
+                    Ok(s) => log::trace!("generated adapter module:\n{s}"),
+                    Err(e) => log::trace!("failed to print adapter module: {e}"),
                 }
             }
 
@@ -257,6 +272,16 @@ fn fact_import_to_core_def(
     import: &fact::Import,
     ty: EntityType,
 ) -> dfg::CoreDef {
+    fn unwrap_memory(def: &dfg::CoreDef) -> dfg::CoreExport<MemoryIndex> {
+        match def {
+            dfg::CoreDef::Export(e) => e.clone().map_index(|i| match i {
+                EntityIndex::Memory(i) => i,
+                _ => unreachable!(),
+            }),
+            _ => unreachable!(),
+        }
+    }
+
     let mut simple_intrinsic = |trampoline: dfg::Trampoline| {
         let signature = ty.unwrap_func();
         let index = dfg
@@ -273,16 +298,6 @@ fn fact_import_to_core_def(
             to,
             to64,
         } => {
-            fn unwrap_memory(def: &dfg::CoreDef) -> dfg::CoreExport<MemoryIndex> {
-                match def {
-                    dfg::CoreDef::Export(e) => e.clone().map_index(|i| match i {
-                        EntityIndex::Memory(i) => i,
-                        _ => unreachable!(),
-                    }),
-                    _ => unreachable!(),
-                }
-            }
-
             let from = dfg.memories.push(unwrap_memory(from));
             let to = dfg.memories.push(unwrap_memory(to));
             let signature = ty.unwrap_func();
@@ -304,17 +319,18 @@ fn fact_import_to_core_def(
         }
         fact::Import::ResourceEnterCall => simple_intrinsic(dfg::Trampoline::ResourceEnterCall),
         fact::Import::ResourceExitCall => simple_intrinsic(dfg::Trampoline::ResourceExitCall),
-        fact::Import::SyncEnterCall => simple_intrinsic(dfg::Trampoline::SyncEnterCall),
-        fact::Import::SyncExitCall { callback } => {
-            simple_intrinsic(dfg::Trampoline::SyncExitCall {
+        fact::Import::PrepareCall { memory } => simple_intrinsic(dfg::Trampoline::PrepareCall {
+            memory: memory.as_ref().map(|v| dfg.memories.push(unwrap_memory(v))),
+        }),
+        fact::Import::SyncStartCall { callback } => {
+            simple_intrinsic(dfg::Trampoline::SyncStartCall {
                 callback: callback.clone().map(|v| dfg.callbacks.push(v)),
             })
         }
-        fact::Import::AsyncEnterCall => simple_intrinsic(dfg::Trampoline::AsyncEnterCall),
-        fact::Import::AsyncExitCall {
+        fact::Import::AsyncStartCall {
             callback,
             post_return,
-        } => simple_intrinsic(dfg::Trampoline::AsyncExitCall {
+        } => simple_intrinsic(dfg::Trampoline::AsyncStartCall {
             callback: callback.clone().map(|v| dfg.callbacks.push(v)),
             post_return: post_return.clone().map(|v| dfg.post_returns.push(v)),
         }),
@@ -380,18 +396,29 @@ impl PartitionAdapterModules {
     }
 
     fn adapter_options(&mut self, dfg: &dfg::ComponentDfg, options: &AdapterOptions) {
-        if let Some(memory) = &options.memory {
-            self.core_export(dfg, memory);
-        }
-        if let Some(def) = &options.realloc {
-            self.core_def(dfg, def);
-        }
         if let Some(def) = &options.callback {
             self.core_def(dfg, def);
         }
         if let Some(def) = &options.post_return {
             self.core_def(dfg, def);
         }
+        match &options.data_model {
+            DataModel::Gc {} => {
+                // Nothing to do here yet.
+            }
+            DataModel::LinearMemory {
+                memory,
+                memory64: _,
+                realloc,
+            } => {
+                if let Some(memory) = memory {
+                    self.core_export(dfg, memory);
+                }
+                if let Some(def) = realloc {
+                    self.core_def(dfg, def);
+                }
+            }
+        }
     }
 
     fn core_def(&mut self, dfg: &dfg::ComponentDfg, def: &dfg::CoreDef) {
