@@ -232,6 +232,16 @@ impl WasmValType {
             size => panic!("invalid int bits for WasmValType: {size}"),
         }
     }
+
+    /// Returns the contained reference type.
+    ///
+    /// Panics if the value type is not a vmgcref
+    pub fn unwrap_ref_type(&self) -> WasmRefType {
+        match self {
+            WasmValType::Ref(ref_type) => *ref_type,
+            _ => panic!("Called WasmValType::unwrap_ref_type on non-reference type"),
+        }
+    }
 }
 
 /// WebAssembly reference type -- equivalent of `wasmparser`'s RefType
@@ -436,6 +446,11 @@ pub enum WasmHeapType {
     ConcreteFunc(EngineOrModuleTypeIndex),
     NoFunc,
 
+    // Exception types.
+    Exn,
+    ConcreteExn(EngineOrModuleTypeIndex),
+    NoExn,
+
     // Continuation types.
     Cont,
     ConcreteCont(EngineOrModuleTypeIndex),
@@ -460,6 +475,7 @@ impl From<WasmHeapTopType> for WasmHeapType {
             WasmHeapTopType::Any => Self::Any,
             WasmHeapTopType::Func => Self::Func,
             WasmHeapTopType::Cont => Self::Cont,
+            WasmHeapTopType::Exn => Self::Exn,
         }
     }
 }
@@ -472,6 +488,7 @@ impl From<WasmHeapBottomType> for WasmHeapType {
             WasmHeapBottomType::None => Self::None,
             WasmHeapBottomType::NoFunc => Self::NoFunc,
             WasmHeapBottomType::NoCont => Self::NoCont,
+            WasmHeapBottomType::NoExn => Self::NoExn,
         }
     }
 }
@@ -494,6 +511,9 @@ impl fmt::Display for WasmHeapType {
             Self::ConcreteArray(i) => write!(f, "array {i}"),
             Self::Struct => write!(f, "struct"),
             Self::ConcreteStruct(i) => write!(f, "struct {i}"),
+            Self::Exn => write!(f, "exn"),
+            Self::ConcreteExn(i) => write!(f, "exn {i}"),
+            Self::NoExn => write!(f, "noexn"),
             Self::None => write!(f, "none"),
         }
     }
@@ -532,9 +552,10 @@ impl WasmHeapType {
     #[inline]
     pub fn is_vmgcref_type(&self) -> bool {
         match self.top() {
-            // All `t <: (ref null any)` and `t <: (ref null extern)` are
-            // represented as `VMGcRef`s.
-            WasmHeapTopType::Any | WasmHeapTopType::Extern => true,
+            // All `t <: (ref null any)`, `t <: (ref null extern)`,
+            // and `t <: (ref null exn)` are represented as
+            // `VMGcRef`s.
+            WasmHeapTopType::Any | WasmHeapTopType::Extern | WasmHeapTopType::Exn => true,
 
             // All `t <: (ref null func)` are not.
             WasmHeapTopType::Func => false,
@@ -572,6 +593,10 @@ impl WasmHeapType {
                 WasmHeapTopType::Cont
             }
 
+            WasmHeapType::Exn | WasmHeapType::ConcreteExn(_) | WasmHeapType::NoExn => {
+                WasmHeapTopType::Exn
+            }
+
             WasmHeapType::Any
             | WasmHeapType::Eq
             | WasmHeapType::I31
@@ -603,6 +628,10 @@ impl WasmHeapType {
                 WasmHeapBottomType::NoCont
             }
 
+            WasmHeapType::Exn | WasmHeapType::ConcreteExn(_) | WasmHeapType::NoExn => {
+                WasmHeapBottomType::NoExn
+            }
+
             WasmHeapType::Any
             | WasmHeapType::Eq
             | WasmHeapType::I31
@@ -624,6 +653,8 @@ pub enum WasmHeapTopType {
     Any,
     /// The common supertype of all function references.
     Func,
+    /// The common supertype of all exception references.
+    Exn,
     /// The common supertype of all continuation references.
     Cont,
 }
@@ -637,6 +668,8 @@ pub enum WasmHeapBottomType {
     None,
     /// The common subtype of all function references.
     NoFunc,
+    /// The common subtype of all exception references.
+    NoExn,
     /// The common subtype of all continuation references.
     NoCont,
 }
@@ -801,6 +834,15 @@ impl WasmContType {
     pub fn new(idx: EngineOrModuleTypeIndex) -> Self {
         WasmContType(idx)
     }
+
+    /// Returns the (module interned) index to the underlying function type.
+    pub fn unwrap_module_type_index(self) -> ModuleInternedTypeIndex {
+        match self.0 {
+            EngineOrModuleTypeIndex::Engine(_) => panic!("not module interned"),
+            EngineOrModuleTypeIndex::Module(idx) => idx,
+            EngineOrModuleTypeIndex::RecGroup(_) => todo!(),
+        }
+    }
 }
 
 impl TypeTrace for WasmContType {
@@ -819,6 +861,79 @@ impl TypeTrace for WasmContType {
     }
 }
 
+/// WebAssembly exception type.
+///
+/// This "exception type" is not a Wasm language-level
+/// concept. Instead, it denotes an *exception object signature* --
+/// the types of the payload values.
+///
+/// In contrast, at the Wasm language level, exception objects are
+/// associated with specific tags, and these tags refer to their
+/// signatures (function types). However, tags are *nominal*: like
+/// memories and tables, a separate instance of a tag exists for every
+/// instance of the defining module, and these tag instances can be
+/// imported and exported. At runtime we handle tags like we do
+/// memories and tables, but these runtime instances do not exist in
+/// the type system here.
+///
+/// Because the Wasm type system does not have concrete `exn` types
+/// (i.e., the heap-type lattice has only top `exn` and bottom
+/// `noexn`), we are free to decide what we mean by "concrete type"
+/// here. Thus, we define an "exception type" to refer to the
+/// type-level *signature*. When a particular *exception object* is
+/// created in a store, it can be associated with a particular *tag
+/// instance* also in that store, and the compatibility is checked
+/// (the tag's function type must match the function type in the
+/// associated WasmExnType).
+#[derive(Debug, Clone, Eq, PartialEq, Hash, Serialize, Deserialize)]
+pub struct WasmExnType {
+    /// The function type from which we get our signature. We hold
+    /// this directly so that we can efficiently derive a FuncType
+    /// without re-interning the field types.
+    pub func_ty: EngineOrModuleTypeIndex,
+    /// The fields (payload values) that make up this exception type.
+    ///
+    /// While we could obtain these by looking up the `func_ty` above,
+    /// we also need to be able to derive a GC object layout from this
+    /// type descriptor without referencing other type descriptors; so
+    /// we directly inline the information here.
+    pub fields: Box<[WasmFieldType]>,
+}
+
+impl fmt::Display for WasmExnType {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        write!(f, "(exn ({})", self.func_ty)?;
+        for ty in self.fields.iter() {
+            write!(f, " {ty}")?;
+        }
+        write!(f, ")")
+    }
+}
+
+impl TypeTrace for WasmExnType {
+    fn trace<F, E>(&self, func: &mut F) -> Result<(), E>
+    where
+        F: FnMut(EngineOrModuleTypeIndex) -> Result<(), E>,
+    {
+        func(self.func_ty)?;
+        for f in self.fields.iter() {
+            f.trace(func)?;
+        }
+        Ok(())
+    }
+
+    fn trace_mut<F, E>(&mut self, func: &mut F) -> Result<(), E>
+    where
+        F: FnMut(&mut EngineOrModuleTypeIndex) -> Result<(), E>,
+    {
+        func(&mut self.func_ty)?;
+        for f in self.fields.iter_mut() {
+            f.trace_mut(func)?;
+        }
+        Ok(())
+    }
+}
+
 /// Represents storage types introduced in the GC spec for array and struct fields.
 #[derive(Debug, Copy, Clone, Eq, PartialEq, Hash, Serialize, Deserialize)]
 pub enum WasmStorageType {
@@ -1008,6 +1123,7 @@ pub enum WasmCompositeInnerType {
     Func(WasmFuncType),
     Struct(WasmStructType),
     Cont(WasmContType),
+    Exn(WasmExnType),
 }
 
 impl fmt::Display for WasmCompositeInnerType {
@@ -1017,6 +1133,7 @@ impl fmt::Display for WasmCompositeInnerType {
             Self::Func(ty) => fmt::Display::fmt(ty, f),
             Self::Struct(ty) => fmt::Display::fmt(ty, f),
             Self::Cont(ty) => fmt::Display::fmt(ty, f),
+            Self::Exn(ty) => fmt::Display::fmt(ty, f),
         }
     }
 }
@@ -1094,6 +1211,24 @@ impl WasmCompositeInnerType {
     pub fn unwrap_cont(&self) -> &WasmContType {
         self.as_cont().unwrap()
     }
+
+    #[inline]
+    pub fn is_exn(&self) -> bool {
+        matches!(self, Self::Exn(_))
+    }
+
+    #[inline]
+    pub fn as_exn(&self) -> Option<&WasmExnType> {
+        match self {
+            Self::Exn(f) => Some(f),
+            _ => None,
+        }
+    }
+
+    #[inline]
+    pub fn unwrap_exn(&self) -> &WasmExnType {
+        self.as_exn().unwrap()
+    }
 }
 
 impl TypeTrace for WasmCompositeType {
@@ -1106,6 +1241,7 @@ impl TypeTrace for WasmCompositeType {
             WasmCompositeInnerType::Func(f) => f.trace(func),
             WasmCompositeInnerType::Struct(a) => a.trace(func),
             WasmCompositeInnerType::Cont(c) => c.trace(func),
+            WasmCompositeInnerType::Exn(e) => e.trace(func),
         }
     }
 
@@ -1118,6 +1254,7 @@ impl TypeTrace for WasmCompositeType {
             WasmCompositeInnerType::Func(f) => f.trace_mut(func),
             WasmCompositeInnerType::Struct(a) => a.trace_mut(func),
             WasmCompositeInnerType::Cont(c) => c.trace_mut(func),
+            WasmCompositeInnerType::Exn(e) => e.trace_mut(func),
         }
     }
 }
@@ -1237,6 +1374,26 @@ impl WasmSubType {
         assert!(!self.composite_type.shared);
         self.composite_type.inner.unwrap_cont()
     }
+
+    #[inline]
+    pub fn is_exn(&self) -> bool {
+        self.composite_type.inner.is_exn() && !self.composite_type.shared
+    }
+
+    #[inline]
+    pub fn as_exn(&self) -> Option<&WasmExnType> {
+        if self.composite_type.shared {
+            None
+        } else {
+            self.composite_type.inner.as_exn()
+        }
+    }
+
+    #[inline]
+    pub fn unwrap_exn(&self) -> &WasmExnType {
+        assert!(!self.composite_type.shared);
+        self.composite_type.inner.unwrap_exn()
+    }
 }
 
 impl TypeTrace for WasmSubType {
@@ -2239,13 +2396,10 @@ pub trait TypeConvert {
                 wasmparser::AbstractHeapType::Array => WasmHeapType::Array,
                 wasmparser::AbstractHeapType::Struct => WasmHeapType::Struct,
                 wasmparser::AbstractHeapType::None => WasmHeapType::None,
-
-                wasmparser::AbstractHeapType::Exn
-                | wasmparser::AbstractHeapType::NoExn
-                | wasmparser::AbstractHeapType::Cont
-                | wasmparser::AbstractHeapType::NoCont => {
-                    return Err(wasm_unsupported!("unsupported heap type {ty:?}"));
-                }
+                wasmparser::AbstractHeapType::Cont => WasmHeapType::Cont,
+                wasmparser::AbstractHeapType::NoCont => WasmHeapType::NoCont,
+                wasmparser::AbstractHeapType::Exn => WasmHeapType::Exn,
+                wasmparser::AbstractHeapType::NoExn => WasmHeapType::NoExn,
             },
             _ => return Err(wasm_unsupported!("unsupported heap type {ty:?}")),
         })
