@@ -10,7 +10,6 @@
 //      _padding: u32, // (On 64-bit systems)
 //      vm_store_context: *const VMStoreContext,
 //      builtin_functions: *mut VMBuiltinFunctionsArray,
-//      callee: *mut VMFunctionBody,
 //      epoch_ptr: *mut AtomicU64,
 //      gc_heap_data: *mut T, // Collector-specific pointer
 //      type_ids: *const VMSharedTypeIndex,
@@ -217,6 +216,11 @@ pub trait PtrSize {
         self.vmstore_context_last_wasm_exit_pc() + self.size()
     }
 
+    /// Return the offset of the `stack_chain` field of `VMStoreContext`.
+    fn vmstore_context_stack_chain(&self) -> u8 {
+        self.vmstore_context_last_wasm_entry_fp() + self.size()
+    }
+
     // Offsets within `VMMemoryDefinition`
 
     /// The offset of the `base` field.
@@ -254,6 +258,122 @@ pub trait PtrSize {
         .unwrap()
     }
 
+    /// Return the size of `VMStackChain`.
+    fn size_of_vmstack_chain(&self) -> u8 {
+        2 * self.size()
+    }
+
+    // Offsets within `VMStackLimits`
+
+    /// Return the offset of `VMStackLimits::stack_limit`.
+    fn vmstack_limits_stack_limit(&self) -> u8 {
+        0
+    }
+
+    /// Return the offset of `VMStackLimits::last_wasm_entry_fp`.
+    fn vmstack_limits_last_wasm_entry_fp(&self) -> u8 {
+        self.size()
+    }
+
+    // Offsets within `VMHostArray`
+
+    /// Return the offset of `VMHostArray::length`.
+    fn vmhostarray_length(&self) -> u8 {
+        0
+    }
+
+    /// Return the offset of `VMHostArray::capacity`.
+    fn vmhostarray_capacity(&self) -> u8 {
+        4
+    }
+
+    /// Return the offset of `VMHostArray::data`.
+    fn vmhostarray_data(&self) -> u8 {
+        8
+    }
+
+    /// Return the size of `VMHostArray`.
+    fn size_of_vmhostarray(&self) -> u8 {
+        8 + self.size()
+    }
+
+    // Offsets within `VMCommonStackInformation`
+
+    /// Return the offset of `VMCommonStackInformation::limits`.
+    fn vmcommon_stack_information_limits(&self) -> u8 {
+        0 * self.size()
+    }
+
+    /// Return the offset of `VMCommonStackInformation::state`.
+    fn vmcommon_stack_information_state(&self) -> u8 {
+        2 * self.size()
+    }
+
+    /// Return the offset of `VMCommonStackInformation::handlers`.
+    fn vmcommon_stack_information_handlers(&self) -> u8 {
+        u8::try_from(align(
+            self.vmcommon_stack_information_state() as u32 + 4,
+            u32::from(self.size()),
+        ))
+        .unwrap()
+    }
+
+    /// Return the offset of `VMCommonStackInformation::first_switch_handler_index`.
+    fn vmcommon_stack_information_first_switch_handler_index(&self) -> u8 {
+        self.vmcommon_stack_information_handlers() + self.size_of_vmhostarray()
+    }
+
+    /// Return the size of `VMCommonStackInformation`.
+    fn size_of_vmcommon_stack_information(&self) -> u8 {
+        u8::try_from(align(
+            self.vmcommon_stack_information_first_switch_handler_index() as u32 + 4,
+            u32::from(self.size()),
+        ))
+        .unwrap()
+    }
+
+    // Offsets within `VMContRef`
+
+    /// Return the offset of `VMContRef::common_stack_information`.
+    fn vmcontref_common_stack_information(&self) -> u8 {
+        0 * self.size()
+    }
+
+    /// Return the offset of `VMContRef::parent_chain`.
+    fn vmcontref_parent_chain(&self) -> u8 {
+        u8::try_from(align(
+            (self.vmcontref_common_stack_information() + self.size_of_vmcommon_stack_information())
+                as u32,
+            u32::from(self.size()),
+        ))
+        .unwrap()
+    }
+
+    /// Return the offset of `VMContRef::last_ancestor`.
+    fn vmcontref_last_ancestor(&self) -> u8 {
+        self.vmcontref_parent_chain() + 2 * self.size()
+    }
+
+    /// Return the offset of `VMContRef::revision`.
+    fn vmcontref_revision(&self) -> u8 {
+        self.vmcontref_last_ancestor() + self.size()
+    }
+
+    /// Return the offset of `VMContRef::stack`.
+    fn vmcontref_stack(&self) -> u8 {
+        self.vmcontref_revision() + 8
+    }
+
+    /// Return the offset of `VMContRef::args`.
+    fn vmcontref_args(&self) -> u8 {
+        self.vmcontref_stack() + 3 * self.size()
+    }
+
+    /// Return the offset of `VMContRef::values`.
+    fn vmcontref_values(&self) -> u8 {
+        self.vmcontref_args() + self.size_of_vmhostarray()
+    }
+
     /// Return the offset to the `magic` value in this `VMContext`.
     #[inline]
     fn vmctx_magic(&self) -> u8 {
@@ -275,17 +395,11 @@ pub trait PtrSize {
         self.vmctx_store_context() + self.size()
     }
 
-    /// Return the offset to the `callee` member in this `VMContext`.
-    #[inline]
-    fn vmctx_callee(&self) -> u8 {
-        self.vmctx_builtin_functions() + self.size()
-    }
-
     /// Return the offset to the `*const AtomicU64` epoch-counter
     /// pointer.
     #[inline]
     fn vmctx_epoch_ptr(&self) -> u8 {
-        self.vmctx_callee() + self.size()
+        self.vmctx_builtin_functions() + self.size()
     }
 
     /// Return the offset to the `*mut T` collector-specific data.
@@ -592,6 +706,18 @@ impl<P: PtrSize> VMOffsets<P> {
         0 * self.pointer_size()
     }
 
+    /// The offset of the `vmctx` field.
+    #[inline]
+    pub fn vmtable_import_vmctx(&self) -> u8 {
+        1 * self.pointer_size()
+    }
+
+    /// The offset of the `index` field.
+    #[inline]
+    pub fn vmtable_import_index(&self) -> u8 {
+        2 * self.pointer_size()
+    }
+
     /// Return the size of `VMTableImport`.
     #[inline]
     pub fn size_of_vmtable_import(&self) -> u8 {
@@ -633,6 +759,18 @@ impl<P: PtrSize> VMOffsets<P> {
         0 * self.pointer_size()
     }
 
+    /// The offset of the `vmctx` field.
+    #[inline]
+    pub fn vmmemory_import_vmctx(&self) -> u8 {
+        1 * self.pointer_size()
+    }
+
+    /// The offset of the `index` field.
+    #[inline]
+    pub fn vmmemory_import_index(&self) -> u8 {
+        2 * self.pointer_size()
+    }
+
     /// Return the size of `VMMemoryImport`.
     #[inline]
     pub fn size_of_vmmemory_import(&self) -> u8 {
@@ -909,6 +1047,28 @@ impl<P: PtrSize> VMOffsets<P> {
     }
 }
 
+/// Offsets for `VMGcHeader`.
+impl<P: PtrSize> VMOffsets<P> {
+    /// Return the offset for the `VMGcHeader::kind` field.
+    #[inline]
+    pub fn vm_gc_header_kind(&self) -> u32 {
+        0
+    }
+
+    /// Return the offset for the `VMGcHeader`'s reserved bits.
+    #[inline]
+    pub fn vm_gc_header_reserved_bits(&self) -> u32 {
+        // NB: The reserved bits are the unused `VMGcKind` bits.
+        self.vm_gc_header_kind()
+    }
+
+    /// Return the offset for the `VMGcHeader::ty` field.
+    #[inline]
+    pub fn vm_gc_header_ty(&self) -> u32 {
+        self.vm_gc_header_kind() + 4
+    }
+}
+
 /// Offsets for `VMDrcHeader`.
 ///
 /// Should only be used when the DRC collector is enabled.
@@ -918,22 +1078,11 @@ impl<P: PtrSize> VMOffsets<P> {
     pub fn vm_drc_header_ref_count(&self) -> u32 {
         8
     }
-}
 
-/// Offsets for `VMGcRefActivationsTable`.
-///
-/// These should only be used when the DRC collector is enabled.
-impl<P: PtrSize> VMOffsets<P> {
-    /// Return the offset for `VMGcRefActivationsTable::next`.
+    /// Return the offset for `VMDrcHeader::next_over_approximated_stack_root`.
     #[inline]
-    pub fn vm_gc_ref_activation_table_next(&self) -> u32 {
-        0
-    }
-
-    /// Return the offset for `VMGcRefActivationsTable::end`.
-    #[inline]
-    pub fn vm_gc_ref_activation_table_end(&self) -> u32 {
-        self.pointer_size().into()
+    pub fn vm_drc_header_next_over_approximated_stack_root(&self) -> u32 {
+        self.vm_drc_header_ref_count() + 8
     }
 }
 
